

part of firebase_dart.api;


class ServerValue {
  static const TIMESTAMP = const {".sv":"timestamp"};
}

class Firebase extends Query {

  final Repo repo;
  final util.Path path;

  const Firebase.fromRepo(this.repo, [this.path])
      : super(repo, path, QueryParams.DEFAULT, false)

  factory Firebase(Uri url, [pathOrContext]) {
    var parsedUrl = util.parseRepoInfo(url);
    var repoInfo = parsedUrl.repoInfo;
    validation.validateUrl(parsedUrl);

    RepoManager repoManager;
    if (pathOrContext != null) {
      if (pathOrContext is RepoManager) {
        repoManager = pathOrContext;
      } else {
        if (pathOrContext is String) {
          repoManager = RepoManager.getInstance();
          repoInfo.persistenceKey = pathOrContext;
        } else {
          throw new Exception(
              "Expected a valid Firebase.Context for second argument to new Firebase()");
        }
      }
    } else {
      repoManager = RepoManager.getInstance();
    }
    return new Firebase.fromRepo(repoManager.getRepo(repoInfo), parsedUrl.path);
  }

  static void goOffline() {
    RepoManager.getInstance().interrupt();
  }

  static void goOnline() {
    RepoManager.getInstance().resume();
  }

  static bool enableLogging;

  String get key => path.isEmpty ? null : path.back;

  Firebase child(path) {
    if (path is num) {
      path = path.toString();
    } else {
      if (path is util.Path) {
        if (this.path.front == null) {
          validation.validateRootPathString("Firebase.child", 1, path, false);
        } else {
          validation.validatePathString("Firebase.child", 1, path, false);
        }
      }
    }
    return new Firebase.fromRepo(this.repo, this.path.child(path));
  }

  Firebase get parent {
    var parentPath = this.path.parent;
    return parentPath == null ? null : new Firebase.fromRepo(this.repo, parentPath);
  }

  Firebase get root {
    var ref = this;
    while (ref.parent != null) {
      ref = ref.parent;
    }
    return ref;
  }

  Future set(newVal, [OnCallback onComplete]) {
    validation.validateWritablePath("Firebase.set", this.path);
    validation.validateFirebaseDataArg(
        "Firebase.set", 1, newVal, this.path, false);
    var deferred = new fb.util.promise.Deferred;

    repo.setWithPriority(path, newVal, null, deferred.wrapCallback(onComplete));
    return deferred.promise;
  }

  Future update(objectToMerge, [OnCallback onComplete]) {
    validation.validateWritablePath("Firebase.update", this.path);
    if (objectToMerge is List) {
      objectToMerge = objectToMerge.asMap();

      util.warn("Passing an Array to Firebase.update() is deprecated. "
          "Use set() if you want to overwrite the existing data, or "
          "an Object with integer keys if you really do want to "
          "only update some of the children.");
    }
    validation.validateFirebaseMergeDataArg(
        "Firebase.update", 1, objectToMerge, this.path, false);
    var deferred = new fb.util.promise.Deferred;
    repo.update(this.path, objectToMerge, deferred.wrapCallback(onComplete));
    return deferred.promise;
  }

  Future setWithPriority(newVal, newPriority, [OnCallback onComplete]) {
    validation.validateWritablePath("Firebase.setWithPriority", this.path);
    validation.validateFirebaseDataArg(
        "Firebase.setWithPriority", 1, newVal, this.path, false);
    validation.validatePriority("Firebase.setWithPriority", 2, newPriority, false);
    if (this.key == ".length" || this.key == ".keys") {
      throw new Exception("Firebase.setWithPriority failed: "
          "$key is a read-only object.");
    }
    var deferred = new fb.util.promise.Deferred;
    repo.setWithPriority(
        this.path, newVal, newPriority, deferred.wrapCallback(onComplete));
    return deferred.promise;
  }

  remove([OnCallback onComplete]) {
    validation.validateWritablePath("Firebase.remove", this.path);
    return this.set(null, onComplete);
  }

  transaction(transactionUpdate,
      [OnCallback onComplete, bool applyLocally = true]) {
    validation.validateWritablePath("Firebase.transaction", this.path);
    if (this.key == ".length" || this.key == ".keys") {
      throw "Firebase.transaction failed: $key is a read-only object.";
    }
    var deferred = new fb.util.promise.Deferred;
    if (onComplete != null) {
      fb.util.promise.attachDummyErrorHandler(deferred.promise);
    }
    var promiseComplete = (error, committed, snapshot)
    {
      if (error) {
        deferred.reject(error);
      } else {
        deferred.resolve(new fb.api.TransactionResult(committed, snapshot));
      }
      if (goog.isFunction(opt_onComplete)) {
        opt_onComplete(error, committed, snapshot);
      }
    };
    this.repo.startTransaction(
        this.path, transactionUpdate, promiseComplete, opt_applyLocally);
    return deferred.promise;
  }

  setPriority(priority, opt_onComplete) {
    fb.util.validation.validateArgCount(
        "Firebase.setPriority", 1, 2, arguments.length);
    fb.core.util.validation.validateWritablePath(
        "Firebase.setPriority", this.path);
    fb.core.util.validation.validatePriority(
        "Firebase.setPriority", 1, priority, false);
    fb.util.validation.validateCallback(
        "Firebase.setPriority", 2, opt_onComplete, true);
    var deferred = new fb.util.promise.Deferred;
    this.repo.setWithPriority(this.path.child(".priority"), priority, null,
        deferred.wrapCallback(opt_onComplete));
    return deferred.promise;
  }

  push(opt_value, opt_onComplete) {
    fb.util.validation.validateArgCount(
        "Firebase.push", 0, 2, arguments.length);
    fb.core.util.validation.validateWritablePath("Firebase.push", this.path);
    fb.core.util.validation.validateFirebaseDataArg(
        "Firebase.push", 1, opt_value, this.path, true);
    fb.util.validation.validateCallback(
        "Firebase.push", 2, opt_onComplete, true);
    var now = this.repo.serverTime();
    var name = fb.core.util.nextPushId(now);
    var pushedRef = this.child(name);
    if (goog.isDefAndNotNull(opt_value)) {
      var self = this;
      var promise = pushedRef.set(opt_value, opt_onComplete).then(() {
      return self.child(name);
      });
      pushedRef.then = goog.bind(promise.then, promise);
      (pushedRef)["catch"] = goog.bind(promise.then, promise, void 0);
      if (goog.isFunction(opt_onComplete)) {
        fb.util.promise.attachDummyErrorHandler(promise);
      }
    }
    return pushedRef;
  }

  onDisconnect() {
    fb.core.util.validation.validateWritablePath(
        "Firebase.onDisconnect", this.path);
    return new fb.api.OnDisconnect(this.repo, this.path);
  }

  auth(cred, opt_onComplete, opt_onCancel) {
    fb.core.util.warn("FirebaseRef.auth() being deprecated. " +
        "Please use FirebaseRef.authWithCustomToken() instead.");
    fb.util.validation.validateArgCount(
        "Firebase.auth", 1, 3, arguments.length);
    fb.core.util.validation.validateCredential("Firebase.auth", 1, cred, false);
    fb.util.validation.validateCallback(
        "Firebase.auth", 2, opt_onComplete, true);
    fb.util.validation.validateCallback(
        "Firebase.auth", 3, opt_onComplete, true);
    var clientOptions = {};
    clientOptions[fb.login.Constants.CLIENT_OPTION_SESSION_PERSISTENCE] =
    "none";
    var self = this;
    var deferred = new fb.util.promise.Deferred;
    self.repo.auth.authenticate(
        cred, {}, clientOptions, deferred.wrapCallback(opt_onComplete),
        opt_onCancel);
    return deferred.promise;
  }

  unauth(opt_onComplete) {
    fb.util.validation.validateArgCount(
        "Firebase.unauth", 0, 1, arguments.length);
    fb.util.validation.validateCallback(
        "Firebase.unauth", 1, opt_onComplete, true);
    var deferred = new fb.util.promise.Deferred;
    this.repo.auth.unauthenticate(deferred.wrapCallback(opt_onComplete));
    return deferred.promise;
  }

  getAuth() {
    fb.util.validation.validateArgCount(
        "Firebase.getAuth", 0, 0, arguments.length);
    return this.repo.auth.getAuth();
  }

  onAuth(callback, opt_context) {
    fb.util.validation.validateArgCount(
        "Firebase.onAuth", 1, 2, arguments.length);
    fb.util.validation.validateCallback("Firebase.onAuth", 1, callback, false);
    fb.util.validation.validateContextObject(
        "Firebase.onAuth", 2, opt_context, true);
    this.repo.auth.on("auth_status", callback, opt_context);
  }

  offAuth(callback, opt_context) {
    fb.util.validation.validateArgCount(
        "Firebase.offAuth", 1, 2, arguments.length);
    fb.util.validation.validateCallback("Firebase.offAuth", 1, callback, false);
    fb.util.validation.validateContextObject(
        "Firebase.offAuth", 2, opt_context, true);
    this.repo.auth.off("auth_status", callback, opt_context);
  }

  authWithCustomToken(token, opt_onComplete, opt_options) {
    fb.util.validation.validateArgCount(
        "Firebase.authWithCustomToken", 1, 3, arguments.length);
    if (arguments.length == 2 && fb.util.obj.isNonNullObject(opt_onComplete)) {
      opt_options = opt_onComplete;
      opt_onComplete = (_) => null;
    }
    fb.core.util.validation.validateCredential(
        "Firebase.authWithCustomToken", 1, token, false);
    fb.util.validation.validateCallback(
        "Firebase.authWithCustomToken", 2, opt_onComplete, true);
    fb.core.util.validation.validateObject(
        "Firebase.authWithCustomToken", 3, opt_options, true);
    var deferred = new fb.util.promise.Deferred;
    this.repo.auth.authenticate(
        token, {}, opt_options || {}, deferred.wrapCallback(opt_onComplete));
    return deferred.promise;
  }

  authWithOAuthPopup(provider, opt_onComplete, opt_options) {
    fb.util.validation.validateArgCount(
        "Firebase.authWithOAuthPopup", 1, 3, arguments.length);
    if (arguments.length == 2 && fb.util.obj.isNonNullObject(opt_onComplete)) {
      opt_options = opt_onComplete;
      opt_onComplete = (_) => null;
    }
    fb.core.util.validation.validateAuthProvider(
        "Firebase.authWithOAuthPopup", 1, provider, false);
    fb.util.validation.validateCallback(
        "Firebase.authWithOAuthPopup", 2, opt_onComplete, true);
    fb.core.util.validation.validateObject(
        "Firebase.authWithOAuthPopup", 3, opt_options, true);
    var deferred = new fb.util.promise.Deferred;
    this.repo.auth.authWithPopup(
        provider, opt_options, deferred.wrapCallback(opt_onComplete));
    return deferred.promise;
  }

  authWithOAuthRedirect(provider, opt_onErr, opt_options) {
    fb.util.validation.validateArgCount(
        "Firebase.authWithOAuthRedirect", 1, 3, arguments.length);
    if (arguments.length == 2 && fb.util.obj.isNonNullObject(opt_onErr)) {
      opt_options = opt_onErr;
      opt_onErr = (_) => null;
    }
    fb.core.util.validation.validateAuthProvider(
        "Firebase.authWithOAuthRedirect", 1, provider, false);
    fb.util.validation.validateCallback(
        "Firebase.authWithOAuthRedirect", 2, opt_onErr, false);
    fb.core.util.validation.validateObject(
        "Firebase.authWithOAuthRedirect", 3, opt_options, true);
    var deferred = new fb.util.promise.Deferred;
    this.repo.auth.authWithRedirect(
        provider, opt_options, deferred.wrapCallback(opt_onErr));
    return deferred.promise;
  }

  authWithOAuthToken(provider, params, opt_onComplete, opt_options) {
    fb.util.validation.validateArgCount(
        "Firebase.authWithOAuthToken", 2, 4, arguments.length);
    if (arguments.length == 3 && fb.util.obj.isNonNullObject(opt_onComplete)) {
      opt_options = opt_onComplete;
      opt_onComplete = (_) => null;
    }
    fb.core.util.validation.validateAuthProvider(
        "Firebase.authWithOAuthToken", 1, provider, false);
    fb.util.validation.validateCallback(
        "Firebase.authWithOAuthToken", 3, opt_onComplete, true);
    fb.core.util.validation.validateObject(
        "Firebase.authWithOAuthToken", 4, opt_options, true);
    var deferred = new fb.util.promise.Deferred;
    if (goog.isString(params)) {
      fb.core.util.validation.validateString(
          "Firebase.authWithOAuthToken", 2, params, false);
      this.repo.auth.authWithCredential(
          provider + "/token", {"access_token":params}, opt_options,
          deferred.wrapCallback(opt_onComplete));
    } else {
      fb.core.util.validation.validateObject(
          "Firebase.authWithOAuthToken", 2, params, false);
      this.repo.auth.authWithCredential(
          provider + "/token", params, opt_options,
          deferred.wrapCallback(opt_onComplete));
    }
    return deferred.promise;
  }

  authAnonymously(opt_onComplete, opt_options) {
    fb.util.validation.validateArgCount(
        "Firebase.authAnonymously", 0, 2, arguments.length);
    if (arguments.length == 1 && fb.util.obj.isNonNullObject(opt_onComplete)) {
      opt_options = opt_onComplete;
      opt_onComplete = (_) => null;
    }
    fb.util.validation.validateCallback(
        "Firebase.authAnonymously", 1, opt_onComplete, true);
    fb.core.util.validation.validateObject(
        "Firebase.authAnonymously", 2, opt_options, true);
    var deferred = new fb.util.promise.Deferred;
    this.repo.auth.authWithCredential(
        "anonymous", {}, opt_options, deferred.wrapCallback(opt_onComplete));
    return deferred.promise;
  }


  authWithPassword(params, opt_onComplete, opt_options) {
    fb.util.validation.validateArgCount(
        "Firebase.authWithPassword", 1, 3, arguments.length);
    if (arguments.length == 2 && fb.util.obj.isNonNullObject(opt_onComplete)) {
      opt_options = opt_onComplete;
      opt_onComplete = (_) => null;
    }
    fb.core.util.validation.validateObject(
        "Firebase.authWithPassword", 1, params, false);
    fb.core.util.validation.validateObjectContainsKey(
        "Firebase.authWithPassword", 1, params, "email", false, "string");
    fb.core.util.validation.validateObjectContainsKey(
        "Firebase.authWithPassword", 1, params, "password", false, "string");
    fb.util.validation.validateCallback(
        "Firebase.authWithPassword", 2, opt_onComplete, true);
    fb.core.util.validation.validateObject(
        "Firebase.authWithPassword", 3, opt_options, true);
    var deferred = new fb.util.promise.Deferred;
    this.repo.auth.authWithCredential(
        "password", params, opt_options, deferred.wrapCallback(opt_onComplete));
    return deferred.promise;
  }

  createUser(params, opt_onComplete) {
    fb.util.validation.validateArgCount(
        "Firebase.createUser", 1, 2, arguments.length);
    fb.core.util.validation.validateObject(
        "Firebase.createUser", 1, params, false);
    fb.core.util.validation.validateObjectContainsKey(
        "Firebase.createUser", 1, params, "email", false, "string");
    fb.core.util.validation.validateObjectContainsKey(
        "Firebase.createUser", 1, params, "password", false, "string");
    fb.util.validation.validateCallback(
        "Firebase.createUser", 2, opt_onComplete, true);
    var deferred = new fb.util.promise.Deferred;
    this.repo.auth.createUser(params, deferred.wrapCallback(opt_onComplete));
    return deferred.promise;
  }

  removeUser(params, opt_onComplete) {
    fb.util.validation.validateArgCount(
        "Firebase.removeUser", 1, 2, arguments.length);
    fb.core.util.validation.validateObject(
        "Firebase.removeUser", 1, params, false);
    fb.core.util.validation.validateObjectContainsKey(
        "Firebase.removeUser", 1, params, "email", false, "string");
    fb.core.util.validation.validateObjectContainsKey(
        "Firebase.removeUser", 1, params, "password", false, "string");
    fb.util.validation.validateCallback(
        "Firebase.removeUser", 2, opt_onComplete, true);
    var deferred = new fb.util.promise.Deferred;
    this.repo.auth.removeUser(params, deferred.wrapCallback(opt_onComplete));
    return deferred.promise;
  }

  changePassword(params, opt_onComplete) {
    fb.util.validation.validateArgCount(
        "Firebase.changePassword", 1, 2, arguments.length);
    fb.core.util.validation.validateObject(
        "Firebase.changePassword", 1, params, false);
    fb.core.util.validation.validateObjectContainsKey(
        "Firebase.changePassword", 1, params, "email", false, "string");
    fb.core.util.validation.validateObjectContainsKey(
        "Firebase.changePassword", 1, params, "oldPassword", false, "string");
    fb.core.util.validation.validateObjectContainsKey(
        "Firebase.changePassword", 1, params, "newPassword", false, "string");
    fb.util.validation.validateCallback(
        "Firebase.changePassword", 2, opt_onComplete, true);
    var deferred = new fb.util.promise.Deferred;
    this.repo.auth.changePassword(
        params, deferred.wrapCallback(opt_onComplete));
    return deferred.promise;
  }

  changeEmail(params, opt_onComplete) {
    fb.util.validation.validateArgCount(
        "Firebase.changeEmail", 1, 2, arguments.length);
    fb.core.util.validation.validateObject(
        "Firebase.changeEmail", 1, params, false);
    fb.core.util.validation.validateObjectContainsKey(
        "Firebase.changeEmail", 1, params, "oldEmail", false, "string");
    fb.core.util.validation.validateObjectContainsKey(
        "Firebase.changeEmail", 1, params, "newEmail", false, "string");
    fb.core.util.validation.validateObjectContainsKey(
        "Firebase.changeEmail", 1, params, "password", false, "string");
    fb.util.validation.validateCallback(
        "Firebase.changeEmail", 2, opt_onComplete, true);
    var deferred = new fb.util.promise.Deferred;
    this.repo.auth.changeEmail(params, deferred.wrapCallback(opt_onComplete));
    return deferred.promise;
  }

  resetPassword(params, opt_onComplete) {
    fb.core.util.validation.validateObject(
        "Firebase.resetPassword", 1, params, false);
    validation.validateObjectContainsKey(
        "Firebase.resetPassword", 1, params, "email", false, "string");
    var deferred = new fb.util.promise.Deferred;
    this.repo.auth.resetPassword(params, deferred.wrapCallback(opt_onComplete));
    return deferred.promise;
  }
}
