
part of firebase_dart.api;



enum EventType {value, child_added, child_removed, child_changed, child_moved}

typedef OnCallback(v);
typedef CancelCallback(Exception e);

class Query {

  final Repo repo;
  final util.Path path;
  QueryParams _queryParams;
  bool _orderByCalled;

  Query(this.repo, this.path, this._queryParams, this._orderByCalled);

  _validateQueryEndpoints(params) {
    var startNode = null;
    var endNode = null;
    if (params.hasStart()) {
      startNode = params.getIndexStartValue();
    }
    if (params.hasEnd()) {
      endNode = params.getIndexEndValue();
    }
    if (params.getIndex() == fb.core.snap.KeyIndex) {
      var tooManyArgsError = "Query: When ordering by key, you may only pass one argument to " +
          "startAt(), endAt(), or equalTo().";
      var wrongArgTypeError = "Query: When ordering by key, the argument passed to startAt(), endAt()," +
          "or equalTo() must be a string.";
      if (params.hasStart()) {
        var startName = params.getIndexStartName();
        if (startName != fb.core.util.MIN_NAME) {
          throw new Error(tooManyArgsError);
        } else {
          if (startNode is String) {
            throw new Error(wrongArgTypeError);
          }
        }
      }
      if (params.hasEnd()) {
        var endName = params.getIndexEndName();
        if (endName != fb.core.util.MAX_NAME) {
          throw new Error(tooManyArgsError);
        } else {
          if (endNode is String) {
            throw new Error(wrongArgTypeError);
          }
        }
      }
    } else {
      if (params.getIndex() == fb.core.snap.PriorityIndex) {
        if (startNode != null &&
            !fb.core.util.validation.isValidPriority(startNode) ||
            endNode != null &&
                !fb.core.util.validation.isValidPriority(endNode)) {
          throw new Error(
              "Query: When ordering by priority, the first argument passed to startAt(), " +
                  "endAt(), or equalTo() must be a valid priority value (null, a number, or a string).");
        }
      } else {
        assert(params.getIndex() is fb.core.snap.PathIndex ||
            params.getIndex() ==
                fb.core.snap.ValueIndex, "unknown index type.");
        if (startNode != null && startNode is Map /*typeof == "object"*/ ||
            endNode != null && endNode is Map /*typeof == "object"*/) {
          throw new Error(
              "Query: First argument passed to startAt(), endAt(), or equalTo() cannot be " +
                  "an object.");
        }
      }
    }
  }

  _validateLimit(params) {
    if (params.hasStart() && params.hasEnd() && params.hasLimit() &&
        !params.hasAnchoredLimit()) {
      throw new Error(
          "Query: Can't combine startAt(), endAt(), and limit(). Use limitToFirst() or limitToLast() instead.");
    }
  }

  _validateNoPreviousOrderByCall(fnName) {
    if (this._orderByCalled == true) {
      throw new Error(fnName + ": You can't combine multiple orderBy calls.");
    }
  }

  QueryParams get queryParams => _queryParams;

  Firebase get ref => new Firebase.fromRepo(this.repo, this.path);

  OnCallback on(EventType eventType, OnCallback callback, {CancelCallback cancelCallback, context}) {
    if (eventType == EventType.value) {
      this.onValueEvent(callback, cancelCallback, context);
    } else {
      var callbacks = {};
      callbacks[eventType] = callback;
      this.onChildEvent(callbacks, cancelCallback, context);
    }
    return callback;
  }

  void onValueEvent(OnCallback callback, CancelCallback cancelCallback, context) {
    var container = new ValueEventRegistration(callback, cancelCallback, context);
    this.repo.addEventCallbackForQuery(this, container);
  }

  void onChildEvent(Map<EventType,OnCallback> callbacks, CancelCallback cancelCallback, context) {
    var container = new ChildEventRegistration(callbacks, cancelCallback, context);
    this.repo.addEventCallbackForQuery(this, container);
  }

  void off([EventType eventType, OnCallback callback, context]) {
    var container = null;
    var callbacks = null;
    if (eventType == EventType.value) {
      container = new ValueEventRegistration(callback, null, context);
    } else {
      if (eventType!=null) {
        if (callback!=null) {
          callbacks = {};
          callbacks[eventType] = callback;
        }
        container = new ChildEventRegistration(callbacks, null, context || null);
      }
    }
    this.repo.removeEventCallbackForQuery(this, container);
  }

  once(EventType eventType, OnCallback userCallback, {cancelCallback, context}) {
    var firstCall = true;
    var deferred = new fb.util.promise.Deferred;
    fb.util.promise.attachDummyErrorHandler(deferred.promise);
    var onceCallback = (snapshot) {
      if (firstCall) {
        firstCall = false;
        this.off(eventType, onceCallback);
        if (userCallback!=null) {
          goog.bind(userCallback, context)(snapshot);
        }
        deferred.resolve(snapshot);
      }
    };
    this.on(eventType, onceCallback, cancelCallback: (err) {
      this.off(eventType, onceCallback);
      if (cancelCallback) {
        goog.bind(cancelCallback, context)(err);
      }
      deferred.reject(err);
    });
    return deferred.promise;
  }


  Query limitToFirst(int limit) {
    if (limit <= 0) {
      throw new Exception("Query.limitToFirst: First argument must be a positive integer.");
    }
    if (this._queryParams.hasLimit) {
      throw new Exception("Query.limitToFirst: Limit was already set (by another call to limitToFirst, or limitToLast).");
    }
    return new Query(this.repo, this.path, this._queryParams.limitToFirst(limit), this._orderByCalled);
  }

  Query limitToLast(int limit) {
    if (limit <= 0) {
      throw new Exception("Query.limitToLast: First argument must be a positive integer.");
    }
    if (this._queryParams.hasLimit) {
      throw new Exception("Query.limitToLast: Limit was already set (by another call to limitToFirst, or limitToLast).");
    }
    return new Query(this.repo, this.path, this._queryParams.limitToLast(limit), this._orderByCalled);
  }

  Query orderByChild(String path) {
    if (path == r"$key") {
      throw new Exception(r'Query.orderByChild: "$key" is invalid.  Use Query.orderByKey() instead.');
    } else {
      if (path == r"$priority") {
        throw new Exception(r'Query.orderByChild: "$priority" is invalid.  Use Query.orderByPriority() instead.');
      } else {
        if (path == r"$value") {
          throw new Exception(r'Query.orderByChild: "$value" is invalid.  Use Query.orderByValue() instead.');
        }
      }
    }
    validation.validatePathString("Query.orderByChild", 1, path, false);
    this._validateNoPreviousOrderByCall("Query.orderByChild");
    var parsedPath = new util.Path.fromString(path);
    if (parsedPath.isEmpty) {
      throw new Exception("Query.orderByChild: cannot pass in empty path.  Use Query.orderByValue() instead.");
    }
    var index = new PathIndex(parsedPath);
    var newParams = this._queryParams.orderBy(index);
    this._validateQueryEndpoints(newParams);
    return new Query(this.repo, this.path, newParams, true);
  }

  Query orderByKey() {
    this._validateNoPreviousOrderByCall("Query.orderByKey");
    var newParams = this._queryParams.orderBy(keyIndex);
    this._validateQueryEndpoints(newParams);
    return new Query(this.repo, this.path, newParams, true);
  }

  Query orderByPriority() {
    this._validateNoPreviousOrderByCall("Query.orderByPriority");
    var newParams = this._queryParams.orderBy(fb.core.snap.PriorityIndex);
    this._validateQueryEndpoints(newParams);
    return new Query(this.repo, this.path, newParams, true);
  }

  Query orderByValue() {
    this._validateNoPreviousOrderByCall("Query.orderByValue");
    var newParams = this._queryParams.orderBy(fb.core.snap.ValueIndex);
    this._validateQueryEndpoints(newParams);
    return new Query(this.repo, this.path, newParams, true);
  }

  Query startAt([value, name]) {
    fb.core.util.validation.validateFirebaseDataArg("Query.startAt", 1, value, this.path, true);
    fb.core.util.validation.validateKey("Query.startAt", 2, name, true);
    var newParams = this._queryParams.startAt(value, name);
    this._validateLimit(newParams);
    this._validateQueryEndpoints(newParams);
    if (this._queryParams.hasStart) {
      throw new Exception("Query.startAt: Starting point was already set (by another call to startAt " + "or equalTo).");
    }
    return new Query(this.repo, this.path, newParams, this._orderByCalled);
  }

  Query endAt([value, opt_name]) {
    fb.core.util.validation.validateFirebaseDataArg("Query.endAt", 1, value, this.path, true);
    fb.core.util.validation.validateKey("Query.endAt", 2, opt_name, true);
    var newParams = this._queryParams.endAt(value, opt_name);
    this._validateLimit(newParams);
    this._validateQueryEndpoints(newParams);
    if (this._queryParams.hasEnd) {
      throw new Exception("Query.endAt: Ending point was already set (by another call to endAt or " + "equalTo).");
    }
    return new Query(this.repo, this.path, newParams, this._orderByCalled);
  }

  Query equalTo(value, [name]) {
    fb.core.util.validation.validateFirebaseDataArg("Query.equalTo", 1, value, this.path, false);
    fb.core.util.validation.validateKey("Query.equalTo", 2, name, true);
    if (this._queryParams.hasStart) {
      throw new Exception("Query.equalTo: Starting point was already set (by another call to endAt or " + "equalTo).");
    }
    if (this._queryParams.hasEnd) {
      throw new Exception("Query.equalTo: Ending point was already set (by another call to endAt or " + "equalTo).");
    }
    return this.startAt(value, name).endAt(value, name);
  }

  queryObject() => this._queryParams.getQueryObject();


  queryIdentifier() {
    var obj = this.queryObject();
    var id = fb.core.util.ObjectToUniqueKey(obj);
    return id == "{}" ? "default" : id;
  }


  int get hashCode => toString().hashCode;

  bool operator==(other) {
    if (other is! Query) return false;
    if (path!=other.path) return false;
    return mapsEqual(queryObject(), other.queryObject());
  }

  toString() => "$path[$queryIdentifier]";
}

