
import 'package:firebase_dart/src/dart/api.dart';
import 'package:firebase_dart/src/dart/core/repo.dart';
import 'package:firebase_dart/src/dart/core/view.dart';
import 'dart:collection';
import 'dart:convert';
import 'dart:async';
import 'package:firebase_dart/src/dart/realtime.dart';
import 'package:firebase_dart/src/dart/core/util.dart' as util;
import 'package:logging/logging.dart';
import 'dart:math' as math;
import 'dart:io';

var RECONNECT_MIN_DELAY = 1E3;
var RECONNECT_MAX_DELAY_DEFAULT = 60 * 5 * 1E3;
var RECONNECT_MAX_DELAY_FOR_ADMINS = 30 * 1E3;
var RECONNECT_DELAY_MULTIPLIER = 1.3;
const RECONNECT_DELAY_RESET_TIMEOUT = 3E4;

typedef Callback(String status, data);


typedef int HashFunction();

class Request {
  final String action;
  final String tag;
  final Query query;
  final util.Path path;
  final hash;
  final data;
  final String cred;
  final HashFunction hashFn;
  final Completer<Response> _completer = new Completer();

  Request(this.action, {this.path, this.tag, this.query, this.hash, this.data, this.cred, this.hashFn});

  Request.auth(String cred) : this("auth", cred: cred);
  Request.unauth() : this("unauth");
  Request.listen(Query query, {HashFunction hashFn, String tag})
      : this("q", tag: tag, query: query, hashFn: hashFn);
  Request.unlisten(Query query, {String tag})
      : this("n", query: query, tag: tag);
  Request.onDisconnectPut(util.Path path, data)
      : this("o", path: path, data: data);
  Request.onDisconnectMerge(util.Path path, data)
      : this("om", path: path, data: data);
  Request.onDisconnectCancel(util.Path path, data)
      : this("oc", path: path, data: data);
  Request.put(util.Path path, data, HashFunction hashFn)
      : this("p", path: path, data: data, hashFn: hashFn);
  Request.merge(util.Path path, data, HashFunction hashFn)
      : this("m", path: path, data: data, hashFn: hashFn);
  Request.stats(stats)
      : this("s", data: stats);

  toJson(int reqNum) {
    var data;
    switch (action) {
      case "auth":
        data = {"cred": cred};
        break;
      case "unauth":
        data = {};
        break;
      case "q": // listen
      case "n": // unlisten
        data = {
          "p": query.path.toString(),
        };
        if (action=="q") {
          data["h"] = hashFn();
        }
        if (tag!=null) {
          data.addAll({
            "q": query.queryObject(),
            "t": tag
          });
        }
        break;
      case "p":
      case "m":
        data = {
          "p": path.toString(),
          "d": this.data
        };
        if (hashFn!=null) {
          data["h"] = hashFn();
        }
        break;
      case "o":  // on disconnect
      case "oc": // on disconnect cancel
      case "om": // on disconnect merge
        data = {
          "p": query.path.toString(),
          "d": this.data
        };
        break;
      case "s":
        data = {
          "c": this.data
        };
    }
    return {
      "a": action,
      "b": data,
      "r": reqNum
    };
  }
}

class Response {
  final payload;
  final String status;
  final Request request;
  final String error;
  final action;

  util.Path get path => new util.Path.fromString(payload["p"]);
  get data => payload["d"];
  String get tag => payload["t"];
  String get queryIdentifier => payload["q"];

  const Response({this.payload, this.status, this.request, this.error, this.action});

  static const disconnect = const Response();


  static const status_ok = "ok";


  bool get isOk => status==status_ok;

  bool get hasWarnings => payload is Map && payload.containsKey("w");


  bool get isResponseToRequest => request!=null;
  bool get isError => error!=null;
  bool get isPush => action!=null;
  bool get hasMessage => message!=null;
  String get message => payload["msg"];

  static const action_set = "d";
  static const action_merge= "m";
  static const action_listen_revoked = "c";
  static const action_auth_revoked = "ac";
  static const action_security_debug = "sd";


  factory Response.fromJson(Map<String,dynamic> json, Map<int, Request> requests) =>
      new Response(
          payload: json["d"],
          status: json["s"],
          request: requests.remove(json["r"]),
          error: json["error"],
          a: json["a"]
      );


  Iterable<String> get warnings {
    if (hasWarnings) {
      var warnings = payload["w"];
      return warnings
        .map((w) {
        switch (w) {
          case "no_index":
            var indexSpec = '".indexOn": "${request.query.queryParams.comparator}"';
            return "Using an unspecified index. Consider adding $indexSpec at "
                "${request.query.path} to your security rules for better performance";

        }
      });
    }
    return [];
  }

}

abstract class Transport {

  Stream get input;
  StreamSink get output;
}

main() async {
  var repoInfo = new RepoInfo("partago-dev.firebaseio.com", true, "partago-dev", true);

  Logger.root.level = Level.ALL;
  Logger.root.onRecord.listen(print);

  var conn = new PersistentConnection(repoInfo, null, (_)=>null, ({serverTimeOffset}) => null);


  var ref = new Query(new Repo(repoInfo, false), new util.Path.fromString("/test"), new QueryParams(), true);
//  await conn.listen(ref, ()=>1, "x");
  await conn.put(new util.Path.fromString("/test"),"hello world 2");
//  var transport = new WebSocketTransport("test",repoInfo);


//  await transport.done;
}

class WebSocketTransport extends Transport {
  static const UPGRADE_TIMEOUT = 6E4;
  static const DELAY_BEFORE_SENDING_EXTRA_REQUESTS = 5E3;
  static const BYTES_SENT_HEALTHY_OVERRIDE = 10 * 1024;
  static const BYTES_RECEIVED_HEALTHY_OVERRIDE = 100 * 1024;
  static const REALTIME_STATE_CONNECTING = 0;
  static const REALTIME_STATE_CONNECTED = 1;
  static const REALTIME_STATE_DISCONNECTED = 2;
  static const MESSAGE_TYPE = "t";
  static const MESSAGE_DATA = "d";
  static const CONTROL_SHUTDOWN = "s";
  static const CONTROL_RESET = "r";
  static const CONTROL_ERROR = "e";
  static const CONTROL_PONG = "o";
  static const SWITCH_ACK = "a";
  static const END_TRANSMISSION = "n";
  static const PING = "p";
  static const SERVER_HELLO = "h";

  WebSocket _socket;
  final RepoInfo _repoInfo;
  final String connId;
  String _sessionId;

  WebSocketTransport(this.connId, this._repoInfo) {
    connect();
  }

  connectionURL_(repoInfo, [opt_transportSessionId, opt_lastSessionId]) {
    var urlParams = {};
    urlParams[Constants.VERSION_PARAM] = Constants.PROTOCOL_VERSION;
/*    if (!NODE_CLIENT && typeof location !== "undefined" && location.href && location.href.indexOf(fb.realtime.Constants.FORGE_DOMAIN) !== -1) {
  urlParams[fb.realtime.Constants.REFERER_PARAM] = fb.realtime.Constants.FORGE_REF;
  }*/
/*  if (opt_transportSessionId) {
  urlParams[fb.realtime.Constants.TRANSPORT_SESSION_PARAM] = opt_transportSessionId;
  }
  if (opt_lastSessionId) {
  urlParams[fb.realtime.Constants.LAST_SESSION_PARAM] = opt_lastSessionId;
  }*/
  return repoInfo.connectionURL(Constants.WEBSOCKET, urlParams);
}
  Future connect() async {
    var url = connectionURL_(_repoInfo);
    print(url);
    var socket = _socket = await WebSocket.connect(url);
    socket.listen((v) {
      print("received $v");
      onPrimaryMessageReceived_(JSON.decode(v));
    });
  }

  final Completer completer = new Completer();

  Future get done => completer.future;

  close() {
    print("close");
    _socket.close();
  }

  onPrimaryMessageReceived_(parsedData) {
    var layer = parsedData["t"];
    var data = parsedData["d"];
    if (layer == "c") {
      this.onControl_(data);
    } else {
      if (layer == "d") {
        this.onDataMessage_(data);
      }
    }
  }
  onDataMessage_(message) {
    print("on data message $message");
    this.onPrimaryResponse_();
    this.onMessage_(message);
  }



  onReset_(host) {
    print("Reset packet received.  New host: " + host);
    this._repoInfo.updateHost(host);
    close();
    connect();
    return;

    if (this.state_ == REALTIME_STATE_CONNECTED) {
      this.close();
    } else {
      this.closeConnections_();
      this.start_();
    }
  }

  onControl_(controlData) {
    print("on control $controlData");
    var cmd = controlData[MESSAGE_TYPE];
    if (controlData.containsKey(MESSAGE_DATA)) {
      var payload = controlData[MESSAGE_DATA];
      if (cmd == SERVER_HELLO) {
        this.onHandshake_(payload);
      } else {
        if (cmd == END_TRANSMISSION) {
          this.log_("recvd end transmission on primary");
          this.rx_ = this.secondaryConn_;
          for (var i = 0;i < this.pendingDataMessages.length;++i) {
            this.onDataMessage_(this.pendingDataMessages[i]);
          }
          this.pendingDataMessages = [];
          this.tryCleanupConnection();
        } else {
          if (cmd == CONTROL_SHUTDOWN) {
            this.onConnectionShutdown_(payload);
          } else {
            if (cmd == CONTROL_RESET) {
              this.onReset_(payload);
            } else {
              if (cmd == CONTROL_ERROR) {
                fb.core.util.error("Server Error: " + payload);
              } else {
                if (cmd == CONTROL_PONG) {
                  this.log_("got pong on primary.");
                  this.onPrimaryResponse_();
                  this.sendPingOnPrimaryIfNecessary_();
                } else {
                  fb.core.util.error("Unknown control packet command: " + cmd);
                }
              }
            }
          }
        }
      }
    }
  }



  onHandshake_(handshake) {
    print("handshake $handshake");
    var timestamp = handshake["ts"];
    var version = handshake["v"];
    var host = handshake["h"];
    _sessionId = handshake["s"];
    _repoInfo.updateHost(host);
    _readyCompleter.complete(new DateTime.fromMillisecondsSinceEpoch(timestamp));
    return;
    if (this.state_ == REALTIME_STATE_CONNECTING) {
      this.conn_.start();
      this.onConnectionEstablished_(this.conn_, timestamp);
      if (fb.realtime.Constants.PROTOCOL_VERSION != version) {
        fb.core.util.warn("Protocol version mismatch detected");
      }
      this.tryStartUpgrade_();
    }
  }

  Future get ready => _readyCompleter.future;
  final Completer _readyCompleter = new Completer();

  Stream get input => _socket;
  StreamSink get output => _socket;
}


class PersistentConnection {

  final Map<Query,Request> _listens = {};
  Request _credential = null;
  final List<Request> _outstandingPuts = [];

  final Map<int, Request> _sentRequests = {};

  bool _connected = false;

  final int id;
  final Logger _log;


  bool _interrupted = false;
  bool _killed = false;
  final Queue _onDisconnectRequestQueue = new Queue();
  num _reconnectDelay = RECONNECT_MIN_DELAY;
  num _maxReconnectDelay = RECONNECT_MAX_DELAY_DEFAULT;
  final onDataUpdate_;
  final _onConnectStatus;
  final _onServerInfoUpdate;
  final _repoInfo;
  var _securityDebugCallback = null;
  var lastSessionId = null;
  Transport _transport = null;
  var _establishConnectionTimer = null;
  bool _visible = false;
  int requestNumber_ = 0;
  bool _firstConnection = true;
  DateTime _lastConnectionAttemptTime = null;
  DateTime _lastConnectionEstablishedTime = null;


  PersistentConnection(this._repoInfo, this.onDataUpdate_, this._onConnectStatus, this._onServerInfoUpdate) :
        _log = new Logger("p:$nextPersistentConnectionId_:"),
        id = nextPersistentConnectionId_++
     {
    this._scheduleConnect(0);
/*
    util.VisibilityMonitor.getInstance().on("visible", this._onVisible, this);
    if (_repoInfo.host.indexOf("fblocal") == -1) {
      util.OnlineMonitor.getInstance().on("online", this._onOnline, this);
    }
*/
  }
  
  static int nextPersistentConnectionId_ = 0;
  static int nextConnectionId_ =0;
  
  Future<Response> sendRequest(Request request) {
    var curReqNum = ++this.requestNumber_;
    _log.fine(request);
    assert(this._connected);//, "sendRequest call when we're not connected not allowed.");
    var msg = JSON.encode({"t":"d", "d": request.toJson(curReqNum)});
    print("send request $msg");
    this._transport.output.add(msg);
    _sentRequests[curReqNum] = request;
    return request._completer.future;
  }

  Future<Response> listen(Query query, HashFunction currentHashFn, String tag) {
    _log.fine("Listen called for $query");

    assert(query.queryParams.isDefault || !query.queryParams.loadsAllData);//, "listen() called for non-default but complete query");
    assert(this._listens[query]==null);//, "listen() called twice for same path/queryId.");
    var listenSpec = new Request.listen(query, hashFn: currentHashFn, tag: tag);
    this._listens[query] = listenSpec;
    if (this._connected) {
      this._sendListen(listenSpec);
    }
    return listenSpec._completer.future;
  }
  
  void _sendListen(Request listenSpec) {
    var query = listenSpec.query;
    _log.fine("Listen on $query");

    this.sendRequest(listenSpec)
        .then((response) {

      response.warnings.forEach(_log.warning);
      var currentListenSpec = _listens[query];
      if (currentListenSpec == listenSpec) {
        _log.fine("listen response", response);
        if (response.isOk) {
          _listens.remove(query);
        }
      }
    });
  }
  

  Future<Response> auth(String cred) {
    _credential = new Request.auth(cred);
    _log.fine("Authenticating using credential: $cred");
    tryAuth();
    _reduceReconnectDelayIfAdminCredential(cred);
    return _credential._completer.future;
  }

  void _reduceReconnectDelayIfAdminCredential(String credential) {
    var isFirebaseSecret = credential.length == 40;
    if (isFirebaseSecret || jwt.isAdmin(credential)) {
      _log.fine("Admin auth credential detected.  Reducing max reconnect time.");
      this._maxReconnectDelay = RECONNECT_MAX_DELAY_FOR_ADMINS;
    }
  }

  Future<Response> unauth() {
    this._credential = null;
    if (this._connected) {
      return this.sendRequest(new Request.unauth());
    }
    return null; // TODO replace with never completing future
  }

  void tryAuth() {
    var self = this;
    if (this._connected && _credential!=null) {
      var request = _credential;

      sendRequest(request)
          .then((res) {
        if (!res.isOk && self._credential == request) {
          self._credential = null;
        }
      });
    }
  }

  Future<Response> unlisten(Query query, String tag) {
    _log.fine("Unlisten called for $query");
    assert(query.queryParams.isDefault || !query.queryParams.loadsAllData);//, "unlisten() called for non-default but complete query");
    var listen = this._listens.remove(query);
    if (listen!=null && this._connected) {
      var request = new Request.unlisten(query, tag: tag);
      _log.fine("Unlisten on ${request.query}");
      return this.sendRequest(request);
    }
    return new Future.value();
  }

  Future<Response> onDisconnectPut(util.Path path, data) =>
      _sendOnDisconnect(new Request.onDisconnectPut(path, data));

  Future<Response> onDisconnectMerge(util.Path path, data) =>
      _sendOnDisconnect(new Request.onDisconnectMerge(path, data));

  Future<Response> onDisconnectCancel(util.Path path, data) =>
      _sendOnDisconnect(new Request.onDisconnectCancel(path, data));


  Future<Response> _sendOnDisconnect(Request request) {
    if (this._connected) {
      _log.fine("onDisconnect $request");
      this.sendRequest(request);
    } else {
      this._onDisconnectRequestQueue.add(request);
    }
    return request._completer.future;
  }

  Future<Response> put(util.Path path, data, [HashFunction hash]) =>
      _putInternal(new Request.put(path, data, hash));

  Future<Response> merge(util.Path path, data, [HashFunction hash]) =>
      _putInternal(new Request.merge(path, data, hash));

  Future<Response> _putInternal(Request request) {
    this._outstandingPuts.add(request);
    if (this._connected) {
      this._sendPut(request);
    } else {
      _log.fine("Buffering put: ${request.path}");
    }
    return request._completer.future;
  }

  final Set<Request> _queuedPuts = new Set();
  _sendPut(Request request) {
    _queuedPuts.add(request);
    this.sendRequest(request)
        .then((response) {
      _log.fine("${request.action} response $response");
      _outstandingPuts.remove(request);
    });
  }

  Future<Response> reportStats(stats) {
    if (_connected) {
      var request = new Request.stats(stats);
      _log.fine("reportStats $request");
      return this.sendRequest(request)
          .then((response) {
        if (!response.isOk) {
          _log.fine("reportStats: Error sending stats: ${response.payload}");
        }
      });
    }
    return null; //TODO
  }

  void _onDataMessage(message) {
    var response = new Response.fromJson(message, _sentRequests);

    if (response.isResponseToRequest) {
      _log.fine("from server: " + JSON.encode(response));

    } else {
      if (response.isError) {
        throw new Exception("A server-side error has occurred: ${response.error}");
      } else {
        if (response.isPush) {
          this._onDataPush(response);
        }
      }
    }
  }

  void _onDataPush(Response response) {
    _log.fine("handleServerMessage $response");

    switch (response.action) {
      case Response.action_set:
        this.onDataUpdate_(response.path, response.data, false, response.tag);
        break;
      case Response.action_merge:
        this.onDataUpdate_(response.path, response.data, true, response.tag);
        break;
      case Response.action_listen_revoked:
        this._onListenRevoked(response.path, response.queryIdentifier);
        break;
      case Response.action_auth_revoked:
        this._onAuthRevoked(body["s"], response.data);
        break;
      case Response.action_security_debug:
        this._onSecurityDebugPacket(response.payload);
        break;
      default:
        throw new Exception("Unrecognized action received from server: ${response.action}\nAre you using the latest client?");
    }
  }

  void _onReady(DateTime timestamp) {
    _log.fine("connection ready");
    _connected = true;
    _lastConnectionEstablishedTime = new DateTime.now();

    this.handleTimestamp_(timestamp);

    this.lastSessionId = _transport._sessionId;
    if (this._firstConnection) {
      this._sendConnectStats();
    }
    this._restoreState();
    this._firstConnection = false;
    this._onConnectStatus(true);
  }

  void _scheduleConnect(num timeout) {
    assert(this._transport==null);//, "Scheduling a connect when we're already connected/ing?");

    var future = this._establishConnectionTimer = new Future.delayed(
        new Duration(milliseconds: timeout.floor())
    );

    future.then((_) {
      if (future!=_establishConnectionTimer) {
        return;
      }
      _establishConnectionTimer = null;
      _establishConnection();
    });
  }

  void _onVisible(bool visible) {
    if (visible && !this._visible && this._reconnectDelay == this._maxReconnectDelay) {
      _log.fine("Window became visible.  Reducing delay.");
      this._reconnectDelay = RECONNECT_MIN_DELAY;
      if (this._transport==null) {
        this._scheduleConnect(0);
      }
    }
    this._visible = visible;
  }

  void _onOnline(bool online) {
    if (online) {
      _log.fine("Browser went online.");
      _reconnectDelay = RECONNECT_MIN_DELAY;
      if (_transport==null) {
        _scheduleConnect(0);
      }
    } else {
      _log.fine("Browser went offline.  Killing connection.");
      _transport?.close();
    }
  }

  void _onRealtimeDisconnect() {
    _log.fine("data client disconnected");
    _connected = false;
    _transport = null;
    _cancelSentTransactions();
    if (_shouldReconnect()) {
      if (!_visible) {
        _log.fine("Window isn't visible.  Delaying reconnect.");
        this._reconnectDelay = this._maxReconnectDelay;
        this._lastConnectionAttemptTime = new DateTime.now();
      } else {
        if (this._lastConnectionEstablishedTime!=null) {
          var timeSinceLastConnectSucceeded = new DateTime.now().difference(this._lastConnectionEstablishedTime);
          if (timeSinceLastConnectSucceeded.inMilliseconds > RECONNECT_DELAY_RESET_TIMEOUT) {
            this._reconnectDelay = RECONNECT_MIN_DELAY;
          }
          this._lastConnectionEstablishedTime = null;
        }
      }
      var timeSinceLastConnectAttempt = new DateTime.now().difference(this._lastConnectionAttemptTime);
      var reconnectDelay = math.max(0, this._reconnectDelay - timeSinceLastConnectAttempt);
      reconnectDelay = new math.Random().nextDouble() * reconnectDelay;
      _log.fine("Trying to reconnect in " + reconnectDelay + "ms");
      this._scheduleConnect(reconnectDelay);
      this._reconnectDelay = math.min(this._maxReconnectDelay, this._reconnectDelay * RECONNECT_DELAY_MULTIPLIER);
    }
    this._onConnectStatus(false);
  }

  Future _establishConnection() async {
    if (this._shouldReconnect()) {
      _log.fine("Making a connection attempt");
      this._lastConnectionAttemptTime = new DateTime.now();
      this._lastConnectionEstablishedTime = null;
      var connId = "$id:${nextConnectionId_++}";
      this._transport = new WebSocketTransport(connId, _repoInfo);

      _transport.ready.then(_onReady);
/*      _transport.input.listen(
          _onDataMessage
      );*/

/*
      new Connection(connId, this._repoInfo, _onDataMessage,
          _onReady, _onRealtimeDisconnect, _onKill, lastSessionId);
*/
    }
  }

  void _onKill(String reason) {
    _log.warning("$reason ($_repoInfo)");
    _killed = true;
  }

  void interrupt() {
    this._interrupted = true;
    if (_realtime!=null) {
      this._realtime.close();
    } else {
      _establishConnectionTimer = null;
      if (_connected) {
        _onRealtimeDisconnect();
      }
    }
  }


  void resume() {
    this._interrupted = false;
    this._reconnectDelay = RECONNECT_MIN_DELAY;
    if (this._realtime==null) {
      this._scheduleConnect(0);
    }
  }

  void handleTimestamp_(DateTime timestamp) {
    var delta = timestamp.difference(new DateTime.now());
    _onServerInfoUpdate(serverTimeOffset: delta);
  }

  void _cancelSentTransactions() {
    _sentRequests.values.forEach((r)=>r._completer.complete(Response.disconnect));
    _sentRequests.clear();
  }

  void _onListenRevoked(Response response) {
    var listen = _listens.remove(response.query);
    listen._completer.complete(response);
  }

  void _onAuthRevoked(Response response) {
    var cred = this._credential;
    _credential = null;
    if (cred!=null) {
      cred._completer.complete(response);
    }
  }

  void _onSecurityDebugPacket(Response response) {
    if (this._securityDebugCallback) {
      this._securityDebugCallback(response.payload);
    } else {
      if (response.hasMessage) {
        _log.info(response.message);
      }
    }
  }

  void _restoreState() {
    this.tryAuth();
    _listens.values.forEach(_sendListen);
    _outstandingPuts.forEach(_sendPut);
    _onDisconnectRequestQueue.forEach(_sendOnDisconnect);
  }

  static const CLIENT_VERSION = "2.4.0";

  void _sendConnectStats() {
    var stats = {};

    stats["sdk.js." + CLIENT_VERSION.replaceAll(".", "-")] = 1;
    this.reportStats(stats);
  }

  bool _shouldReconnect() {
    var online = true;//OnlineMonitor.getInstance().currentlyOnline();
    return!this._killed && !this._interrupted && online;
  }

}


