
import 'storage.dart' as storage;
import 'package:firebase_dart/src/dart/realtime.dart' as realtime;
import 'snap.dart' as snap;
import 'util.dart' as util;
import 'dart:async';
import 'dart:math' as math;
import 'package:firebase_dart/src/dart/firebase.dart';
import 'sync.dart';

class RepoInfo {
  final String host;
  final String domain;
  final bool secure;
  final String namespace;
  final bool webSocketOnly;
  final String persistenceKey;
  String _internalHost;

  String get internalHost => _internalHost;

  RepoInfo(String host, this.secure, this.namespace, this.webSocketOnly, [this.persistenceKey = ""]) :
      host = host.toLowerCase(), domain = host.substring(host.indexOf(".")+1).toLowerCase(),
        _internalHost = storage.persistentStorage["host:" + host.toLowerCase()] ?? host.toLowerCase();

  bool get needsQueryParam => host!=internalHost;

  bool get isCacheableHost => internalHost.substring(0,2)=="s-";

  bool get isDemoHost => domain=="firebaseio-demo.com";

  bool get isCustomHost => domain != "firebaseio.com" && this.domain != "firebaseio-demo.com";

  void updateHost(String newHost) {
    if (newHost != this.internalHost) {
      this._internalHost = newHost;
      if (this.isCacheableHost) {
        storage.persistentStorage["host:" + this.host] = this.internalHost;
      }
    }
  }

  connectionURL(String type, Map<String,dynamic> params) {
    var connURL;
    if (type == realtime.Constants.WEBSOCKET) {
      connURL = (this.secure ? "wss://" : "ws://") + this.internalHost + "/.ws?";
    } else {
      if (type == realtime.Constants.LONG_POLLING) {
        connURL = (this.secure ? "https://" : "http://") + this.internalHost + "/.lp?";
      } else {
        throw new Exception("Unknown connection type: " + type);
      }
    }
    if (this.needsQueryParam) {
      params["ns"] = this.namespace;
    }
    var pairs = [];
    params.forEach((index, element) {
      pairs.add("$index=$element");
    });
    return connURL + pairs.join("&");
  }
  
  
  toString() {
    var str = (this.secure ? "https://" : "http://") + this.host;
    if (persistenceKey!=null) {
      str += "<" + this.persistenceKey + ">";
    }
    return str;
  }
  
  int get hashCode => toString().hashCode;
  
  bool operator==(other) => other is RepoInfo&&toString()==other.toString();

}



class Repo {

  final RepoInfo repoInfo;
  final bool forceRestClient;
  var _stats;
  stats.StatsListener _statsListener;
  stats.StatsReporter _statsReporter;
  var _eventQueue = [];//new EventQueue();
  var _nextWriteId = 1;
  var _persistentConnection = null;
  var _server = null;
  var _onDisconnect;
  var auth;
  int dataUpdateCount;
  var _interceptServerDataCallback;
  SyncTree _serverSyncTree;
  var _infoData;
  var _infoSyncTree;

  Repo(this.repoInfo, this.forceRestClient) {
    _stats = StatsManager.getCollection(repoInfo);
    if (forceRestClient || util.beingCrawled()) {
      this._server = new core.ReadonlyRestClient(repoInfo, _onDataUpdate);
      new Future.microtask(()=>_onConnectStatus(true));
    } else {
      this._persistentConnection = new core.PersistentConnection(repoInfo,
          _onDataUpdate, _onConnectStatus, _onServerInfoUpdate);
      this._server = this._persistentConnection;
    }
    this._statsReporter = stats.StatsManager.getOrCreateReporter(repoInfo, () {
      return new stats.StatsReporter(this._stats, this._server);
    });
    this._transactions_init();
    this._infoData = new core.SnapshotHolder();
    this._infoSyncTree = new core.SyncTree(
        startListening:(query, tag, currentHashFn, onComplete) {
          var infoEvents = [];
          var node = _infoData.getNode(query.path);
          if (!node.isEmpty()) {
            infoEvents = _infoSyncTree.applyServerOverwrite(query.path, node);
            new Future.microtask(()=>onComplete("ok"));
          }
          return infoEvents;
        }
    );
    this._updateInfo("connected", false);
    this._onDisconnect = new core.SparseSnapshotTree();
    this.auth = new login.AuthenticationManager(repoInfo, this._server.auth,
        this._server.unauth, this._onAuthStatus);
    this.dataUpdateCount = 0;
    this._interceptServerDataCallback = null;
    this._serverSyncTree = new SyncTree(
        startListening: (query, tag, currentHashFn, onComplete) {
          _server.listen(query, currentHashFn, tag, (status, data) {
            var events = onComplete(status, data);
            _eventQueue.raiseEventsForChangedPath(query.path, events);
          });
          return[];
        },
        stopListening: (query, tag) {
          _server.unlisten(query, tag);
        }
    );
  }

  toString() {
    return(repoInfo.secure ? "https://" : "http://") + repoInfo.host;
  }

  String get name => repoInfo.namespace;

  int get serverTime {
    var offsetNode = _infoData.getNode(new util.Path.fromString(".info/serverTimeOffset"));
    var offset = (offsetNode.val()) ?? 0;
    return new DateTime.now().millisecondsSinceEpoch + offset;
  }

  generateServerValues() {
    return util.ServerValues.generateWithValues({"timestamp":this.serverTime});
  }

  _onDataUpdate(pathString, data, isMerge, tag) {
    this.dataUpdateCount++;
    var path = new util.Path.fromString(pathString);
    data = this._interceptServerDataCallback ? this._interceptServerDataCallback(pathString, data) : data;
    var events = [];
    if (tag) {
      if (isMerge) {
        var taggedChildren = data.map((raw) {
          return new snap.Node.fromJson(raw);
        });
        events = this._serverSyncTree.applyTaggedQueryMerge(path, taggedChildren, tag);
      } else {
        var taggedSnap = new snap.Node.fromJson(data);
        events = this._serverSyncTree.applyTaggedQueryOverwrite(path, taggedSnap, tag);
      }
    } else {
      if (isMerge) {
        var changedChildren = data.map((raw) {
          return new snap.Node.fromJson(raw);
        });
        events = this._serverSyncTree.applyServerMerge(path, changedChildren);
      } else {
        var _snap = new snap.Node.fromJson(data);
        events = this._serverSyncTree.applyServerOverwrite(path, _snap);
      }
    }
    var affectedPath = path;
    if (events.length > 0) {
      affectedPath = this._rerunTransactions(path);
    }
    this._eventQueue.raiseEventsForChangedPath(affectedPath, events);
  }

  _interceptServerData(callback) {
    this._interceptServerDataCallback = callback;
  }

  _onConnectStatus(connectStatus) {
    this._updateInfo("connected", connectStatus);
    if (connectStatus == false) {
      this._runOnDisconnectEvents();
    }
  }

  _onServerInfoUpdate(Map updates) {
    updates.forEach((key, value) {
      _updateInfo(key, value);
    });
  }
  _onAuthStatus(authStatus) {
    this._updateInfo("authenticated", authStatus);
  }
  _updateInfo(String pathString, value) {
    var path = new util.Path.fromString("/.info/" + pathString);
    var newNode = new snap.Node.fromJson(value);
    this._infoData.updateSnapshot(path, newNode);
    var events = this._infoSyncTree.applyServerOverwrite(path, newNode);
    this._eventQueue.raiseEventsForChangedPath(path, events);
  }

  _getNextWriteId() => _nextWriteId++;


  setWithPriority(util.Path path, newVal, newPriority, onComplete) {
    this._log("set", path:path.toString(), value:newVal, priority:newPriority);

    var serverValues = this.generateServerValues();
    var newNodeUnresolved = new snap.Node.fromJson(newVal, newPriority);

    var newNode = util.ServerValues.resolveDeferredValueSnapshot(newNodeUnresolved, serverValues);
    var writeId = this._getNextWriteId();
    var events = this._serverSyncTree.applyUserOverwrite(path, newNode, writeId, true);
    this._eventQueue.queueEvents(events);
    var self = this;
    this._server.put(path.toString(), newNodeUnresolved.val(true), (status, errorReason) {
      var success = status == "ok";
      if (!success) {
        util.warn("set at " + path + " failed: " + status);
      }
      var clearEvents = self._serverSyncTree.ackUserWrite(writeId, !success);
      self._eventQueue.raiseEventsForChangedPath(path, clearEvents);
      self.callOnCompleteCallback(onComplete, status, errorReason);
    });
    var affectedPath = this._abortTransactions(path);
    this._rerunTransactions(affectedPath);
    this._eventQueue.raiseEventsForChangedPath(affectedPath, []);
  }

  update(util.Path path, Map<String,dynamic> childrenToMerge, onComplete) {
    this._log("update", path:path.toString(), value:childrenToMerge);

    var empty = childrenToMerge.isEmpty;
    var serverValues = this.generateServerValues();
    var changedChildren = <String,Node>{};

    childrenToMerge.forEach((changedKey, changedValue) {
      empty = false;
      var newNodeUnresolved = new snap.Node.fromJson(changedValue);
      changedChildren[changedKey] = util.ServerValues.resolveDeferredValueSnapshot(newNodeUnresolved, serverValues);
    });


    if (!empty) {
      var writeId = this._getNextWriteId();
      var events = this._serverSyncTree.applyUserMerge(path, changedChildren, writeId);
      this._eventQueue.queueEvents(events);
      var self = this;
      this._server.merge(path.toString(), childrenToMerge, (status, errorReason) {
        var success = status == "ok";
        if (!success) {
          util.warn("update at " + path + " failed: " + status);
        }
        var clearEvents = self._serverSyncTree.ackUserWrite(writeId, !success);
        var affectedPath = path;
        if (clearEvents.length > 0) {
          affectedPath = self._rerunTransactions(path);
        }
        self._eventQueue.raiseEventsForChangedPath(affectedPath, clearEvents);
        self.callOnCompleteCallback(onComplete, status, errorReason);
      });
      var affectedPath = this._abortTransactions(path);
      this._rerunTransactions(affectedPath);
      this._eventQueue.raiseEventsForChangedPath(path, []);
    } else {
      util.log("update() called with empty data.  Don't do anything.");
      this.callOnCompleteCallback(onComplete, "ok");
    }
  }

  _runOnDisconnectEvents() {
    this._log("onDisconnectEvents");
    var self = this;
    var serverValues = this.generateServerValues();
    var resolvedOnDisconnectTree = util.ServerValues.resolveDeferredValueTree(this._onDisconnect, serverValues);
    var events = [];
    resolvedOnDisconnectTree.forEachTree(util.Path.empty, (path, snap) {
      events..addAll(self._serverSyncTree.applyServerOverwrite(path, snap));
      var affectedPath = self._abortTransactions(path);
      self._rerunTransactions(affectedPath);
    });
    this._onDisconnect = new core.SparseSnapshotTree();
    this._eventQueue.raiseEventsForChangedPath(util.Path.empty, events);
  }

  onDisconnectCancel(path, onComplete) {
    var self = this;
    this._server.onDisconnectCancel(path.toString(), (status, errorReason) {
      if (status == "ok") {
        self._onDisconnect.forget(path);
      }
      self.callOnCompleteCallback(onComplete, status, errorReason);
    });
  }

  onDisconnectSet(path, value, onComplete) {
    var self = this;
    var newNode = new snap.Node.fromJson(value);
    this._server.onDisconnectPut(path.toString(), newNode.val(true), (status, errorReason) {
      if (status == "ok") {
        self._onDisconnect.remember(path, newNode);
      }
      self.callOnCompleteCallback(onComplete, status, errorReason);
    });
  }

  onDisconnectSetWithPriority(path, value, priority, onComplete) {
    var self = this;
    var newNode = new snap.Node.fromJson(value, priority);
    this._server.onDisconnectPut(path.toString(), newNode.val(true), (status, errorReason) {
      if (status == "ok") {
        self._onDisconnect.remember(path, newNode);
      }
      self.callOnCompleteCallback(onComplete, status, errorReason);
    });
  }

  onDisconnectUpdate(path, childrenToMerge, onComplete) {
    var empty = true;
    for (var childName in childrenToMerge) {
      empty = false;
    }
    if (empty) {
      util.log("onDisconnect().update() called with empty data.  Don't do anything.");
      this.callOnCompleteCallback(onComplete, "ok");
      return;
    }
    var self = this;
    this._server.onDisconnectMerge(path.toString(), childrenToMerge, (status, errorReason) {
      if (status == "ok") {
        for (var childName in childrenToMerge) {
          var newChildNode = new snap.Node.fromJson(childrenToMerge[childName]);
          self._onDisconnect.remember(path.child(childName), newChildNode);
        }
      }
      self.callOnCompleteCallback(onComplete, status, errorReason);
    });
  }

  addEventCallbackForQuery(query, eventRegistration) {
    var events;
    if (query.path.getFront() == ".info") {
      events = this._infoSyncTree.addEventRegistration(query, eventRegistration);
    } else {
      events = this._serverSyncTree.addEventRegistration(query, eventRegistration);
    }
    this._eventQueue.raiseEventsAtPath(query.path, events);
  }

  removeEventCallbackForQuery(query, eventRegistration) {
    var events;
    if (query.path.getFront() == ".info") {
      events = this._infoSyncTree.removeEventRegistration(query, eventRegistration);
    } else {
      events = this._serverSyncTree.removeEventRegistration(query, eventRegistration);
    }
    this._eventQueue.raiseEventsAtPath(query.path, events);
  }

  interrupt() => _persistentConnection?.interrupt();

  resume() => _persistentConnection?.resume();

  stats(showDelta) {
    if (console==null) {
      return;
    }
    var stats;
    if (showDelta) {
      _statsListener ?= new stats.StatsListener(this._stats);
      stats = this._statsListener.get();
    } else {
      stats = this._stats.get();
    }
    var longestName = stats.keys.reduce((int previousValue, String currentValue) {
      return math.max(currentValue.length, previousValue);
    }, 0);

    for (var stat in stats) {
      var value = stats[stat];
      for (var i = stat.length;i < longestName + 2;i++) {
        stat += " ";
      }
      console.log(stat + value);
    }
  }

  statsIncrementCounter(metric) {
    this._stats.incrementCounter(metric);
    this.statsReporter_.includeStat(metric);
  }

  _log(var_args) {
    var prefix = "";
    if (this._persistentConnection) {
      prefix = this._persistentConnection.id + ":";
    }
    util.log(prefix, arguments);
  }

  callOnCompleteCallback(callback, status, errorReason) {
    if (callback) {
      util.exceptionGuard(() {
        if (status == "ok") {
          callback(null);
        } else {
          var code = (status || "error").toUpperCase();
          var message = code;
          if (errorReason) {
            message += ": " + errorReason;
          }
          var error = new Exception(message);
          error.code = code;
          callback(error);
        }
      });
    }
  }


  // transaction

  _transactions_init() {
    this._transactionQueueTree = new util.Tree();
  }

  startTransaction(util.Path path, transactionUpdate, onComplete, applyLocally) {
    this._log("transaction on " + path);
    var valueCallback = () {
    };
    var watchRef = new Firebase(this, path);
    watchRef.on("value", valueCallback);
    var unwatcher = () {
      watchRef.off("value", valueCallback);
    };
    var transaction = new Transaction(path:path, update:transactionUpdate,
        onComplete:onComplete, status:null, order:util.LUIDGenerator(),
        applyLocally:applyLocally, retryCount:0, unwatcher:unwatcher,
        abortReason:null, currentWriteId:null, currentInputSnapshot:null,
        currentOutputSnapshotRaw:null, currentOutputSnapshotResolved:null);

    var currentState = this._getLatestState(path);
    transaction.currentInputSnapshot = currentState;
    var newVal = transaction.update(currentState.val());
    if (!goog.isDef(newVal)) {
      transaction.unwatcher();
      transaction.currentOutputSnapshotRaw = null;
      transaction.currentOutputSnapshotResolved = null;
      if (transaction.onComplete) {
        var snapshot = new api.DataSnapshot((transaction.currentInputSnapshot), new Firebase(this, transaction.path), fb.core.snap.PriorityIndex);
        transaction.onComplete(null, false, snapshot);
      }
    } else {
      util.validation.validateFirebaseData("transaction failed: Data returned ", newVal, transaction.path);
      transaction.status = TransactionStatus.RUN;
      var queueNode = this.transactionQueueTree_.subTree(path);
      var nodeQueue = queueNode.getValue() || [];
      nodeQueue.add(transaction);
      queueNode.setValue(nodeQueue);
      var priorityForNode;
      if (newVal is Map && newVal != null && fb.util.obj.contains(newVal, ".priority")) {
        priorityForNode = fb.util.obj.get(newVal, ".priority");
        assert(util.validation.isValidPriority(priorityForNode))
        //, "Invalid priority returned by transaction. " + "Priority must be a valid string, finite number, server value, or null.");
      } else {
        var currentNode = this._serverSyncTree.calcCompleteEventCache(path) || fb.core.snap.EMPTY_NODE;
        priorityForNode = currentNode.getPriority().val();
      }
      priorityForNode = (priorityForNode);
      var serverValues = this.generateServerValues();
      var newNodeUnresolved = new snap.Node.fromJson(newVal, priorityForNode);
      var newNode = util.ServerValues.resolveDeferredValueSnapshot(newNodeUnresolved, serverValues);
      transaction.currentOutputSnapshotRaw = newNodeUnresolved;
      transaction.currentOutputSnapshotResolved = newNode;
      transaction.currentWriteId = this._getNextWriteId();
      var events = this._serverSyncTree.applyUserOverwrite(path, newNode, transaction.currentWriteId, transaction.applyLocally);
      this._eventQueue.raiseEventsForChangedPath(path, events);
      this._sendReadyTransactions();
    }
  }

  _getLatestState(path, excludeSets) {
    return this._serverSyncTree.calcCompleteEventCache(path, excludeSets) || fb.core.snap.EMPTY_NODE;
  }

  _sendReadyTransactions(opt_node) {
    var node = (opt_node || this.transactionQueueTree_);
    if (!opt_node) {
      this._pruneCompletedTransactionsBelowNode(node);
    }
    if (node.getValue() != null) {
      var queue = this._buildTransactionQueue(node);
      assert(queue.length > 0);//, "Sending zero length transaction queue");
      var allRun = goog.array.every(queue, (transaction) {
        return transaction.status == TransactionStatus.RUN;
      });
      if (allRun) {
        this.sendTransactionQueue_(node.path(), queue);
      }
    } else {
      if (node.hasChildren()) {
        var self = this;
        node.forEachChild((childNode) {
          self._sendReadyTransactions(childNode);
        });
      }
    }
  }

  sendTransactionQueue_(path, queue) {
    var setsToIgnore = goog.array.map(queue, (txn) {
      return txn.currentWriteId;
    });
    var latestState = this._getLatestState(path, setsToIgnore);
    var snapToSend = latestState;
    var latestHash = latestState.hash();
    for (var i = 0;i < queue.length;i++) {
      var txn = queue[i];
      assert(txn.status == TransactionStatus.RUN);
      //, "tryToSendTransactionQueue_: items in queue should all be run.");
      txn.status = TransactionStatus.SENT;
      txn.retryCount++;
      var relativePath = util.Path.relativePath(path, txn.path);
      snapToSend = snapToSend.updateChild(relativePath, (txn.currentOutputSnapshotRaw));
    }
    var dataToSend = snapToSend.val(true);
    var pathToSend = path;
    var self = this;
    this._server.put(pathToSend.toString(), dataToSend, (status) {
      self._log("transaction put response", {path:pathToSend.toString(), status:status});
      var events = [];
      if (status == "ok") {
        var callbacks = [];
        for (var i = 0;i < queue.length;i++) {
          queue[i].status = TransactionStatus.COMPLETED;
          events.addAll(self._serverSyncTree.ackUserWrite(queue[i].currentWriteId));
          if (queue[i].onComplete) {
            var node = (queue[i].currentOutputSnapshotResolved);
            var ref = new Firebase(self, queue[i].path);
            var snapshot = new DataSnapshot(node, ref, snap.priorityIndex);
            callbacks.add(goog.bind(queue[i].onComplete, null, null, true, snapshot));
          }
          queue[i].unwatcher();
        }
        self._pruneCompletedTransactionsBelowNode(self.transactionQueueTree_.subTree(path));
        self._sendReadyTransactions();
        self._eventQueue.raiseEventsForChangedPath(path, events);
        for (var i = 0;i < callbacks.length;i++) {
          util.exceptionGuard(callbacks[i]);
        }
      } else {
        if (status == "datastale") {
          for (var i = 0;i < queue.length;i++) {
            if (queue[i].status == TransactionStatus.SENT_NEEDS_ABORT) {
              queue[i].status = TransactionStatus.NEEDS_ABORT;
            } else {
              queue[i].status = TransactionStatus.RUN;
            }
          }
        } else {
          util.warn("transaction at " + pathToSend.toString() + " failed: " + status);
          for (var i = 0;i < queue.length;i++) {
            queue[i].status = TransactionStatus.NEEDS_ABORT;
            queue[i].abortReason = status;
          }
        }
        self._rerunTransactions(path);
      }
    }, latestHash);
  }

  _rerunTransactions(changedPath) {
    var rootMostTransactionNode = this._getAncestorTransactionNode(changedPath);
    var path = rootMostTransactionNode.path();
    var queue = this._buildTransactionQueue(rootMostTransactionNode);
    this._rerunTransactionQueue(queue, path);
    return path;
  }

  _rerunTransactionQueue(queue, path) {
    if (queue.length == 0) {
      return;
    }
    var callbacks = [];
    var events = [];
    var setsToIgnore = goog.array.map(queue, (q) {
      return q.currentWriteId;
    });
    for (var i = 0;i < queue.length;i++) {
      var transaction = queue[i];
      var relativePath = util.Path.relativePath(path, transaction.path);
      var abortTransaction = false, abortReason;
      assert(relativePath != null);//, "rerunTransactionsUnderNode_: relativePath should not be null.");
      if (transaction.status == TransactionStatus.NEEDS_ABORT) {
        abortTransaction = true;
        abortReason = transaction.abortReason;
        events = events.concat(this._serverSyncTree.ackUserWrite(transaction.currentWriteId, true));
      } else {
        if (transaction.status == TransactionStatus.RUN) {
          if (transaction.retryCount >= fb.core.Repo._MAX_TRANSACTION_RETRIES) {
            abortTransaction = true;
            abortReason = "maxretry";
            events = events.concat(this._serverSyncTree.ackUserWrite(transaction.currentWriteId, true));
          } else {
            var currentNode = this._getLatestState(transaction.path, setsToIgnore);
            transaction.currentInputSnapshot = currentNode;
            var newData = queue[i].update(currentNode.val());
            if (goog.isDef(newData)) {
              util.validation.validateFirebaseData("transaction failed: Data returned ", newData, transaction.path);
              var newDataNode = new snap.Node.fromJson(newData);
              var hasExplicitPriority = newData is Map && newData != null && fb.util.obj.contains(newData, ".priority");
              if (!hasExplicitPriority) {
                newDataNode = newDataNode.updatePriority(currentNode.getPriority());
              }
              var oldWriteId = transaction.currentWriteId;
              var serverValues = this.generateServerValues();
              var newNodeResolved = util.ServerValues.resolveDeferredValueSnapshot(newDataNode, serverValues);
              transaction.currentOutputSnapshotRaw = newDataNode;
              transaction.currentOutputSnapshotResolved = newNodeResolved;
              transaction.currentWriteId = this._getNextWriteId();
              goog.array.remove(setsToIgnore, oldWriteId);
              events = events.concat(this._serverSyncTree.applyUserOverwrite(transaction.path, newNodeResolved, transaction.currentWriteId, transaction.applyLocally));
              events = events.concat(this._serverSyncTree.ackUserWrite(oldWriteId, true));
            } else {
              abortTransaction = true;
              abortReason = "nodata";
              events = events.concat(this._serverSyncTree.ackUserWrite(transaction.currentWriteId, true));
            }
          }
        }
      }
      this._eventQueue.raiseEventsForChangedPath(path, events);
      events = [];
      if (abortTransaction) {
        queue[i].status = TransactionStatus.COMPLETED;
        ((unwatcher) {
          setTimeout(unwatcher, Math.floor(0));
        })(queue[i].unwatcher);
        if (queue[i].onComplete) {
          if (abortReason == "nodata") {
            var ref = new Firebase(this, queue[i].path);
            var lastInput = (queue[i].currentInputSnapshot);
            var snapshot = new fb.api.DataSnapshot(lastInput, ref, snap.priorityIndex);
            callbacks.add(goog.bind(queue[i].onComplete, null, null, false, snapshot));
          } else {
            callbacks.add(goog.bind(queue[i].onComplete, null, new Error(abortReason), false, null));
          }
        }
      }
    }
    this._pruneCompletedTransactionsBelowNode(this.transactionQueueTree_);
    for (var i = 0;i < callbacks.length;i++) {
      util.exceptionGuard(callbacks[i]);
    }
    this._sendReadyTransactions();
  }


  _getAncestorTransactionNode(path) {
    var front;
    var transactionNode = this.transactionQueueTree_;
    while ((front = path.getFront()) != null && transactionNode.getValue() == null) {
      transactionNode = transactionNode.subTree(front);
      path = path.popFront();
    }
    return transactionNode;
  }

  _buildTransactionQueue(transactionNode) {
    var transactionQueue = [];
    this.aggregateTransactionQueuesForNode_(transactionNode, transactionQueue);
    transactionQueue.sort((a, b) {
      return a.order - b.order;
    });
    return transactionQueue;
  }

  aggregateTransactionQueuesForNode_(node, queue) {
    var nodeQueue = node.getValue();
    if (nodeQueue != null) {
      for (var i = 0;i < nodeQueue.length;i++) {
        queue.add(nodeQueue[i]);
      }
    }
    var self = this;
    node.forEachChild((child) {
      self.aggregateTransactionQueuesForNode_(child, queue);
    });
  }

  _pruneCompletedTransactionsBelowNode(node) {
    var queue = node.getValue();
    if (queue) {
      var to = 0;
      for (var from = 0;from < queue.length;from++) {
        if (queue[from].status != TransactionStatus.COMPLETED) {
          queue[to] = queue[from];
          to++;
        }
      }
      queue.length = to;
      node.setValue(queue.length > 0 ? queue : null);
    }
    var self = this;
    node.forEachChild((childNode) {
      self._pruneCompletedTransactionsBelowNode(childNode);
    });
  }

  _abortTransactions(path) {
    var affectedPath = this._getAncestorTransactionNode(path).path();
    var transactionNode = this.transactionQueueTree_.subTree(path);
    var self = this;
    transactionNode.forEachAncestor((node) {
      self._abortTransactionsOnNode(node);
    });
    this._abortTransactionsOnNode(transactionNode);
    transactionNode.forEachDescendant((node) {
      self._abortTransactionsOnNode(node);
    });
    return affectedPath;
  }

  _abortTransactionsOnNode(node) {
    var queue = node.value;
    if (queue != null) {
      var callbacks = [];
      var events = [];
      var lastSent = -1;
      for (var i = 0;i < queue.length;i++) {
        if (queue[i].status == TransactionStatus.SENT_NEEDS_ABORT) {
        } else {
          if (queue[i].status == TransactionStatus.SENT) {
            assert(lastSent == i - 1);//, "All SENT items should be at beginning of queue.");
            lastSent = i;
            queue[i].status = TransactionStatus.SENT_NEEDS_ABORT;
            queue[i].abortReason = "set";
          } else {
            assert(queue[i].status == TransactionStatus.RUN);//, "Unexpected transaction status in abort");
            queue[i].unwatcher();
            events.addAll(this._serverSyncTree.ackUserWrite(queue[i].currentWriteId, true));
            if (queue[i].onComplete) {
              var snapshot = null;
              callbacks.add(goog.bind(queue[i].onComplete, null, new Exception("set"), false, snapshot));
            }
          }
        }
      }
      if (lastSent == -1) {
        node.setValue(null);
      } else {
        queue.length = lastSent + 1;
      }
      this._eventQueue.raiseEventsForChangedPath(node.path, events);
      callbacks.forEach(util.exceptionGuard);
    }
  }

}

enum TransactionStatus {RUN, SENT, COMPLETED, SENT_NEEDS_ABORT, NEEDS_ABORT}

const _MAX_TRANSACTION_RETRIES = 25;

class RepoManager {

  final Map<RepoInfo,Repo> _repos = {};
  static bool _useRestClient = false;

  const RepoManager();

  void interrupt() => _repos.values.forEach((r)=>r.interrupt());

  void resume() => _repos.values.forEach((r)=>r.resume());

  getRepo(RepoInfo repoInfo) =>
      _repos.putIfAbsent(repoInfo, () => new Repo(repoInfo, _useRestClient));

  void forceRestClient() {
    _useRestClient = true;
  }

  static RepoManager getInstance() => const RepoManager();

}