part of firebase_dart.core.snap;

class ChildrenNode extends Node {

  final SortedMap<String,Node> _children;
  final Node _priorityNode;
  final Map<Index,SortedMap<NamedNode,Node>> _indexMap;

  const ChildrenNode._(this._children, this._priorityNode, this._indexMap);

  factory ChildrenNode(SortedMap<String,Node> children, Node priorityNode,
      Map<Index,SortedMap<NamedNode,Node>> indexMap) {
    if (priorityNode!=null) {
      validatePriorityNode(priorityNode);
    }
    if (children.isEmpty) {
      assert(priorityNode==null || priorityNode.isEmpty);//, "An empty node cannot have a priority");
    }
    return new ChildrenNode._(children, priorityNode, indexMap);
  }


  bool get isLeafNode => false;

  Node get priority => _priorityNode ?? Node.empty;

  ChildrenNode updatePriority(Node newPriorityNode) {
    if (this._children.isEmpty) {
      return this;
    } else {
      return new ChildrenNode(this._children, newPriorityNode, this._indexMap);
    }
  }

  Node getImmediateChild(String childName) {
    if (childName == ".priority") {
      return this.priority;
    } else {
      var child = this._children[childName];
      return child == null ? EMPTY_NODE : child;
    }
  }

  Node getChild(util.Path path) {
    var front = path.front;
    if (front == null) {
      return this;
    }
    return this.getImmediateChild(front).getChild(path.popFront());
  }

  bool hasChild(String childName) => _children.containsKey(childName);

  Node updateImmediateChild(String childName, Node newChildNode) {
    new SplayTreeMap.from()
    assert(newChildNode!=null); //, "We should always be passing snapshot nodes");
    if (childName == ".priority") {
      return this.updatePriority(newChildNode);
    } else {
      var namedNode = new NamedNode(childName, newChildNode);
      var oldValue = _children[childName];
      Map<String,Node> newChildren;
      Map<Index,SortedMap<NamedNode,Node>> newIndexMap;
      Node newPriority;
      if (newChildNode.isEmpty) {
        if (oldValue==null) return this;

        newChildren = _children.withKeyRemoved(childName);
        newIndexMap = new Map.fromIterables(
            _indexMap.keys,
            _indexMap.values.map((v)=>v.withKeyRemoved(namedNode)));
      } else {
        var oldNamedNode = oldValue!=null ? new NamedNode(childName, oldValue) : null;

        newChildren = _children.withKeyValueAdded(childName, newChildNode);
        newIndexMap = new Map.fromIterables(
            _indexMap.keys,
            _indexMap.values.map((v) {
              if (oldValue!=null) {
                v = v.withKeyRemoved(oldNamedNode);
              }
              return v.withKeyValueAdded(namedNode, newChildNode);
            }));
      }
      newPriority = newChildren.isEmpty ? EMPTY_NODE : this._priorityNode;
      return new ChildrenNode(newChildren, newPriority, newIndexMap);
    }
  }

  Node updateChild(util.Path path, Node newChildNode) {
    var front = path.front;
    if (front == null) {
      return newChildNode;
    } else {
      assert(path.front != ".priority" || path.length == 1); //, ".priority must be the last token in a path");
      var newImmediateChild = this.getImmediateChild(front).updateChild(path.popFront(), newChildNode);
      return this.updateImmediateChild(front, newImmediateChild);
    }
  }

  bool get isEmpty => _children.isEmpty;

  int get numChildren => _children.length;

  static final _INTEGER_REGEXP = new RegExp(r"^(0|[1-9]\d*)$");

  dynamic val([bool exportFormat = false]) {
    if (this.isEmpty) {
      return null;
    }
    var obj = {};
    var numKeys = 0, maxKey = 0, allIntegerKeys = true;

    forEachChild(priorityIndex, (String key, Node childNode) {
      obj[key] = childNode.val(exportFormat);
      numKeys++;
      if (allIntegerKeys && ChildrenNode._INTEGER_REGEXP.hasMatch(key)) {
        maxKey = math.max(maxKey, int.parse(key));
      } else {
        allIntegerKeys = false;
      }
    });
    if (!exportFormat && allIntegerKeys && maxKey < 2 * numKeys) {
      var array = [];
      for (var key in obj) {
        array[key] = obj[key];
      }
      return array;
    } else {
      if (exportFormat && !this.priority.isEmpty) {
        obj[".priority"] = this.priority.val();
      }
      return obj;
    }
  }

  int get hashCode {
    if (this._lazyHash == null) {
      var toHash = "";
      if (!this.priority.isEmpty) {
        toHash += "priority:" + priorityHashText((this.priority.val())) + ":";
      }
      this.forEachChild(priorityIndex, (key, childNode) {
        var childHash = childNode.hashCode;
        if (childHash != "") {
          toHash += ":" + key + ":" + childHash;
        }
      });
      this._lazyHash = toHash == "" ? "" : util.sha1(toHash);
    }
    return this._lazyHash;
  }

  getPredecessorChildName(String childName, Node childNode, Index index) {
    var idx = this._resolveIndex(index);
    if (idx) {
      var predecessor = idx.lastKeyBefore(new NamedNode(childName, childNode));
      return predecessor ? predecessor.name : null;
    } else {
      return this._children.lastKeyBefore(childName);
    }
  }

  getFirstChildName(indexDefinition) {
    var idx = this._resolveIndex(indexDefinition);
    if (idx) {
      var minKey = idx.firstKey();
      return minKey && minKey.name;
    } else {
      return this._children.firstKey();
    }
  }

  getFirstChild(indexDefinition) {
    var minKey = this.getFirstChildName(indexDefinition);
    if (minKey) {
      return new NamedNode(minKey, this._children[minKey]);
    } else {
      return null;
    }
  }

  getLastChildName(indexDefinition) {
    var idx = this._resolveIndex(indexDefinition);
    if (idx) {
      var maxKey = idx.lastKey();
      return maxKey && maxKey.name;
    } else {
      return this._children.lastKey;
    }
  }

  getLastChild(indexDefinition) {
    var maxKey = this.getLastChildName(indexDefinition);
    if (maxKey) {
      return new NamedNode(maxKey, this._children[maxKey]);
    } else {
      return null;
    }
  }

  void forEachChild(Index index, void action(String, Node)) {
    var idx = this._resolveIndex(index);
    if (idx) {
      return idx.keys.forEach((wrappedNode) {
        return action(wrappedNode.name, wrappedNode.node);
      });
    } else {
      return this._children.forEach(action);
    }
  }

  getIterator(indexDefinition) {
    return this.getIteratorFrom(indexDefinition.minPost(), indexDefinition);
  }

  iteratorFrom(startPost, indexDefinition) {
    var idx = this._resolveIndex(indexDefinition);
    if (idx) {
      return idx.iteratorFrom(startPost).map((key) {
        return key;
      });
    } else {
      var iterator = this._children.iteratorFrom(startPost.name, NamedNode.wrap);
      var next = iterator.peek();
      while (next != null && indexDefinition.compare(next, startPost) < 0) {
        iterator.getNext();
        next = iterator.peek();
      }
      return iterator;
    }
  }

  getReverseIterator(indexDefinition) {
    return this.getReverseIteratorFrom(indexDefinition.maxPost(), indexDefinition);
  }

  getReverseIteratorFrom(endPost, indexDefinition) {
    var idx = this._resolveIndex(indexDefinition);
    if (idx) {
      return idx.getReverseIteratorFrom(endPost, (key) {
        return key;
      });
    } else {
      var iterator = this._children.getReverseIteratorFrom(endPost.name, NamedNode.wrap);
      var next = iterator.peek();
      while (next != null && indexDefinition.compare(next, endPost) > 0) {
        iterator.getNext();
        next = iterator.peek();
      }
      return iterator;
    }
  }

  int compareTo(Node other) {
    if (this.isEmpty) {
      if (other.isEmpty) {
        return 0;
      } else {
        return-1;
      }
    } else {
      if (other.isLeafNode || other.isEmpty) {
        return 1;
      } else {
        if (other == MAX_NODE) {
          return-1;
        } else {
          return 0;
        }
      }
    }
  }

  Node withIndex(Index indexDefinition) {
    if (indexDefinition == keyIndex || _indexMap.containsKey(indexDefinition)) {
      return this;
    } else {
      var newIndexMap = new Map.from(this._indexMap)..[indexDefinition] = this._children;
      return new ChildrenNode(this._children, this._priorityNode, newIndexMap);
    }
  }

  bool isIndexed(Index index) {
    return index == keyIndex || this._indexMap.containsKey(index);
  }

  bool operator==(Node other) {
    if (other == this) {
      return true;
    } else {
      if (other.isLeafNode) {
        return false;
      } else {
        var otherChildrenNode = (other as ChildrenNode);
        if (this.priority!=(otherChildrenNode.priority)) {
          return false;
        } else {
          if (this._children.length == otherChildrenNode._children.length) {
            var thisIter = this.getIterator(priorityIndex);
            var otherIter = otherChildrenNode.getIterator(priorityIndex);
            var thisCurrent = thisIter.getNext();
            var otherCurrent = otherIter.getNext();
            while (thisCurrent && otherCurrent) {
              if (thisCurrent.name != otherCurrent.name || !thisCurrent.node.equals(otherCurrent.node)) {
                return false;
              }
              thisCurrent = thisIter.getNext();
              otherCurrent = otherIter.getNext();
            }
            return thisCurrent == null && otherCurrent == null;
          } else {
            return false;
          }
        }
      }
    }
  }

  SortedMap<NamedNode,Node> _resolveIndex(Index indexDefinition) {
    if (indexDefinition == keyIndex) {
      return null;
    } else {
      return this._indexMap[indexDefinition];
    }
  }

  toString() => JSON.encode(this.val(true));
}
