
import 'package:firebase_dart/src/dart/core/util.dart' as util;
import 'package:firebase_dart/src/dart/core/util/tree.dart';
import 'package:firebase_dart/src/dart/core/snap.dart';
import 'package:firebase_dart/src/dart/core/view.dart';
import 'package:firebase_dart/src/dart/core/operation.dart';
import 'package:firebase_dart/src/dart/api.dart';
import 'package:firebase_dart/src/dart/core/persistentconnection.dart';

class RemovedAndEvents {
  final List<Query> removed;
  final List<CancelEvent> events;
  RemovedAndEvents(this.removed, this.events);
}

class ListenProvider {
  final Function startListening; //TODO signature
  final Function stopListening; //TODO signature

  ListenProvider({this.startListening, this.stopListening});
}

class SyncPoint {
  final views_ = <String,View>{};

  bool get isEmpty => views_.isEmpty;

  List<Event> applyOperation(Operation operation, WriteTree writesCache,
      [Node completeServerCache]) {
    var queryId = operation.source.queryId;
    if (queryId != null) {
      var view = views_[queryId];

      assert(view != null); //, "SyncTree gave us an op for an invalid query.");
      return view.applyOperation(
          operation, writesCache, completeServerCache);
    } else {
      var events = [];
      views_.values.forEach((view) {
        events.addAll(view.applyOperation(
            operation, writesCache, completeServerCache));
      });
      return events;
    }
  }

  List<Event> addEventRegistration(Query query, EventRegistration eventRegistration,
      WriteTree writesCache, Node serverCache, bool serverCacheComplete) {
    var queryId = query.queryIdentifier();
    var view = views_[queryId];
    if (!view) {
      var eventCache = writesCache.calcCompleteEventCache(
          serverCacheComplete ? serverCache : null);
      var eventCacheComplete = false;
      if (eventCache) {
        eventCacheComplete = true;
      } else {
        if (serverCache is ChildrenNode) {
          eventCache = writesCache.calcCompleteEventChildren(serverCache);
          eventCacheComplete = false;
        } else {
          eventCache = EMPTY_NODE;
          eventCacheComplete = false;
        }
      }
      var viewCache = new ViewCache(
          new CacheNode((eventCache), eventCacheComplete, false),
          new CacheNode(
              (serverCache), serverCacheComplete, false));
      view = new View(query, viewCache);
      this.views_[queryId] = view;
    }
    view.addEventRegistration(eventRegistration);
    return view.getInitialEvents(eventRegistration);
  }

  RemovedAndEvents removeEventRegistration(Query query, EventRegistration eventRegistration, Exception cancelError) {
    var queryId = query.queryIdentifier();
    var removed = <Query>[];
    var cancelEvents = [];
    var hadCompleteView = this.hasCompleteView();
    if (queryId == "default") {
      var self = this;
      views_.forEach((viewQueryId, view) {
        cancelEvents.addAll(
            view.removeEventRegistration(eventRegistration, cancelError));
        if (view.isEmpty) {
          self.views_.remove(viewQueryId);
          if (!view.query.queryParams.loadsAllData) {
            removed.add(view.query);
          }
        }
      });
    } else {
      var view = views_[queryId];
      if (view) {
        cancelEvents.addAll(
            view.removeEventRegistration(eventRegistration, cancelError));
        if (view.isEmpty) {
          this.views_.remove(queryId);
          if (!view.query.queryParams.loadsAllData) {
            removed.add(view.query);
          }
        }
      }
    }
    if (hadCompleteView && !this.hasCompleteView()) {
      removed.add(new Firebase.fromRepo(query.repo, query.path));
    }
    return new RemovedAndEvents(removed, cancelEvents);
  }


  Iterable<View> getQueryViews() {
    return views_.values.where((view) {
      return !view.query.queryParams.loadsAllData;
    });
  }

  Node getCompleteServerCache(util.Path path) {
    Node serverCache = null;
    views_.forEach((key, view) {
      serverCache = serverCache ?? view.getCompleteServerCache(path);
    });
    return serverCache;
  }

  View viewForQuery(Query query) {
    var params = query.queryParams;
    if (params.loadsAllData) {
      return this.getCompleteView();
    } else {
      var queryId = query.queryIdentifier();
      return views_[queryId];
    }
  }

  bool viewExistsForQuery(Query query) {
    return this.viewForQuery(query) != null;
  }

  bool hasCompleteView() {
    return this.getCompleteView() != null;
  }

  View getCompleteView() {
    return views_.values.firstWhere((view) {
      return view.query.queryParams.loadsAllData;
    }, orElse: ()=>null);
  }
}

class SyncTree {
  
  final ListenProvider listenProvider_;
  ImmutableTree<SyncPoint> syncPointTree_ = ImmutableTree.empty;
  final WriteTree pendingWriteTree_ = new WriteTree();
  final tagToQueryMap_ = <String,QueryKey>{};
  final queryToTagMap_ = <QueryKey,String>{};

  SyncTree(this.listenProvider_);
  
  List<Event> applyUserOverwrite(util.Path path, Node newData, int writeId, [bool visible = true]) {
    this.pendingWriteTree_.addOverwrite(path, newData, writeId, visible);
    if (!visible) {
      return [];
    } else {
      return this.applyOperationToSyncPoints_(
          new Overwrite(OperationSource.user, path, newData));
    }
  }

  List<Event> applyUserMerge(util.Path path, Map<String,Node> changedChildren,
      int writeId) {
    this.pendingWriteTree_.addMerge(path, changedChildren, writeId);
    var changeTree = new ImmutableTree.fromObject(changedChildren);
    return this.applyOperationToSyncPoints_(
        new Merge(OperationSource.user, path, changeTree));
  }

  List<Event> ackUserWrite(int writeId, [bool revert = false]) {
    var write = this.pendingWriteTree_.getWrite(writeId);
    var needToReevaluate = this.pendingWriteTree_.removeWrite(writeId);
    if (!needToReevaluate) {
      return [];
    } else {
      var affectedTree = ImmutableTree.empty;
      if (write.snap != null) {
        affectedTree = affectedTree.set(util.Path.empty, true);
      } else {
        write.children.forEach((pathString, node) {
          affectedTree =
              affectedTree.set(new util.Path.fromString(pathString), node);
        });
      }
      return this.applyOperationToSyncPoints_(
          new AckUserWrite(write.path, affectedTree, revert));
    }
  }

  List<Event> applyServerOverwrite(util.Path path, Node newData) {
    return this.applyOperationToSyncPoints_(
        new Overwrite(OperationSource.server, path, newData));
  }

  List<Event> applyServerMerge(util.Path path, Map<String,Node> changedChildren) {
    var changeTree = new ImmutableTree.fromObject(changedChildren);
    return this.applyOperationToSyncPoints_(
        new Merge(OperationSource.server, path, changeTree));
  }

  List<Event> applyListenComplete(util.Path path) {
    return this.applyOperationToSyncPoints_(
        new ListenComplete(OperationSource.server, path));
  }

  List<Event> applyTaggedQueryOverwrite(util.Path path, Node snap, String tag) {
    var queryKey = this.queryKeyForTag_(tag);
    if (queryKey != null) {
      var queryPath = queryKey.path,
          queryId = queryKey.queryId;
      var relativePath = util.Path.relativePath(queryPath, path);
      var op = new Overwrite(
          new OperationSource.forServerTaggedQuery(queryId), relativePath,
          snap);
      return this.applyTaggedOperation_(queryPath, queryId, op);
    } else {
      return [];
    }
  }

  List<Event> applyTaggedQueryMerge(util.Path path,
      Map<String,Node> changedChildren, String tag) {
    var queryKey = this.queryKeyForTag_(tag);
    if (queryKey) {
      var queryPath = queryKey.path,
          queryId = queryKey.queryId;
      var relativePath = util.Path.relativePath(queryPath, path);
      var changeTree = new ImmutableTree.fromObject(changedChildren);
      var op = new Merge(
          new OperationSource.forServerTaggedQuery(queryId), relativePath,
          changeTree);
      return this.applyTaggedOperation_(queryPath, queryId, op);
    } else {
      return [];
    }
  }

  List<Event> applyTaggedListenComplete(util.Path path, String tag) {
    var queryKey = this.queryKeyForTag_(tag);
    if (queryKey) {
      var queryPath = queryKey.path,
          queryId = queryKey.queryId;
      var relativePath = util.Path.relativePath(queryPath, path);
      var op = new ListenComplete(
          new OperationSource.forServerTaggedQuery(queryId), relativePath);
      return this.applyTaggedOperation_(queryPath, queryId, op);
    } else {
      return [];
    }
  }

  List<Event> addEventRegistration(Query query, EventRegistration eventRegistration) {
    var path = query.path;
    Node serverCache = null;
    var foundAncestorDefaultView = false;
    this.syncPointTree_.forEachOnPath(path, (pathToSyncPoint, sp) {
      var relativePath = util.Path.relativePath(pathToSyncPoint, path);
      serverCache = serverCache ?? sp.getCompleteServerCache(relativePath);
      foundAncestorDefaultView =
          foundAncestorDefaultView || sp.hasCompleteView();
    });
    var syncPoint = this.syncPointTree_.get(path);
    if (!syncPoint) {
      syncPoint = new SyncPoint();
      this.syncPointTree_ = this.syncPointTree_.set(path, syncPoint);
    } else {
      foundAncestorDefaultView =
          foundAncestorDefaultView || syncPoint.hasCompleteView();
      serverCache = serverCache ??
          syncPoint.getCompleteServerCache(util.Path.empty);
    }
    var serverCacheComplete;
    if (serverCache != null) {
      serverCacheComplete = true;
    } else {
      serverCacheComplete = false;
      serverCache = EMPTY_NODE;
      var subtree = this.syncPointTree_.subtree(path);
      subtree.forEachChild((childName, childSyncPoint) {
        var completeCache = childSyncPoint.getCompleteServerCache(
            util.Path.empty);
        if (completeCache) {
          serverCache =
              serverCache.updateImmediateChild(childName, completeCache);
        }
      });
    }
    var viewAlreadyExists = syncPoint.viewExistsForQuery(query);
    if (!viewAlreadyExists && !query.queryParams.loadsAllData) {
      var queryKey = new QueryKey.fromQuery(query);
      assert(!this.queryToTagMap_.containsKey(queryKey));
          //"View does not exist, but we have a tag");
      var tag = getNextQueryTag_();
      this.queryToTagMap_[queryKey] = tag;
      this.tagToQueryMap_["_" + tag] = queryKey;
    }
    var writesCache = this.pendingWriteTree_.childWrites(path);
    var events = syncPoint.addEventRegistration(
        query, eventRegistration, writesCache, serverCache,
        serverCacheComplete);
    if (!viewAlreadyExists && !foundAncestorDefaultView) {
      var view = (syncPoint.viewForQuery(query));
      events.addAll(this.setupListener_(query, view));
    }
    return events;
  }

  List<Event> removeEventRegistration(Query query,
      EventRegistration eventRegistration, Exception cancelError) {
    var path = query.path;
    var maybeSyncPoint = this.syncPointTree_.get(path);
    var cancelEvents = [];
    if (maybeSyncPoint && (query.queryIdentifier() == "default" ||
        maybeSyncPoint.viewExistsForQuery(query))) {
      var removedAndEvents = maybeSyncPoint.removeEventRegistration(
          query, eventRegistration, cancelError);
      if (maybeSyncPoint.isEmpty) {
        this.syncPointTree_ = this.syncPointTree_.remove(path);
      }
      var removed = removedAndEvents.removed;
      cancelEvents = removedAndEvents.events;
      var removingDefault = removed.firstWhere((q)=>q.queryParams.loadsAllData, orElse: ()=>null) !=null;
      var covered = this.syncPointTree_.findOnPath(
          path, (relativePath, parentSyncPoint) {
        return parentSyncPoint.hasCompleteView();
      });
      if (removingDefault && !covered) {
        var subtree = this.syncPointTree_.subtree(path);
        if (!subtree.isEmpty) {
          var newViews = this.collectDistinctViewsForSubTree_(subtree);
          for (var i = 0; i < newViews.length; ++i) {
            var view = newViews[i],
                newQuery = view.query;
            var listener = this.createListenerForView_(view);
            this.listenProvider_.startListening(
                this.queryForListening_(newQuery), this.tagForQuery_(newQuery),
                listener.hashFn, listener.onComplete);
          }
        } else {}
      }
      if (!covered && removed.length > 0 && cancelError==null) {
        if (removingDefault) {
          var defaultTag = null;
          this.listenProvider_.stopListening(
              this.queryForListening_(query), defaultTag);
        } else {
          removed.forEach((queryToRemove) {
            var queryIdToRemove = queryToRemove.queryIdentifier();
            var tagToRemove = queryToTagMap_[new QueryKey.fromQuery(
                queryToRemove)];
            listenProvider_.stopListening(
                queryForListening_(queryToRemove), tagToRemove);
          });
        }
      }
      this.removeTags_(removed);
    }
    return cancelEvents;
  }

  Node calcCompleteEventCache(util.Path path, [writeIdsToExclude = false]) {
    var includeHiddenSets = true;
    var writeTree = this.pendingWriteTree_;
    Node serverCache = this.syncPointTree_.findOnPath(
        path, (pathSoFar, syncPoint) {
      var relativePath = util.Path.relativePath(pathSoFar, path);
      var serverCache = syncPoint.getCompleteServerCache(relativePath);
      if (serverCache) {
        return serverCache;
      }
    });
    return writeTree.calcCompleteEventCache(
        path, serverCache, writeIdsToExclude, includeHiddenSets);
  }

  List<View> collectDistinctViewsForSubTree_(ImmutableTree<SyncPoint> subtree) {
    return subtree.fold((relativePath, SyncPoint maybeChildSyncPoint,
        Map<String,List<View>> childMap) {
      if (maybeChildSyncPoint?.hasCompleteView()==true) {
        var completeView = maybeChildSyncPoint.getCompleteView();
        return [completeView];
      } else {
        List<View> views = <View>[];
        if (maybeChildSyncPoint!=null) {
          views = maybeChildSyncPoint.getQueryViews();
        }
        childMap.forEach((key,childViews) {
          views.addAll(childViews);
        });
        return views;
      }
    });
  }

  void removeTags_(List<Query> queries) {
    for (var j = 0; j < queries.length; ++j) {
      var removedQuery = queries[j];
      if (!removedQuery.queryParams.loadsAllData) {
        var removedQueryKey = new QueryKey.fromQuery(removedQuery);
        var removedQueryTag = this.queryToTagMap_[removedQueryKey];

        this.queryToTagMap_.remove(removedQueryKey);

        this.tagToQueryMap_.remove("_" + removedQueryTag);
      }
    }
  }

  Query queryForListening_(Query query) {
    if (query.queryParams.loadsAllData &&
        !query.queryParams.isDefault) {
      return (query.ref);
    } else {
      return query;
    }
  }

  List<Event> setupListener_(Query query, View view) {
    var path = query.path;
    var tag = this.tagForQuery_(query);
    var listener = this.createListenerForView_(view);
    List<Event> events = this.listenProvider_.startListening(
        this.queryForListening_(query), tag, listener.hashFn,
        listener.onComplete);
    var subtree = this.syncPointTree_.subtree(path);
    if (tag) {
      assert(!subtree.value
          .hasCompleteView()); //, "If we're adding a query, it shouldn't be shadowed");
    } else {
      var queriesToStop = subtree.fold((relativePath, maybeChildSyncPoint,
          childMap) {
        if (!relativePath.isEmpty && maybeChildSyncPoint &&
            maybeChildSyncPoint.hasCompleteView()) {
          return [maybeChildSyncPoint.getCompleteView().query];
        } else {
          var queries = [];
          if (maybeChildSyncPoint) {
            queries.addAll(
                maybeChildSyncPoint.getQueryViews().map((view) => view.query));
          }
          childMap.forEach((key, childQueries) {
            queries.addAll(childQueries);
          });
          return queries;
        }
      });
      for (var i = 0; i < queriesToStop.length; ++i) {
        var queryToStop = queriesToStop[i];
        this.listenProvider_.stopListening(this.queryForListening_(queryToStop),
            this.tagForQuery_(queryToStop));
      }
    }
    return events;
  }

  ListenSpec createListenerForView_(View view) {
    var query = view.query;
    var tag = this.tagForQuery_(query);
    return new ListenSpec(hashFn:() {
      var cache = view.serverCache ?? EMPTY_NODE;
      return cache.hashCode;
    }, onComplete:(status, data) {
      if (status == "ok") {
        if (tag) {
          return applyTaggedListenComplete(query.path, tag);
        } else {
          return applyListenComplete(query.path);
        }
      } else {
        var error = util.errorForServerCode(status);
        return removeEventRegistration(query, null, error);
      }
    });
  }

  QueryKey queryKeyForTag_(String tag) => tagToQueryMap_["_$tag"];

  String tagForQuery_(Query query) {
    var queryKey = new QueryKey.fromQuery(query);
    return this.queryToTagMap_[queryKey];
  }

  static int nextQueryTag_ = 1;

  static getNextQueryTag_() {
    return nextQueryTag_++;
  }

  List<Event> applyTaggedOperation_(util.Path queryPath, String queryId, Operation operation) {
    var syncPoint = this.syncPointTree_.get(queryPath);
    assert(syncPoint!=null); //, "Missing sync point for query tag that we're tracking");
    var writesCache = this.pendingWriteTree_.childWrites(queryPath);
    return syncPoint.applyOperation(operation, writesCache, null);
  }

  List<Event> applyOperationToSyncPoints_(Operation operation) {
    return this.applyOperationHelper_(operation, this.syncPointTree_, null,
        this.pendingWriteTree_.childWrites(util.Path.empty));
  }

  List<Event> applyOperationHelper_(Operation operation,
      ImmutableTree<SyncPoint> syncPointTree, Node serverCache,
      WriteTreeRef writesCache) {
    if (operation.path.isEmpty) {
      return this.applyOperationDescendantsHelper_(
          operation, syncPointTree, serverCache, writesCache);
    } else {
      var syncPoint = syncPointTree.get(util.Path.empty);
      if (serverCache == null && syncPoint != null) {
        serverCache = syncPoint.getCompleteServerCache(util.Path.empty);
      }
      var events = [];
      var childName = operation.path.front;
      var childOperation = operation.operationForChild(childName);
      var childTree = syncPointTree.children[childName];
      if (childTree && childOperation) {
        var childServerCache = serverCache?.getImmediateChild(childName);
        var childWritesCache = writesCache.child(childName);
        events.addAll(this.applyOperationHelper_(
            childOperation, childTree, childServerCache, childWritesCache));
      }
      if (syncPoint!=null) {
        events.addAll(
            syncPoint.applyOperation(operation, writesCache, serverCache));
      }
      return events;
    }
  }

  List<Event> applyOperationDescendantsHelper_(Operation operation,
      ImmutableTree<SyncPoint> syncPointTree, Node serverCache,
      WriteTreeRef writesCache) {
    var syncPoint = syncPointTree.get(util.Path.empty);
    if (serverCache == null && syncPoint != null) {
      serverCache = syncPoint.getCompleteServerCache(util.Path.empty);
    }
    var events = [];
    var self = this;
    syncPointTree.children.forEach((childName, childTree) {
      var childServerCache = serverCache?.getImmediateChild(childName);
      var childWritesCache = writesCache.child(childName);
      var childOperation = operation.operationForChild(childName);
      if (childOperation) {
        events.addAll(self.applyOperationDescendantsHelper_(
            childOperation, childTree, childServerCache, childWritesCache));
      }
    });
    if (syncPoint) {
      events.addAll(
          syncPoint.applyOperation(operation, writesCache, serverCache));
    }
    return events;
  }

}


class QueryKey {
  final String queryId;
  final util.Path path;

  QueryKey(this.queryId, this.path);

  QueryKey.fromQuery(Query query) : this.path = query.path, queryId = query.queryIdentifier();

  static QueryKey parse(String str) {
    var splitIndex = str.indexOf(r"$");
    assert(splitIndex != -1 &&
        splitIndex < str.length - 1); //, "Bad queryKey.");
    return new QueryKey(
      str.substring(splitIndex + 1),
      new util.Path.fromString(str.substring(0, splitIndex))
    );
  }

  toString() => "$path\$$queryId";


}