
import 'dart:collection';

abstract class Pair<K,V> {
  K get key;
  V get value;
}

class LLRBNode<K,V> extends IterableBase<Pair<K,V>> implements Pair<K,V> {
  static const RED = true;
  static const BLACK = false;

  final Comparator<K> comparator;

  K get key => null;
  V get value => null;

  factory LLRBNode.nil({Comparator<K> comparator}) => new LLRBNode._(comparator ?? Comparable.compare);

  const LLRBNode._(this.comparator);

  factory LLRBNode(K key, V value, {bool color: RED, Comparator<K> comparator, LLRBNode<K,V> left, LLRBNode<K,V> right}) {
    comparator = comparator ?? Comparable.compare;
    var nil = new LLRBNode._(comparator);
    return new _NonNilNode<K,V>(key, value, color, left ?? nil, right ?? nil, comparator);
  }

  LLRBNode<K,V> copy({K key, V value, bool color, _NonNilNode<K, V> left,
  _NonNilNode<K, V> right}) => this;

  _NonNilNode<K,V> insert(K key, V value) =>
      new LLRBNode<K,V>(key, value, comparator: comparator, left: this, right: this);


  LLRBNode<K,V> remove(K key) => this;

  int get length => 0;

  bool get isEmpty => true;

  LLRBNode<K,V> get _min =>
      left is _NonNilNode ? (left as _NonNilNode)._min : this;

  K firstKey() => _min?.key;

  K lastKey() => right is _NonNilNode ? right.lastKey() : key;

  int get blackDepth => 0;

  bool get isRed => false;

  LLRBNode<K, V> get left => this;
  LLRBNode<K, V> get right => this;


  @override
  Iterator<Pair<K, V>> get iterator => new _Iterator(this, false);

  Iterator<Pair<K, V>> get reversedIterator => new _Iterator(this, true);

  Iterator<Pair<K, V>> iteratorFrom(K key) => new _Iterator(this, false, key);

  Iterator<Pair<K, V>> reversedIteratorFrom(K key) => new _Iterator(this, true, key);

}

class _Iterator<K,V> implements Iterator<LLRBNode<K,V>> {


  final LLRBNode<K, V> _root;
  final bool _reversed;
  Pair<K,V> _current;
  final Queue<LLRBNode<K, V>> _queue = new Queue();

  _Iterator(this._root, this._reversed, [K startKey]) {
    var cmp = 1;
    var n = this._root;
    while (!n.isEmpty) {
      cmp = startKey!=null ? _root.comparator(n.key, startKey) : 1;
      if (_reversed) {
        cmp *= -1;
      }
      if (cmp < 0) {
        if (this._reversed) {
          n = n.left;
        } else {
          n = n.right;
        }
      } else {
        if (cmp == 0) {
          _queue.addLast(n);
          break;
        } else {
          _queue.addLast(n);
          if (_reversed) {
            n = n.right;
          } else {
            n = n.left;
          }
        }
      }
    }
  }

  @override
  Pair<K, V> get current => _current;

  @override
  bool moveNext() {
    if (_queue.isEmpty) {
      return false;
    }
    var node = _queue.removeLast();
    _current = node;
    if (this._reversed) {
      node = node.left;
      while (!node.isEmpty) {
        _queue.addLast(node);
        node = node.right;
      }
    } else {
      node = node.right;
      while (!node.isEmpty) {
        _queue.addLast(node);
        node = node.left;
      }
    }
    return true;
  }

}

class _NonNilNode<K,V> extends LLRBNode<K,V> {

  final K key;
  final V value;
  final bool color;
  final LLRBNode<K, V> left;
  final LLRBNode<K, V> right;

  _NonNilNode(this.key, this.value, this.color, this.left, this.right, Comparator<K> comparator)
      : super._(comparator);

  _NonNilNode<K,V> copy({K key, V value, bool color, _NonNilNode<K, V> left,
      _NonNilNode<K, V> right}) {
    return new _NonNilNode<K,V>(
        key ?? this.key,
        value ?? this.value,
        color ?? this.color,
        left ?? this.left,
        right ?? this.right,
        comparator);
  }

  int get length => left.length + 1 + right.length;

  bool get isEmpty => false;


  _NonNilNode<K, V> insert(K key, V value) {
    var n = this;
    int cmp = comparator(key, n.key);
    if (cmp < 0) {
      n = n.copy(left: n.left.insert(key, value));
    } else {
      if (cmp == 0) {
        n = n.copy(value: value);
      } else {
        n = n.copy(right: n.right.insert(key, value));
      }
    }
    return n._fixUp();
  }

  _NonNilNode<K, V> _removeMin() {
    var n;
    if (left.isEmpty) {
      return left;
    }
    n = this;
    if (!n.left.isRed && !n.left.left.isRed) {
      n = n._moveRedLeft();
    }
    n = n.copy(left: n.left._removeMin());
    return n._fixUp();
  }

  _NonNilNode<K, V> remove(K key) {
    var n, smallest;
    n = this;
    if (comparator(key, n.key) < 0) {
      if (!n.left.isEmpty && !n.left.isRed && !n.left.left.isRed) {
        n = n._moveRedLeft;
      }
      n = n.copy(null, null, null, n.left.remove(key, comparator), null);
    } else {
      if (n.left.isRed) {
        n = n._rotateRight();
      }
      if (!n.right.isEmpty && !n.right.isRed && !n.right.left.isRed) {
        n = n._moveRedRight();
      }
      if (comparator(key, n.key) == 0) {
        if (n.right.isEmpty) {
          return n.right;
        } else {
          smallest = n.right._min;
          n = n.copy(key: smallest.key, value: smallest.value,
              right: n.right._removeMin());
        }
      }
      n = n.copy(right: n.right.remove(key));
    }
    return n._fixUp();
  }

  bool get isRed => color;

  _NonNilNode<K, V> _fixUp() {
    var n = this;
    if (n.right.isRed && !n.left.isRed) {
      n = n._rotateLeft();
    }
    if (n.left.isRed && n.left.left.isRed) {
      n = n._rotateRight();
    }
    if (n.left.isRed && n.right.isRed) {
      n = n._colorFlip();
    }
    return n;
  }

  _NonNilNode<K, V> _moveRedLeft() {
    var n = this._colorFlip();
    if (n.right.left.isRed) {
      n = n.copy(right: n.right._rotateRight());
      n = n._rotateLeft();
      n = n._colorFlip();
    }
    return n;
  }

  _NonNilNode<K, V> _moveRedRight() {
    var n = this._colorFlip();
    if (n.left.left.isRed) {
      n = n._rotateRight();
      n = n._colorFlip();
    }
    return n;
  }

  _NonNilNode<K, V> _rotateLeft() {
    var nl;
    nl = this.copy(color: LLRBNode.RED, right: this.right.left);
    return this.right.copy(color: this.color, left: nl);
  }

  _NonNilNode<K, V> _rotateRight() {
    var nr;
    nr = this.copy(color: LLRBNode.RED, left: this.left.right);
    return this.left.copy(color: this.color, right: nr);
  }

  _NonNilNode<K, V> _colorFlip() {
    var left, right;
    left = this.left.copy(color: !this.left.isRed);
    right = this.right.copy(color: !this.right.isRed);
    return this.copy(color: !this.color, left: left, right: right);
  }


  int get blackDepth {
    assert(!isRed || !left.isRed); // Red node has red child
    assert(!right.isRed); // Right child is red
    var blackDepth = this.left.blackDepth;
    assert(blackDepth==right.blackDepth); // Black depths differ
    return blackDepth + (this.isRed ? 0 : 1);
  }
}
