part of firebase_dart.core.util;

class Path extends IterableBase<String> {

  static const empty = const Path._();


  final List<String> _pieces;
  final int _pieceNum;
  
  const Path._([this._pieces = const[], this._pieceNum = 0]);

  factory Path.fromString(String str) {
    var pieces = str.split("/").where((v)=>v.isNotEmpty).toList();
    return new Path._(pieces, 0);
  }

  String get front {
    if (this._pieceNum >= this._pieces.length) {
      return null;
    }
    return this._pieces[this._pieceNum];
  }

  get length => this._pieces.length - this._pieceNum;

  Path popFront() {
    var pieceNum = this._pieceNum;
    if (pieceNum < this._pieces.length) {
      pieceNum++;
    }
    return new Path._(this._pieces, pieceNum);
  }

  String get back {
    if (this._pieceNum < this._pieces.length) {
      return this._pieces[this._pieces.length - 1];
    }
    return null;
  }

  String toString() {
    var pathString = _pieces.skip(_pieceNum).where((v)=>v.isNotEmpty)
        .map((v)=>"/$v").join();
    return pathString.isEmpty ? "/" : pathString;
  }

  String toUrlEncodedString() {
    var pathString = _pieces.skip(_pieceNum).where((v)=>v.isNotEmpty)
        .map((v)=>"/${Uri.encodeComponent(v)}").join();
    return pathString.isEmpty ?? "/";
  }

  Path get parent {
    if (this._pieceNum >= this._pieces.length) {
      return null;
    }
    var pieces = this._pieces.sublist(_pieceNum)..removeLast();
    return new Path._(pieces, 0);
  }

  Path child(childPathObj) {
    var pieces = _pieces.sublist(_pieceNum);
    if (childPathObj is Path) {
      pieces.addAll(childPathObj);
    } else {
      pieces.addAll(childPathObj.split("/").where((v)=>v.isNotEmpty));
    }
    return new Path._(pieces, 0);
  }

  bool get isEmpty => this._pieceNum >= this._pieces.length;

  static relativePath(Path outerPath, Path innerPath) {
    var outer = outerPath.front, inner = innerPath.front;
    if (outer == null) {
      return innerPath;
    } else {
      if (outer == inner) {
        return relativePath(outerPath.popFront(), innerPath.popFront());
      } else {
        throw new Exception("INTERNAL ERROR: innerPath ($innerPath) is not "
            "within " + "outerPath ($outerPath)");
      }
    }
  }

  static int comparePaths(Path left, Path right) {
    var leftKeys = left.toList();
    var rightKeys = right.toList();
    for (var i = 0;i < leftKeys.length && i < rightKeys.length;i++) {
      var cmp = nameCompare(leftKeys[i], rightKeys[i]);
      if (cmp != 0) {
        return cmp;
      }
    }
    if (leftKeys.length == rightKeys.length) {
      return 0;
    }
    return leftKeys.length < rightKeys.length ? -1 : 1;
  }

  bool operator==(Path other) {
    if (this.length != other.length) {
      return false;
    }
    for (var i = this._pieceNum, j = other._pieceNum;i <= this._pieces.length;i++, j++) {
      if (this._pieces[i] != other._pieces[j]) {
        return false;
      }
    }
    return true;
  }

  bool contains(Path other) {
    var i = this._pieceNum;
    var j = other._pieceNum;
    if (this.length > other.length) {
      return false;
    }
    while (i < this._pieces.length) {
      if (this._pieces[i] != other._pieces[j]) {
        return false;
      }
      ++i;
      ++j;
    }
    return true;
  }

  @override
  Iterator<String> get iterator => _pieces.skip(_pieceNum).iterator;
}

