
import 'package:firebase_dart/src/dart/core/snap.dart';
import 'package:firebase_dart/src/dart/core/view.dart';
import 'package:firebase_dart/src/dart/core/util.dart' as util;
import 'sorted_map.dart';
import 'dart:convert';

class PathAndValue {  //TODO replace with Pair
  final util.Path path;
  final value;

  PathAndValue(this.path, this.value);
}

class ImmutableTree<T> {

  final T value;
  final SortedMap<String,ImmutableTree<T>> children;

  static const empty = const ImmutableTree(null);

  const ImmutableTree(this.value, [this.children = const SortedMap()]);

  factory ImmutableTree.fromObject(Map<String, Node> obj) {
    var tree = empty;
    obj.forEach((childPath, childSnap) {
      tree = tree.set(new util.Path.fromString(childPath), childSnap);
    });
    return tree;
  }

  bool get isEmpty => value == null && this.children.isEmpty;

  PathAndValue findRootMostMatchingPathAndValue(util.Path relativePath, bool predicate(v)) {
    if (this.value != null && predicate(this.value)) {
      return new PathAndValue(util.Path.empty, this.value);
    } else {
      if (relativePath.isEmpty) {
        return null;
      } else {
        var front = relativePath.front;
        var child = this.children[front];
        if (child != null) {
          var childExistingPathAndValue = child.findRootMostMatchingPathAndValue(relativePath.popFront(), predicate);
          if (childExistingPathAndValue != null) {
            var fullPath = (new util.Path.fromString(front)).child(childExistingPathAndValue.path);
            return new PathAndValue(fullPath, childExistingPathAndValue.value);
          } else {
            return null;
          }
        } else {
          return null;
        }
      }
    }
  }

  findRootMostValueAndPath(util.Path relativePath) {
    return this.findRootMostMatchingPathAndValue(relativePath, (_)=>true);
  }

  ImmutableTree<T> subtree(util.Path relativePath) {
    if (relativePath.isEmpty) {
      return this;
    } else {
      var front = relativePath.front;
      var childTree = this.children[front];
      if (childTree != null) {
        return childTree.subtree(relativePath.popFront());
      } else {
        return empty;
      }
    }
  }

  ImmutableTree set(util.Path relativePath, T toSet) {
    if (relativePath.isEmpty) {
      return new ImmutableTree(toSet, this.children);
    } else {
      var front = relativePath.front;
      ImmutableTree child = this.children[front] ?? empty;
      var newChild = child.set(relativePath.popFront(), toSet);
      var newChildren = this.children.withKeyValueAdded(front, newChild);
      return new ImmutableTree(this.value, newChildren);
    }
  }

  ImmutableTree remove(util.Path relativePath) {
    if (relativePath.isEmpty) {
      if (this.children.isEmpty) {
        return empty;
      } else {
        return new ImmutableTree(null, this.children);
      }
    } else {
      var front = relativePath.front;
      var child = this.children[front];
      if (child) {
        var newChild = child.remove(relativePath.popFront());
        SortedMap newChildren;
        if (newChild.isEmpty) {
          newChildren = this.children.withKeyRemoved(front);
        } else {
          newChildren = this.children.withKeyValueAdded(front, newChild);
        }
        if (this.value == null && newChildren.isEmpty) {
          return empty;
        } else {
          return new ImmutableTree(this.value, newChildren);
        }
      } else {
        return this;
      }
    }
  }

  T get(util.Path relativePath) {
    if (relativePath.isEmpty) {
      return this.value;
    } else {
      var front = relativePath.front;
      var child = this.children[front];

      if (child!=null) {
        return child.get(relativePath.popFront());
      } else {
        return null;
      }
    }
  }

  setTree(util.Path relativePath, ImmutableTree newTree) {
    if (relativePath.isEmpty) {
      return newTree;
    } else {
      var front = relativePath.front;
      var child = this.children[front] ?? empty;
      var newChild = child.setTree(relativePath.popFront(), newTree);
      var newChildren;
      if (newChild.isEmpty()) {
        newChildren = this.children.withKeyRemoved(front);
      } else {
        newChildren = this.children.withKeyValueAdded(front, newChild);
      }
      return new ImmutableTree(this.value, newChildren);
    }
  }

  fold(fn(util.Path pathSoFar, T v, Map<String,dynamic> accum)) {
    return this._fold(util.Path.empty, fn);
  }

  _fold(util.Path pathSoFar, fn(util.Path pathSoFar, T v, Map<String,dynamic> accum)) {
    Map<String,dynamic> accum = {};
    this.children.forEach((childKey, childTree) {
      accum[childKey] = childTree._fold(pathSoFar.child(childKey), fn);
    });
    return fn(pathSoFar, this.value, accum);
  }

  findOnPath(util.Path path, bool f(util.Path p, T v)) { // replace with iterator and map
    return this._findOnPath(path, util.Path.empty, f);
  }

  _findOnPath(util.Path pathToFollow, util.Path pathSoFar, bool f(util.Path p, T v)) {
    var result = this.value!=null ? f(pathSoFar, this.value) : false;
    if (result) {
      return result;
    } else {
      if (pathToFollow.isEmpty) {
        return null;
      } else {
        var front = pathToFollow.front;
        var nextChild = this.children[front];
        if (nextChild) {
          return nextChild._findOnPath(pathToFollow.popFront(), pathSoFar.child(front), f);
        } else {
          return null;
        }
      }
    }
  }

  void forEachOnPath(util.Path path, void f(util.Path p, T v)) {
    this._forEachOnPath(path, util.Path.empty, f);
  }

  void _forEachOnPath(util.Path pathToFollow, util.Path currentRelativePath, void f(util.Path p,T v)) {
    if (pathToFollow.isEmpty) {
      return;
    } else {
      if (this.value!=null) {
        f(currentRelativePath, this.value);
      }
      var front = pathToFollow.front;
      var nextChild = this.children[front];
      nextChild?._forEachOnPath(pathToFollow.popFront(), currentRelativePath.child(front), f);
    }
  }

  forEach(void f(util.Path p,v)) {
    this._forEach(util.Path.empty, f);
  }

  _forEach(util.Path currentRelativePath, void f(util.Path p,v)) {
    this.children.forEach((childName, childTree) {
      childTree._forEach(currentRelativePath.child(childName), f);
    });
    if (this.value!=null) {
      f(currentRelativePath, this.value);
    }
  }

  forEachChild(void f(String name, T child)) {
    this.children.forEach((childName, childTree) {
      if (childTree.value) {
        f(childName, childTree.value);
      }
    });
  }

  toString() {
    var json = {};
    this.forEach((util.Path relativePath, value) {
      var pathString = relativePath.toString();
      json[pathString] = value.toString();
    });
    return JSON.encode(json);

  }
}


class CompoundWrite {
  final ImmutableTree<Node> writeTree_;
  
  const CompoundWrite(this.writeTree_);
  
  static const empty = const CompoundWrite(const ImmutableTree(null));

  CompoundWrite addWrite(util.Path path, Node node) {
    if (path.isEmpty) {
      return new CompoundWrite(new ImmutableTree(node));
    } else {
      var rootmost = this.writeTree_.findRootMostValueAndPath(path);
      if (rootmost != null) {
        var rootMostPath = rootmost.path, value = rootmost.value;
        var relativePath = util.Path.relativePath(rootMostPath, path);
        value = value.updateChild(relativePath, node);
        return new CompoundWrite(this.writeTree_.set(rootMostPath, value));
      } else {
        var subtree = new ImmutableTree(node);
        var newWriteTree = this.writeTree_.setTree(path, subtree);
        return new CompoundWrite(newWriteTree);
      }
    }
  }
  
  CompoundWrite addWrites(util.Path path, Map<String,Node> updates) {
    var newWrite = this;
    updates.forEach((childKey, node) {
      newWrite = newWrite.addWrite(path.child(childKey), node);
    });
    return newWrite;
  }
  
  CompoundWrite removeWrite(util.Path path) {
    if (path.isEmpty) {
      return CompoundWrite.empty;
    } else {
      var newWriteTree = this.writeTree_.setTree(path, ImmutableTree.empty);
      return new CompoundWrite(newWriteTree);
    }
  }
  
  bool hasCompleteWrite(util.Path path) {
    return this.getCompleteNode(path) != null;
  }
  
  Node getCompleteNode(util.Path path) {
    var rootmost = this.writeTree_.findRootMostValueAndPath(path);
    if (rootmost != null) {
      return this.writeTree_.get(rootmost.path).getChild(util.Path.relativePath(rootmost.path, path));
    } else {
      return null;
    }
  }
  
  List<NamedNode> getCompleteChildren() {
    List<NamedNode> children = [];
    var node = this.writeTree_.value;
    if (node != null) {
      if (!node.isLeafNode) {
        node = (node);
        node.forEachChild(priorityIndex, (childName, childNode) {
          children.add(new NamedNode(childName, childNode));
        });
      }
    } else {
      this.writeTree_.children.forEach((childName, childTree) {
      if (childTree.value != null) {
        children.add(new NamedNode(childName, childTree.value));
      }
      });
    }
    return children;
  }
  
  CompoundWrite childCompoundWrite(util.Path path) {
    if (path.isEmpty) {
      return this;
    } else {
      var shadowingNode = this.getCompleteNode(path);
      if (shadowingNode != null) {
        return new CompoundWrite(new ImmutableTree(shadowingNode));
      } else {
        return new CompoundWrite(this.writeTree_.subtree(path));
      }
    }
  }
  
  bool get isEmpty {
    return this.writeTree_.isEmpty;
  }
  
  Node apply(Node node) {
    return CompoundWrite.applySubtreeWrite_(util.Path.empty, this.writeTree_, node);
  }
  
  static Node applySubtreeWrite_(util.Path relativePath,
      ImmutableTree<Node> writeTree, Node node) {
    if (writeTree.value != null) {
      return node.updateChild(relativePath, writeTree.value);
    } else {
      var priorityWrite = null;
      writeTree.children.forEach((childKey, childTree) {
        if (childKey == ".priority") {
          assert(childTree.value != null);//, "Priority writes must always be leaf nodes");
          priorityWrite = childTree.value;
        } else {
          node = CompoundWrite.applySubtreeWrite_(relativePath.child(childKey), childTree, node);
        }
      });
      if (!node.getChild(relativePath).isEmpty && priorityWrite != null) {
        node = node.updateChild(relativePath.child(".priority"), (priorityWrite));
      }
      return node;
    }
  }

}

class WriteRecord {

  final util.Path path;
  final Map<String,Node> children;
  final int writeId;
  final bool visible;
  final Node snap;

  WriteRecord({this.path, this.children, this.writeId, this.visible, this.snap});
}

class WriteTree {

  CompoundWrite visibleWrites_ = CompoundWrite.empty;
  final List<WriteRecord> allWrites_ = <WriteRecord>[]; // TODO replace with (linked?) map
  var lastWriteId_ = -1;


  WriteTreeRef childWrites(util.Path path) {
    return new WriteTreeRef(path, this);
  }

  void addOverwrite(util.Path path, Node snap, int writeId, [bool visible = true]) {
    assert(writeId >
        this.lastWriteId_); //, "Stacking an older write on top of newer ones");
    this.allWrites_.add(new WriteRecord(
        path:path, snap:snap, writeId:writeId, visible:visible));
    if (visible) {
      this.visibleWrites_ = this.visibleWrites_.addWrite(path, snap);
    }
    this.lastWriteId_ = writeId;
  }

  void addMerge(util.Path path, Map<String,Node> changedChildren, int writeId) {
    assert(writeId >
        this.lastWriteId_); //, "Stacking an older merge on top of newer ones");
    this.allWrites_.add(new WriteRecord(path:path, children:changedChildren,
        writeId:writeId, visible:true));
    this.visibleWrites_ = this.visibleWrites_.addWrites(path, changedChildren);
    this.lastWriteId_ = writeId;
  }

  WriteRecord getWrite(int writeId) {
    for (var i = 0; i < this.allWrites_.length; i++) {
      var record = this.allWrites_[i];
      if (record.writeId == writeId) {
        return record;
      }
    }
    return null;
  }

  bool removeWrite(int writeId) {
    var writeToRemove = allWrites_.firstWhere((s) => s.writeId==writeId);
    var idx = allWrites_.indexOf(writeToRemove);
    this.allWrites_.removeAt(idx);
    var removedWriteWasVisible = writeToRemove.visible;
    var removedWriteOverlapsWithOtherWrites = false;
    var i = this.allWrites_.length - 1;
    while (removedWriteWasVisible && i >= 0) {
      var currentWrite = this.allWrites_[i];
      if (currentWrite.visible) {
        if (i >= idx &&
            this.recordContainsPath_(currentWrite, writeToRemove.path)) {
          removedWriteWasVisible = false;
        } else {
          if (writeToRemove.path.contains(currentWrite.path)) {
            removedWriteOverlapsWithOtherWrites = true;
          }
        }
      }
      i--;
    }
    if (!removedWriteWasVisible) {
      return false;
    } else {
      if (removedWriteOverlapsWithOtherWrites) {
        this.resetTree_();
        return true;
      } else {
        if (writeToRemove.snap) {
          this.visibleWrites_ =
              this.visibleWrites_.removeWrite(writeToRemove.path);
        } else {
          var children = writeToRemove.children;
          var self = this;
          children.forEach((childName, childSnap) {
            self.visibleWrites_ = self.visibleWrites_.removeWrite(
                writeToRemove.path.child(childName));
          });
        }
        return true;
      }
    }
  }

  Node getCompleteWriteData(util.Path path) {
    return this.visibleWrites_.getCompleteNode(path);
  }

  Node calcCompleteEventCache(util.Path treePath, Node completeServerCache,
      List<int> writeIdsToExclude, bool includeHiddenWrites) {
    if (writeIdsToExclude==null && !includeHiddenWrites) {
      var shadowingNode = this.visibleWrites_.getCompleteNode(treePath);
      if (shadowingNode != null) {
        return shadowingNode;
      } else {
        var subMerge = this.visibleWrites_.childCompoundWrite(treePath);
        if (subMerge.isEmpty) {
          return completeServerCache;
        } else {
          if (completeServerCache == null &&
              !subMerge.hasCompleteWrite(util.Path.empty)) {
            return null;
          } else {
            var layeredCache = completeServerCache ?? EMPTY_NODE;
            return subMerge.apply(layeredCache);
          }
        }
      }
    } else {
      var merge = this.visibleWrites_.childCompoundWrite(treePath);
      if (!includeHiddenWrites && merge.isEmpty) {
        return completeServerCache;
      } else {
        if (!includeHiddenWrites && completeServerCache == null &&
            !merge.hasCompleteWrite(util.Path.empty)) {
          return null;
        } else {
          var filter =
              (WriteRecord write) {
            return (write.visible || includeHiddenWrites) &&
                (writeIdsToExclude==null || writeIdsToExclude.contains(write.writeId)) &&
                (write.path.contains(treePath) ||
                    treePath.contains(write.path));
          };
          var mergeAtPath = WriteTree.layerTree_(
              this.allWrites_, filter, treePath);
          var layeredCache = completeServerCache ?? EMPTY_NODE;
          return mergeAtPath.apply(layeredCache);
        }
      }
    }
  }

  Node calcCompleteEventChildren(util.Path treePath, [Node completeServerChildren]) {
    var completeChildren = EMPTY_NODE;
    var topLevelSet = this.visibleWrites_.getCompleteNode(treePath);
    if (topLevelSet) {
      if (!topLevelSet.isLeafNode) {
        topLevelSet.forEachChild(priorityIndex, (childName, childSnap) {
          completeChildren =
              completeChildren.updateImmediateChild(childName, childSnap);
        });
      }
      return completeChildren;
    } else {
      if (completeServerChildren!=null) {
        var merge = this.visibleWrites_.childCompoundWrite(treePath);
        completeServerChildren.forEachChild(
            priorityIndex, (childName, childNode) {
          var node = merge.childCompoundWrite(new util.Path.fromString(childName)).apply(
              childNode);
          completeChildren =
              completeChildren.updateImmediateChild(childName, node);
        });
        merge.getCompleteChildren().forEach((namedNode) {
          completeChildren = completeChildren.updateImmediateChild(
              namedNode.name, namedNode.node);
        });
        return completeChildren;
      } else {
        var merge = this.visibleWrites_.childCompoundWrite(treePath);
        merge.getCompleteChildren().forEach((namedNode) {
          completeChildren = completeChildren.updateImmediateChild(
              namedNode.name, namedNode.node);
        });
        return completeChildren;
      }
    }
  }

  Node calcEventCacheAfterServerOverwrite(util.Path treePath, util.Path childPath,
      Node existingEventSnap, Node existingServerSnap) {
    assert(existingEventSnap!=null ||
        existingServerSnap!=null); //"Either existingEventSnap or existingServerSnap must exist");
    var path = treePath.child(childPath);
    if (this.visibleWrites_.hasCompleteWrite(path)) {
      return null;
    } else {
      var childMerge = this.visibleWrites_.childCompoundWrite(path);
      if (childMerge.isEmpty) {
        return existingServerSnap.getChild(childPath);
      } else {
        return childMerge.apply(existingServerSnap.getChild(childPath));
      }
    }
  }

  Node calcCompleteChild(util.Path treePath, String childKey, CacheNode existingServerSnap) {
    var path = treePath.child(childKey);
    var shadowingNode = this.visibleWrites_.getCompleteNode(path);
    if (shadowingNode != null) {
      return shadowingNode;
    } else {
      if (existingServerSnap.isCompleteForChild(childKey)) {
        var childMerge = this.visibleWrites_.childCompoundWrite(path);
        return childMerge.apply(
            existingServerSnap.node.getImmediateChild(childKey));
      } else {
        return null;
      }
    }
  }

  Node shadowingWrite(util.Path path) {
    return this.visibleWrites_.getCompleteNode(path);
  }

  List<Node> calcIndexedSlice(util.Path treePath, Node completeServerData,
      String startPost, int count, bool reverse, Index index) {
    Node toIterate;
    var merge = this.visibleWrites_.childCompoundWrite(treePath);
    var shadowingNode = merge.getCompleteNode(util.Path.empty);
    if (shadowingNode != null) {
      toIterate = shadowingNode;
    } else {
      if (completeServerData != null) {
        toIterate = merge.apply(completeServerData);
      } else {
        return [];
      }
    }
    toIterate = toIterate.withIndex(index);
    if (!toIterate.isEmpty && !toIterate.isLeafNode) {
      List<Node> nodes = [];
      var cmp = index.compare;
      var iter = reverse
          ? toIterate.reverseIteratorFrom(startPost, index)
          : toIterate.iteratorFrom(startPost, index);
      var next = iter.getNext();
      while (next && nodes.length < count) {
        if (cmp(next, startPost) != 0) {
          nodes.add(next);
        }
        next = iter.getNext();
      }
      return nodes;
    } else {
      return [];
    }
  }

  bool recordContainsPath_(WriteRecord writeRecord, util.Path path) {
    if (writeRecord.snap!=null) {
      return writeRecord.path.contains(path);
    } else {
      return writeRecord.children.keys.any((childName) {
        return writeRecord.path.child(childName).contains(path);
      });
    }
  }

  void resetTree_() {
    this.visibleWrites_ = layerTree_(
        this.allWrites_, DefaultFilter_, util.Path.empty);
    if (this.allWrites_.length > 0) {
      this.lastWriteId_ = this.allWrites_[this.allWrites_.length - 1].writeId;
    } else {
      this.lastWriteId_ = -1;
    }
  }

  static bool DefaultFilter_(WriteRecord write) {
    return write.visible;
  }

  static layerTree_(List<WriteRecord> writes, bool filter(WriteRecord r),
      util.Path treeRoot) {
    var compoundWrite = CompoundWrite.empty;
    for (var i = 0; i < writes.length; ++i) {
      var write = writes[i];
      if (filter(write)) {
        var writePath = write.path;
        util.Path relativePath;
        if (write.snap) {
          if (treeRoot.contains(writePath)) {
            relativePath = util.Path.relativePath(treeRoot, writePath);
            compoundWrite = compoundWrite.addWrite(relativePath, write.snap);
          } else {
            if (writePath.contains(treeRoot)) {
              relativePath = util.Path.relativePath(writePath, treeRoot);
              compoundWrite = compoundWrite.addWrite(
                  util.Path.empty, write.snap.getChild(relativePath));
            } else {}
          }
        } else {
          if (write.children) {
            if (treeRoot.contains(writePath)) {
              relativePath = util.Path.relativePath(treeRoot, writePath);
              compoundWrite =
                  compoundWrite.addWrites(relativePath, write.children);
            } else {
              if (writePath.contains(treeRoot)) {
                relativePath = util.Path.relativePath(writePath, treeRoot);
                if (relativePath.isEmpty) {
                  compoundWrite =
                      compoundWrite.addWrites(util.Path.empty, write.children);
                } else {
                  var child = write.children[relativePath.front];
                  if (child) {
                    var deepNode = child.getChild(relativePath.popFront());
                    compoundWrite =
                        compoundWrite.addWrite(util.Path.empty, deepNode);
                  }
                }
              } else {}
            }
          } else {
            throw new Exception(
                "WriteRecord should have .snap or .children");
          }
        }
      }
    }
    return compoundWrite;
  }
}


class WriteTreeRef {
  final treePath_;
  final WriteTree writeTree_;

  WriteTreeRef(this.treePath_, this.writeTree_);

  Node calcCompleteEventCache(Node completeServerCache, [List<int> writeIdsToExclude = const[],
      bool includeHiddenWrites = false]) {
    return this.writeTree_.calcCompleteEventCache(
        this.treePath_, completeServerCache, writeIdsToExclude,
        includeHiddenWrites);
  }

  Node calcCompleteEventChildren(Node completeServerChildren) {
    return this.writeTree_.calcCompleteEventChildren(
        this.treePath_, completeServerChildren);
  }

  Node calcEventCacheAfterServerOverwrite(util.Path path, Node existingEventSnap,
      Node existingServerSnap) {
    return this.writeTree_.calcEventCacheAfterServerOverwrite(
        this.treePath_, path, existingEventSnap, existingServerSnap);
  }

  Node shadowingWrite(util.Path path) {
    return this.writeTree_.shadowingWrite(this.treePath_.child(path));
  }

  List<Node> calcIndexedSlice(Node completeServerData, String startPost,
      int count, bool reverse, Index index) {
    return this.writeTree_.calcIndexedSlice(
        this.treePath_, completeServerData, startPost, count, reverse, index);
  }

  Node calcCompleteChild(String childKey, CacheNode existingServerCache) {
    return this.writeTree_.calcCompleteChild(
        this.treePath_, childKey, existingServerCache);
  }

  WriteTreeRef child(String childName) {
    return new WriteTreeRef(
        this.treePath_.child(childName), this.writeTree_);
  }
}


class TreeNode {
  final Map<String,TreeNode> children = {};
  int childCount = 0;
  var value;
  
}

class Tree {
  final String name_;
  final Tree parent_;
  final TreeNode node_;

  Tree([this.name_ = "", this.parent_, TreeNode node]) : node_ = node ?? new TreeNode();

  Tree subTree(pathObj) {
    util.Path path = pathObj is util.Path ? pathObj : new util.Path.fromString(pathObj);
    var child = this, next;
    while ((next = path.front) != null) {
      var childNode = node_.children[next] ?? new TreeNode();
      child = new Tree(next, child, childNode);
      path = path.popFront();
    }
    return child;
  }

  getValue() {
    return this.node_.value;
  }

  void setValue(value) {
    this.node_.value = value;
    this.updateParents_();
  }

  void clear() {
    this.node_.value = null;
    this.node_.children.clear();
    this.node_.childCount = 0;
    this.updateParents_();
  }

  bool hasChildren() {
    return this.node_.childCount > 0;
  }

  bool get isEmpty {
    return this.getValue() == null && !this.hasChildren();
  }

  void forEachChild(void action(Tree tree)) {
    node_.children.forEach((child, childTree) {
      action(new Tree(child, this, childTree));
    });
  }

  void forEachDescendant(void action(Tree tree), [bool includeSelf = false, bool childrenFirst = false]) {
    if (includeSelf && !childrenFirst) {
      action(this);
    }
    this.forEachChild((child) {
      child.forEachDescendant(action, true, childrenFirst);
    });
    if (includeSelf && childrenFirst) {
      action(this);
    }
  }

  bool forEachAncestor(bool action(Tree tree), [bool includeSelf = false]) {
    Tree node = includeSelf ? this : this.parent();
    while (node != null) {
      if (action(node)) {
        return true;
      }
      node = node.parent();
    }
    return false;
  }

  void forEachImmediateDescendantWithValue(void action(Tree tree)) {
    this.forEachChild((child) {
      if (child.getValue() != null) {
        action(child);
      } else {
        child.forEachImmediateDescendantWithValue(action);
      }
    });
  }

  util.Path path() {
    return new util.Path.fromString(
        this.parent_ == null ? this.name_ : this.parent_.path() + "/" +
            this.name_);
  }

  String name() {
    return this.name_;
  }


  Tree parent() {
    return this.parent_;
  }


  void updateParents_() {
    if (this.parent_ != null) {
      this.parent_.updateChild_(this.name_, this);
    }
  }

  void updateChild_(String childName, Tree child) {
    var childEmpty = child.isEmpty;
    var childExists = node_.children.containsKey(childName);
    if (childEmpty && childExists) {
      this.node_.children.remove(childName);
      this.node_.childCount--;
      this.updateParents_();
    } else {
      if (!childEmpty && !childExists) {
        this.node_.children[childName] = child.node_;
        this.node_.childCount++;
        this.updateParents_();
      }
    }
  }
}