
part of firebase_dart.core.view;

class IndexedFilter extends NodeFilter {
  final Index index;


  IndexedFilter(this.index);

  updateChild(Node snap, String key, Node newChild, util.Path affectedPath,
      CompleteChildSource source, [ChildChangeAccumulator changeAccumulator]) {
    assert(snap.isIndexed(index));//, "A node must be indexed if only a child is updated");
    var oldChild = snap.getImmediateChild(key);
    if (oldChild.getChild(affectedPath)==(newChild.getChild(affectedPath))) {
      if (oldChild.isEmpty == newChild.isEmpty) {
        return snap;
      }
    }
    if (changeAccumulator != null) {
      if (newChild.isEmpty) {
        if (snap.hasChild(key)) {
          changeAccumulator.trackChildChange(new Change.childRemoved(key, oldChild));
        } else {
          assert(snap.isLeafNode);//, "A child remove without an old child only makes sense on a leaf node");
        }
      } else {
        if (oldChild.isEmpty) {
          changeAccumulator.trackChildChange(new Change.childAdded(key, newChild));
        } else {
          changeAccumulator.trackChildChange(new Change.childChanged(key, newChild, oldChild));
        }
      }
    }
    if (snap.isLeafNode && newChild.isEmpty) {
      return snap;
    } else {
      return snap.updateImmediateChild(key, newChild).withIndex(index);
    }
  }

  updateFullNode(Node oldSnap, Node newSnap,
      [ChildChangeAccumulator changeAccumulator]) {
    if (changeAccumulator != null) {
      if (!oldSnap.isLeafNode) {
        oldSnap.forEachChild(priorityIndex, (key, childNode) {
          if (!newSnap.hasChild(key)) {
            changeAccumulator.trackChildChange(new Change.childRemoved(key, childNode));
          }
        });
      }
      if (!newSnap.isLeafNode) {
        newSnap.forEachChild(priorityIndex, (key, childNode) {
          if (oldSnap.hasChild(key)) {
            var oldChild = oldSnap.getImmediateChild(key);
            if (oldChild!=(childNode)) {
              changeAccumulator.trackChildChange(new Change.childChanged(key, childNode, oldChild));
            }
          } else {
            changeAccumulator.trackChildChange(new Change.childAdded(key, childNode));
          }
        });
      }
    }
    return newSnap.withIndex(this.index);
  }

  Node updatePriority(Node oldSnap, Node newPriority) {
    if (oldSnap.isEmpty) {
      return EMPTY_NODE;
    } else {
      return oldSnap.updatePriority(newPriority);
    }
  }

  bool get filtersNodes => false;

  IndexedFilter get indexedFilter => this;


}

class RangedFilter extends NodeFilter {

  final IndexedFilter indexedFilter;
  final NamedNode startPost;
  final NamedNode endPost;

  RangedFilter(QueryParams params) : indexedFilter = new IndexedFilter(params.index),
    startPost = _getStartPost(params), endPost = _getEndPost(params);

  bool matches(NamedNode node) =>
      index.compare(startPost, node) <= 0 && index.compare(node, endPost) <= 0;


  updateChild(Node snap, String key, Node newChild, util.Path affectedPath,
      CompleteChildSource source, [ChildChangeAccumulator changeAccumulator]) {
    if (!this.matches(new NamedNode(key, newChild))) {
      newChild = EMPTY_NODE;
    }
    return indexedFilter.updateChild(snap, key, newChild, affectedPath, source, changeAccumulator);
  }


  updateFullNode(Node oldSnap, Node newSnap,
      [ChildChangeAccumulator changeAccumulator]) {
    if (newSnap.isLeafNode) {
      newSnap = EMPTY_NODE;
    }
    var filtered = newSnap.withIndex(index);
    filtered = filtered.updatePriority(EMPTY_NODE);
    var self = this;
    newSnap.forEachChild(priorityIndex, (key, childNode) {
      if (!self.matches(new NamedNode(key, childNode))) {
        filtered = filtered.updateImmediateChild(key, EMPTY_NODE);
      }
    });
    return indexedFilter.updateFullNode(oldSnap, filtered, changeAccumulator);
  }

  updatePriority(Node oldSnap, Node newPriority) => oldSnap;

  bool get filtersNodes => true;

  static _getStartPost(QueryParams params) {
    if (params.hasStart) {
      var startName = params.indexStartName;
      return params.index.makePost(params.indexStartValue, startName);
    } else {
      return params.index.minPost;
    }
  }

  static _getEndPost(QueryParams params) {
    if (params.hasEnd) {
      var endName = params.indexEndName;
      return params.index.makePost(params.indexEndValue, endName);
    } else {
      return params.index.maxPost;
    }
  }

}

abstract class NodeFilter {

  updateChild(Node snap, String key, Node newChild, util.Path affectedPath,
      CompleteChildSource source, [ChildChangeAccumulator changeAccumulator]);

  updateFullNode(Node oldSnap, Node newSnap,
      [ChildChangeAccumulator changeAccumulator]);

  updatePriority(Node oldSnap, Node newPriority);

  bool get filtersNodes;

  IndexedFilter get indexedFilter;

  Index get index => indexedFilter.index;

}

class LimitedFilter extends NodeFilter {

  final RangedFilter rangedFilter;

  final int limit;
  final bool reverse;

  IndexedFilter get indexedFilter => rangedFilter.indexedFilter;

  LimitedFilter(QueryParams params) : rangedFilter = new RangedFilter(params),
    limit = params.limit, reverse = !params.isViewFromLeft;

  updateChild(Node snap, String key, Node newChild, util.Path affectedPath,
      CompleteChildSource source, [ChildChangeAccumulator changeAccumulator]) {
    if (!rangedFilter.matches(new NamedNode(key, newChild))) {
      newChild = EMPTY_NODE;
    }
    if (snap.getImmediateChild(key)==(newChild)) {
      return snap;
    } else {
      if (snap.numChildren < this.limit) {
        return this.indexedFilter.updateChild(snap, key, newChild, affectedPath, source, changeAccumulator);
      } else {
        return this._fullLimitUpdateChild(snap, key, newChild, source, changeAccumulator);
      }
    }
  }

  updateFullNode(Node oldSnap, Node newSnap,
      [ChildChangeAccumulator changeAccumulator]) {
    var filtered;
    if (newSnap.isLeafNode || newSnap.isEmpty) {
      filtered = EMPTY_NODE.withIndex(index);
    } else {
      if (this.limit * 2 < newSnap.numChildren && newSnap.isIndexed(this.index)) {
        filtered = EMPTY_NODE.withIndex(index);
        var iterator;
        newSnap = (newSnap);
        if (this.reverse) {
          iterator = newSnap.reverseIteratorFrom(this.rangedFilter.endPost, index);
        } else {
          iterator = newSnap.iteratorFrom(rangedFilter.startPost, index);
        }
        var count = 0;
        while (iterator.hasNext() && count < this.limit) {
          var next = iterator.getNext();
          var inRange;
          if (this.reverse) {
            inRange = index.compare(this.rangedFilter.startPost, next) <= 0;
          } else {
            inRange = index.compare(next, this.rangedFilter.endPost) <= 0;
          }
          if (inRange) {
            filtered = filtered.updateImmediateChild(next.name, next.node);
            count++;
          } else {
            break;
          }
        }
      } else {
        filtered = newSnap.withIndex(this.index);
        filtered = (filtered.updatePriority(EMPTY_NODE));
        var startPost;
        var endPost;
        var cmp;
        var iterator;
        if (this.reverse) {
          iterator = filtered.getReverseIterator(this.index);
          startPost = this.rangedFilter.endPost;
          endPost = this.rangedFilter.startPost;
          var indexCompare = this.index.compare;
          cmp = (a, b) {
            return indexCompare(b, a);
          };
        } else {
          iterator = filtered.getIterator(this.index);
          startPost = this.rangedFilter.startPost;
          endPost = this.rangedFilter.endPost;
          cmp = this.index.compare;
        }
        var count = 0;
        var foundStartPost = false;
        var next;
        while (iterator.hasNext()) {
          next = iterator.getNext();
          if (!foundStartPost && cmp(startPost, next) <= 0) {
            foundStartPost = true;
          }
          var inRange = foundStartPost && count < this.limit && cmp(next, endPost) <= 0;
          if (inRange) {
            count++;
          } else {
            filtered = filtered.updateImmediateChild(next.name, EMPTY_NODE);
          }
        }
      }
    }
    return indexedFilter.updateFullNode(oldSnap, filtered, changeAccumulator);
  }

  updatePriority(Node oldSnap, Node newPriority) => oldSnap;

  bool get filtersNodes => true;

  _fullLimitUpdateChild(Node snap, String childKey, Node childSnap,
      CompleteChildSource source, ChildChangeAccumulator changeAccumulator) {
    var cmp;
    if (this.reverse) {
      var indexCmp = this.index.compare;
      cmp = (a, b) {
        return indexCmp(b, a);
      };
    } else {
      cmp = this.index.compare;
    }
    var oldEventCache = (snap);
    assert(oldEventCache.numChildren == this.limit);
    var newChildNamedNode = new NamedNode(childKey, childSnap);
    var windowBoundary = (this.reverse ? oldEventCache.getFirstChild(this.index) :
    oldEventCache.getLastChild(this.index));
    var inRange = this.rangedFilter.matches(newChildNamedNode);
    if (oldEventCache.hasChild(childKey)) {
      var oldChildSnap = oldEventCache.getImmediateChild(childKey);
      var nextChild = source.getChildAfterChild(this.index, windowBoundary, this.reverse);
      while (nextChild != null && (nextChild.name == childKey || oldEventCache.hasChild(nextChild.name))) {
        nextChild = source.getChildAfterChild(this.index, nextChild, this.reverse);
      }
      var compareNext = nextChild == null ? 1 : cmp(nextChild, newChildNamedNode);
      var remainsInWindow = inRange && !childSnap.isEmpty && compareNext >= 0;
      if (remainsInWindow) {
        if (changeAccumulator != null) {
          changeAccumulator.trackChildChange(new Change.childChanged(childKey, childSnap, oldChildSnap));
        }
        return oldEventCache.updateImmediateChild(childKey, childSnap);
      } else {
        if (changeAccumulator != null) {
          changeAccumulator.trackChildChange(new Change.childRemoved(childKey, oldChildSnap));
        }
        var newEventCache = oldEventCache.updateImmediateChild(childKey, EMPTY_NODE);
        var nextChildInRange = nextChild != null && this.rangedFilter.matches(nextChild);
        if (nextChildInRange) {
          if (changeAccumulator != null) {
            changeAccumulator.trackChildChange(new Change.childAdded(nextChild.name, nextChild.node));
          }
          return newEventCache.updateImmediateChild(nextChild.name, nextChild.node);
        } else {
          return newEventCache;
        }
      }
    } else {
      if (childSnap.isEmpty) {
        return snap;
      } else {
        if (inRange) {
          if (cmp(windowBoundary, newChildNamedNode) >= 0) {
            if (changeAccumulator != null) {
              changeAccumulator.trackChildChange(new Change.childRemoved(windowBoundary.name, windowBoundary.node));
              changeAccumulator.trackChildChange(new Change.childAdded(childKey, childSnap));
            }
            return oldEventCache.updateImmediateChild(childKey, childSnap)
                .updateImmediateChild(windowBoundary.name, EMPTY_NODE);
          } else {
            return snap;
          }
        } else {
          return snap;
        }
      }
    }
  }

}

