part of firebase_dart.core.view;

class ProcessorResult {
  final ViewCache viewCache;
  final changes;

  ProcessorResult(this.viewCache, this.changes);
}

class ViewProcessor {
  final NodeFilter filter;

  const ViewProcessor(this.filter);

  void assertIndexed(ViewCache viewCache) {
    assert(viewCache.eventCache.node.isIndexed(this.filter.index));//, "Event snap not indexed");
    assert(viewCache.serverCache.node.isIndexed(this.filter.index));//, "Server snap not indexed");
  }

  ProcessorResult applyOperation(ViewCache oldViewCache, Operation operation,
      WriteTreeRef writesCache, [Node completeCache]) {
    var accumulator = new ChildChangeAccumulator();
    var newViewCache, filterServerNode;
    if (operation is Overwrite) {
      var overwrite = (operation);
      if (overwrite.source.fromUser) {
        newViewCache = this._applyUserOverwrite(oldViewCache, overwrite.path,
            overwrite.snap, writesCache, completeCache, accumulator);
      } else {
        assert(overwrite.source.fromServer);//, "Unknown source.");
        filterServerNode = overwrite.source.tagged
            || oldViewCache.serverCache.isFiltered && !overwrite.path.isEmpty;
        newViewCache = this._applyServerOverwrite(oldViewCache, overwrite.path,
            overwrite.snap, writesCache, completeCache, filterServerNode,
            accumulator);
      }
    } else {
      if (operation is Merge) {
        var merge = (operation);
        if (merge.source.fromUser) {
          newViewCache = this._applyUserMerge(oldViewCache, merge.path,
              merge.children, writesCache, completeCache, accumulator);
        } else {
          assert(merge.source.fromServer);//, "Unknown source.");
          filterServerNode = merge.source.tagged || oldViewCache.serverCache.isFiltered;
          newViewCache = this._applyServerMerge(oldViewCache, merge.path,
              merge.children, writesCache, completeCache, filterServerNode, accumulator);
        }
      } else {
        if (operation is AckUserWrite) {
          var ackUserWrite = (operation);
          if (!ackUserWrite.revert) {
            newViewCache = this._ackUserWrite(oldViewCache, ackUserWrite.path,
                ackUserWrite.affectedTree, writesCache, completeCache, accumulator);
          } else {
            newViewCache = this._revertUserWrite(oldViewCache, ackUserWrite.path,
                writesCache, completeCache, accumulator);
          }
        } else {
          if (operation.type == OperationType.LISTEN_COMPLETE) {
            newViewCache = this._listenComplete(oldViewCache, operation.path,
                writesCache, completeCache, accumulator);
          } else {
            throw new Exception("Unknown operation type: $operation");
          }
        }
      }
    }
    var changes = accumulator.changes.toList();
    this._maybeAddValueEvent(oldViewCache, newViewCache, changes);
    return new ProcessorResult(newViewCache, changes);
  }

  void _maybeAddValueEvent(ViewCache oldViewCache, ViewCache newViewCache, List<Change> accumulator) {
    var eventSnap = newViewCache.eventCache;
    if (eventSnap.isFullyInitialized) {
      var isLeafOrEmpty = eventSnap.node.isLeafNode || eventSnap.node.isEmpty;
      var oldCompleteSnap = oldViewCache.completeEventSnap;
      if (accumulator.length > 0 || !oldViewCache.eventCache.isFullyInitialized
          || isLeafOrEmpty && eventSnap.node!=((oldCompleteSnap))
          || eventSnap.node.priority!=(oldCompleteSnap.priority)) {
        accumulator.add(new Change.value((newViewCache.completeEventSnap)));
      }
    }
  }

  ViewCache _generateEventCacheAfterServerEvent(ViewCache viewCache, util.Path changePath,
      WriteTreeRef writesCache, CompleteChildSource source, ChildChangeAccumulator accumulator) {
    var oldEventSnap = viewCache.eventCache;
    if (writesCache.shadowingWrite(changePath) != null) {
      return viewCache;
    } else {
      var newEventCache, serverNode;
      if (changePath.isEmpty) {
        assert(viewCache.serverCache.isFullyInitialized);//, "If change path is empty, we must have complete server data");
        if (viewCache.serverCache.isFiltered) {
          var serverCache = viewCache.completeServerSnap;
          var completeChildren = serverCache is ChildrenNode ? serverCache : EMPTY_NODE;
          var completeEventChildren = writesCache.calcCompleteEventChildren(completeChildren);
          newEventCache = this.filter.updateFullNode(viewCache.eventCache.node, completeEventChildren, accumulator);
        } else {
          var completeNode = (writesCache.calcCompleteEventCache(viewCache.completeServerSnap));
          newEventCache = this.filter.updateFullNode(viewCache.eventCache.node, completeNode, accumulator);
        }
      } else {
        var childKey = changePath.front;
        if (childKey == ".priority") {
          assert(changePath.length == 1);//, "Can't have a priority with additional path components");
          var oldEventNode = oldEventSnap.node;
          serverNode = viewCache.serverCache.node;
          var updatedPriority = writesCache.calcEventCacheAfterServerOverwrite(changePath, oldEventNode, serverNode);
          if (updatedPriority != null) {
            newEventCache = this.filter.updatePriority(oldEventNode, updatedPriority);
          } else {
            newEventCache = oldEventSnap.node;
          }
        } else {
          var childChangePath = changePath.popFront();
          var newEventChild;
          if (oldEventSnap.isCompleteForChild(childKey)) {
            serverNode = viewCache.serverCache.node;
            var eventChildUpdate = writesCache.calcEventCacheAfterServerOverwrite(
                changePath, oldEventSnap.node, serverNode);
            if (eventChildUpdate != null) {
              newEventChild = oldEventSnap.node.getImmediateChild(childKey)
                  .updateChild(childChangePath, eventChildUpdate);
            } else {
              newEventChild = oldEventSnap.node.getImmediateChild(childKey);
            }
          } else {
            newEventChild = writesCache.calcCompleteChild(childKey, viewCache.serverCache);
          }
          if (newEventChild != null) {
            newEventCache = this.filter.updateChild(oldEventSnap.node, childKey,
                newEventChild, childChangePath, source, accumulator);
          } else {
            newEventCache = oldEventSnap.node;
          }
        }
      }
      return viewCache.updateEventSnap(newEventCache,
          oldEventSnap.isFullyInitialized || changePath.isEmpty,
          this.filter.filtersNodes);
    }
  }

  ViewCache _applyServerOverwrite(ViewCache oldViewCache, util.Path changePath,
      Node changedSnap, WriteTreeRef writesCache, Node completeCache,
      bool filterServerNode, ChildChangeAccumulator accumulator) {
    var oldServerSnap = oldViewCache.serverCache;
    var newServerCache;
    var serverFilter = filterServerNode ? this.filter : this.filter.indexedFilter;
    if (changePath.isEmpty) {
      newServerCache = serverFilter.updateFullNode(oldServerSnap.node, changedSnap, null);
    } else {
      if (serverFilter.filtersNodes && !oldServerSnap.isFiltered) {
        var newServerNode = oldServerSnap.node.updateChild(changePath, changedSnap);
        newServerCache = serverFilter.updateFullNode(oldServerSnap.node, newServerNode, null);
      } else {
        var childKey = changePath.front;
        if (!oldServerSnap.isCompleteForPath(changePath) && changePath.length > 1) {
          return oldViewCache;
        }
        var childChangePath = changePath.popFront();
        var childNode = oldServerSnap.node.getImmediateChild(childKey);
        var newChildNode = childNode.updateChild(childChangePath, changedSnap);
        if (childKey == ".priority") {
          newServerCache = serverFilter.updatePriority(oldServerSnap.node, newChildNode);
        } else {
          newServerCache = serverFilter.updateChild(oldServerSnap.node, childKey,
              newChildNode, childChangePath, NO_COMPLETE_CHILD_SOURCE, null);
        }
      }
    }
    var newViewCache = oldViewCache.updateServerSnap(newServerCache,
        oldServerSnap.isFullyInitialized || changePath.isEmpty, serverFilter.filtersNodes);
    var source = new WriteTreeCompleteChildSource(writesCache, newViewCache, completeCache);
    return this._generateEventCacheAfterServerEvent(newViewCache, changePath, writesCache, source, accumulator);
  }

  ViewCache _applyUserOverwrite(ViewCache oldViewCache, util.Path changePath,
      Node changedSnap, WriteTreeRef writesCache, Node completeCache,
      ChildChangeAccumulator accumulator) {
    var oldEventSnap = oldViewCache.eventCache;
    var newViewCache, newEventCache;
    var source = new WriteTreeCompleteChildSource(writesCache, oldViewCache, completeCache);
    if (changePath.isEmpty) {
      newEventCache = this.filter.updateFullNode(oldViewCache.eventCache.node, changedSnap, accumulator);
      newViewCache = oldViewCache.updateEventSnap(newEventCache, true, this.filter.filtersNodes);
    } else {
      var childKey = changePath.front;
      if (childKey == ".priority") {
        newEventCache = this.filter.updatePriority(oldViewCache.eventCache.node, changedSnap);
        newViewCache = oldViewCache.updateEventSnap(newEventCache,
            oldEventSnap.isFullyInitialized, oldEventSnap.isFiltered);
      } else {
        var childChangePath = changePath.popFront();
        var oldChild = oldEventSnap.node.getImmediateChild(childKey);
        var newChild;
        if (childChangePath.isEmpty) {
          newChild = changedSnap;
        } else {
          var childNode = source.getCompleteChild(childKey);
          if (childNode != null) {
            if (childChangePath.back == ".priority"
                && childNode.getChild((childChangePath.parent)).isEmpty) {
              newChild = childNode;
            } else {
              newChild = childNode.updateChild(childChangePath, changedSnap);
            }
          } else {
            newChild = EMPTY_NODE;
          }
        }
        if (oldChild!=(newChild)) {
          var newEventSnap = this.filter.updateChild(oldEventSnap.node,
              childKey, newChild, childChangePath, source, accumulator);
          newViewCache = oldViewCache.updateEventSnap(newEventSnap,
              oldEventSnap.isFullyInitialized, this.filter.filtersNodes);
        } else {
          newViewCache = oldViewCache;
        }
      }
    }
    return newViewCache;
  }

  static _cacheHasChild(ViewCache viewCache, String childKey) {
    return viewCache.eventCache.isCompleteForChild(childKey);
  }

  ViewCache _applyUserMerge(ViewCache viewCache, util.Path path,
      ImmutableTree<Node> changedChildren, WriteTreeRef writesCache, Node serverCache, ChildChangeAccumulator accumulator) {
    var curViewCache = viewCache;
    changedChildren.forEach((relativePath, childNode) {
      var writePath = path.child(relativePath);
      if (_cacheHasChild(viewCache, writePath.front)) {
        curViewCache = _applyUserOverwrite(curViewCache, writePath, childNode, writesCache, serverCache, accumulator);
      }
    });
    changedChildren.forEach((relativePath, childNode) {
      var writePath = path.child(relativePath);
      if (!_cacheHasChild(viewCache, writePath.front)) {
        curViewCache = _applyUserOverwrite(curViewCache, writePath, childNode, writesCache, serverCache, accumulator);
      }
    });
    return curViewCache;
  }

  Node _applyMerge(Node node, Map<util.Path, Node> merge) {
    merge.forEach((relativePath, childNode) {
      node = node.updateChild(relativePath, childNode);
    });
    return node;
  }

  ViewCache _applyServerMerge(ViewCache viewCache, util.Path path, changedChildren,
      WriteTreeRef writesCache, Node serverCache, bool filterServerNode, ChildChangeAccumulator accumulator) {
    if (viewCache.serverCache.node.isEmpty && !viewCache.serverCache.isFullyInitialized) {
      return viewCache;
    }
    var curViewCache = viewCache;
    var viewMergeTree;
    if (path.isEmpty) {
      viewMergeTree = changedChildren;
    } else {
      viewMergeTree = ImmutableTree.empty.setTree(path, changedChildren);
    }
    var serverNode = viewCache.serverCache.node;
    var self = this;
    viewMergeTree.children.inorderTraversal((childKey, childTree) {
      if (serverNode.hasChild(childKey)) {
        var serverChild = viewCache.serverCache.node.getImmediateChild(childKey);
        var newChild = self._applyMerge(serverChild, childTree);
        curViewCache = self._applyServerOverwrite(curViewCache,
            new util.Path.fromString(childKey), newChild, writesCache, serverCache, filterServerNode, accumulator);
      }
    });
    viewMergeTree.children.inorderTraversal((childKey, childMergeTree) {
      var isUnknownDeepMerge = !viewCache.serverCache.isCompleteForChild(childKey) && childMergeTree.value == null;
      if (!serverNode.hasChild(childKey) && !isUnknownDeepMerge) {
        var serverChild = viewCache.serverCache.node.getImmediateChild(childKey);
        var newChild = self._applyMerge(serverChild, childMergeTree);
        curViewCache = self._applyServerOverwrite(curViewCache, new util.Path.fromString(childKey), newChild, writesCache, serverCache, filterServerNode, accumulator);
      }
    });
    return curViewCache;
  }

  ViewCache _ackUserWrite(ViewCache viewCache, util.Path ackPath, ImmutableTree<Node> affectedTree,
      WriteTreeRef writesCache, Node completeCache, ChildChangeAccumulator accumulator) {
    if (writesCache.shadowingWrite(ackPath) != null) {
      return viewCache;
    }
    var filterServerNode = viewCache.serverCache.isFiltered;
    var serverCache = viewCache.serverCache;
    if (affectedTree.value != null) {
      if (ackPath.isEmpty && serverCache.isFullyInitialized || serverCache.isCompleteForPath(ackPath)) {
        return this._applyServerOverwrite(viewCache, ackPath,
            serverCache.node.getChild(ackPath), writesCache, completeCache, filterServerNode, accumulator);
      } else {
        if (ackPath.isEmpty) {
          var changedChildren = ImmutableTree.empty;
          serverCache.node.forEachChild(keyIndex, (name, node) {
            changedChildren = changedChildren.set(new util.Path.fromString(name), node);
          });
          return this._applyServerMerge(viewCache, ackPath, changedChildren,
              writesCache, completeCache, filterServerNode, accumulator);
        } else {
          return viewCache;
        }
      }
    } else {
      var changedChildren = ImmutableTree.empty;
      affectedTree.forEach((mergePath, value) {
        var serverCachePath = ackPath.child(mergePath);
        if (serverCache.isCompleteForPath(serverCachePath)) {
          changedChildren = changedChildren.set(mergePath, serverCache.node.getChild(serverCachePath));
        }
      });
      return this._applyServerMerge(viewCache, ackPath, changedChildren, writesCache, completeCache, filterServerNode, accumulator);
    }
  }

  ViewCache _revertUserWrite(ViewCache viewCache, util.Path path, WriteTreeRef writesCache,
      Node completeServerCache, ChildChangeAccumulator accumulator) {
    var complete;
    if (writesCache.shadowingWrite(path) != null) {
      return viewCache;
    } else {
      var source = new WriteTreeCompleteChildSource(writesCache, viewCache, completeServerCache);
      var oldEventCache = viewCache.eventCache.node;
      var newEventCache;
      if (path.isEmpty || path.front == ".priority") {
        var newNode;
        if (viewCache.serverCache.isFullyInitialized) {
          newNode = writesCache.calcCompleteEventCache(viewCache.completeServerSnap);
        } else {
          var serverChildren = viewCache.serverCache.node;
          assert(serverChildren is ChildrenNode);//, "serverChildren would be complete if leaf node");
          newNode = writesCache.calcCompleteEventChildren((serverChildren));
        }
        newNode = (newNode);
        newEventCache = this.filter.updateFullNode(oldEventCache, newNode, accumulator);
      } else {
        var childKey = path.front;
        var newChild = writesCache.calcCompleteChild(childKey, viewCache.serverCache);
        if (newChild == null && viewCache.serverCache.isCompleteForChild(childKey)) {
          newChild = oldEventCache.getImmediateChild(childKey);
        }
        if (newChild != null) {
          newEventCache = this.filter.updateChild(oldEventCache, childKey,
              newChild, path.popFront(), source, accumulator);
        } else {
          if (viewCache.eventCache.node.hasChild(childKey)) {
            newEventCache = this.filter.updateChild(oldEventCache, childKey,
                EMPTY_NODE, path.popFront(), source, accumulator);
          } else {
            newEventCache = oldEventCache;
          }
        }
        if (newEventCache.isEmpty() && viewCache.serverCache.isFullyInitialized) {
          complete = writesCache.calcCompleteEventCache(viewCache.completeServerSnap);
          if (complete.isLeafNode) {
            newEventCache = this.filter.updateFullNode(newEventCache, complete, accumulator);
          }
        }
      }
      complete = viewCache.serverCache.isFullyInitialized || writesCache.shadowingWrite(util.Path.empty) != null;
      return viewCache.updateEventSnap(newEventCache, complete, this.filter.filtersNodes);
    }
  }

  ViewCache _listenComplete(ViewCache viewCache, util.Path path, WriteTreeRef writesCache,
      Node serverCache, ChildChangeAccumulator accumulator) {
    var oldServerNode = viewCache.serverCache;
    var newViewCache = viewCache.updateServerSnap(oldServerNode.node,
        oldServerNode.isFullyInitialized || path.isEmpty, oldServerNode.isFiltered);
    return this._generateEventCacheAfterServerEvent(newViewCache, path, writesCache,
        NO_COMPLETE_CHILD_SOURCE, accumulator);
  }

}

