part of firebase_dart.core.view;

class View {

  final Query query;
  ViewProcessor _processor;
  ViewCache _viewCache;
  List<EventRegistration> _eventRegistrations = [];
  EventGenerator _eventGenerator;

  View(this.query, ViewCache initialViewCache) {
    var params = query.queryParams;
    var indexFilter = new IndexedFilter(params.comparator);
    var filter = params.nodeFilter;
    this._processor = new ViewProcessor(filter);
    var initialServerCache = initialViewCache.serverCache;
    var initialEventCache = initialViewCache.eventCache;
    var serverSnap = indexFilter.updateFullNode(EMPTY_NODE, initialServerCache.node, null);
    var eventSnap = filter.updateFullCollection(EMPTY_NODE, initialEventCache.node, null);
    var newServerCache = new CacheNode(serverSnap, initialServerCache.isFullyInitialized, indexFilter.filtersNodes);
    var newEventCache = new CacheNode(eventSnap, initialEventCache.isFullyInitialized, filter.filtersNodes());
    this._viewCache = new ViewCache(newEventCache, newServerCache);
    this._eventGenerator = new EventGenerator(query);
  }

  Node get serverCache => _viewCache.serverCache.node;

  Node getCompleteServerCache(util.Path path) {
    var cache = _viewCache.completeServerSnap;
    if (cache!=null) {
      if (this.query.queryParams.loadsAllData || !path.isEmpty
          && !cache.getImmediateChild(path.front).isEmpty) {
        return cache.getChild(path);
      }
    }
    return null;
  }

  bool get isEmpty => _eventRegistrations.isEmpty;

  void addEventRegistration(EventRegistration eventRegistration) {
    _eventRegistrations.add(eventRegistration);
  }

  List<CancelEvent> removeEventRegistration(EventRegistration eventRegistration, Exception cancelError) {
    var cancelEvents = [];
    if (cancelError!=null) {
      assert(eventRegistration == null);//, "A cancel should cancel all event registrations.");
      var path = this.query.path;
      _eventRegistrations.forEach((registration) {
        var maybeEvent = registration.createCancelEvent(cancelError, path);
        if (maybeEvent!=null) {
          cancelEvents.add(maybeEvent);
        }
      });
    }
    if (eventRegistration!=null) {
      var remaining = [];
      for (var i = 0;i < this._eventRegistrations.length;++i) {
        var existing = this._eventRegistrations[i];
        if (!existing.matches(eventRegistration)) {
          remaining.add(existing);
        } else {
          if (eventRegistration.hasAnyCallback) {
            remaining.addAll(this._eventRegistrations.sublist(i + 1));
            break;
          }
        }
      }
      this._eventRegistrations = remaining;
    } else {
      this._eventRegistrations = [];
    }
    return cancelEvents;
  }

  List<Event> applyOperation(Operation operation, WriteTreeRef writesCache, [Node completeServerCache]) {
    if (operation.type == OperationType.MERGE && operation.source.queryId != null) {
      assert(this._viewCache.completeServerSnap!=null);//, "We should always have a full cache before handling merges");
      assert(this._viewCache.completeEventSnap!=null);//, "Missing event cache, even though we have a server cache");
    }
    var oldViewCache = this._viewCache;
    var result = this._processor.applyOperation(oldViewCache, operation, writesCache, completeServerCache);
    this._processor.assertIndexed(result.viewCache);
    assert(result.viewCache.serverCache.isFullyInitialized || !oldViewCache.serverCache.isFullyInitialized);
    // , "Once a server snap is complete, it should never go back");
    this._viewCache = result.viewCache;
    return this._generateEventsForChanges(result.changes, result.viewCache.eventCache.node, null);
  }

  List<Event> getInitialEvents(EventRegistration registration) {
    var eventSnap = this._viewCache.eventCache;
    var initialChanges = [];
    if (!eventSnap.node.isLeafNode) {
      var eventNode = (eventSnap.node);
      eventNode.forEachChild(priorityIndex, (key, childNode) {
        initialChanges.add(new Change.childAdded(key, childNode));
      });
    }
    if (eventSnap.isFullyInitialized) {
      initialChanges.add(new Change.value(eventSnap.node));
    }
    return this._generateEventsForChanges(initialChanges, eventSnap.node, registration);
  }

  _generateEventsForChanges(List<Change> changes, Node eventCache, [EventRegistration eventRegistration]) {
    var registrations = eventRegistration!=null ? [eventRegistration] : this._eventRegistrations;
    return _eventGenerator.generateEventsForChanges(changes, eventCache, registrations);
  }

}

