
import 'dart:math' as math;
import 'core/repo.dart';

abstract class Transport {

  Transport(connId, RepoInfo repoInfo, sessionId);

  open(onMessage, onDisconnect);

  start();

  close();

  send(data);

  int bytesReceived;
  int bytesSent;
}

class Constants {
  static const PROTOCOL_VERSION = "5";
  static const VERSION_PARAM = "v";
  static const TRANSPORT_SESSION_PARAM = "s";
  static const REFERER_PARAM = "r";
  static const FORGE_REF = "f";
  static const FORGE_DOMAIN = "firebaseio.com";
  static const LAST_SESSION_PARAM = "ls";
  static const WEBSOCKET = "websocket";
  static const LONG_POLLING = "long_polling";

}

class PacketReceiver {

  final onMessage_;
  final pendingResponses = [];
  var currentResponseNum = 0;
  var closeAfterResponse = -1;
  var onClose = null;

  PacketReceiver(this.onMessage_);

  closeAfter(responseNum, callback) {
    this.closeAfterResponse = responseNum;
    this.onClose = callback;
    if (this.closeAfterResponse < this.currentResponseNum) {
      this.onClose();
      this.onClose = null;
    }
  }

  handleResponse(requestNum, data) {
    this.pendingResponses[requestNum] = data;
    while (this.pendingResponses[this.currentResponseNum]) {
      var toProcess = this.pendingResponses[this.currentResponseNum];
      this.pendingResponses.remove(this.currentResponseNum);
      for (var i = 0;i < toProcess.length;++i) {
        if (toProcess[i]) {
          var self = this;
          fb.core.util.exceptionGuard(() {
            self.onMessage_(toProcess[i]);
          });
        }
      }
      if (this.currentResponseNum == this.closeAfterResponse) {
        if (this.onClose) {
          clearTimeout(this.onClose);
          this.onClose();
          this.onClose = null;
        }
        break;
      }
      this.currentResponseNum++;
    }
  }

}

const FIREBASE_LONGPOLL_START_PARAM = "start";
const FIREBASE_LONGPOLL_CLOSE_COMMAND = "close";
const FIREBASE_LONGPOLL_COMMAND_CB_NAME = "pLPCommand";
const FIREBASE_LONGPOLL_DATA_CB_NAME = "pRTLPCB";
const FIREBASE_LONGPOLL_ID_PARAM = "id";
const FIREBASE_LONGPOLL_PW_PARAM = "pw";
const FIREBASE_LONGPOLL_SERIAL_PARAM = "ser";
const FIREBASE_LONGPOLL_CALLBACK_ID_PARAM = "cb";
const FIREBASE_LONGPOLL_SEGMENT_NUM_PARAM = "seg";
const FIREBASE_LONGPOLL_SEGMENTS_IN_PACKET = "ts";
const FIREBASE_LONGPOLL_DATA_PARAM = "d";
const FIREBASE_LONGPOLL_DISCONN_FRAME_PARAM = "disconn";
const FIREBASE_LONGPOLL_DISCONN_FRAME_REQUEST_PARAM = "dframe";
const MAX_URL_DATA_SIZE = 1870;
const SEG_HEADER_SIZE = 30;
const MAX_PAYLOAD_SIZE = MAX_URL_DATA_SIZE - SEG_HEADER_SIZE;
const KEEPALIVE_REQUEST_INTERVAL = 25E3;
const LP_CONNECT_TIMEOUT = 3E4;

class BrowserPollConnection {
  final connId;
  final log_ = util.logWrapper(connId);
  final repoInfo;
  var bytesSent = 0;
  var bytesReceived = 0;
  final stats_;
  final transportSessionId;
  bool eventConnected_ = false;
  final lastSessionId;

  urlFn(params) {
    return repoInfo.connectionURL(Constants.LONG_POLLING, params);
  }

  BrowserPollConnection(this.connId, this.repoInfo, [this.transportSessionId, this.lastSessionId])
      : log_ = util.logWrapper(connId), stats_ = new StatsManager.getCollection(repoInfo);


  open(onMessage, onDisconnect) {
    this.curSegmentNum = 0;
    this.onDisconnect_ = onDisconnect;
    this.myPacketOrderer = new fb.realtime.polling.PacketReceiver(onMessage);
    this.isClosed_ = false;
    var self = this;
    this.connectTimeoutTimer_ = setTimeout(() {
      self.log_("Timed out trying to connect.");
      self.onClosed_();
      self.connectTimeoutTimer_ = null;
    }, math.floor(LP_CONNECT_TIMEOUT));
    fb.core.util.executeWhenDOMReady(() {
      if (self.isClosed_) {
        return;
      }
      self.scriptTagHolder = new FirebaseIFrameScriptHolder((command, arg1, arg2, arg3, arg4) {
        self.incrementIncomingBytes_(arguments);
        if (!self.scriptTagHolder) {
          return;
        }
        if (self.connectTimeoutTimer_) {
          clearTimeout(self.connectTimeoutTimer_);
          self.connectTimeoutTimer_ = null;
        }
        self.everConnected_ = true;
        if (command == FIREBASE_LONGPOLL_START_PARAM) {
          self.id = arg1;
          self.password = arg2;
        } else {
          if (command == FIREBASE_LONGPOLL_CLOSE_COMMAND) {
            if (arg1) {
              self.scriptTagHolder.sendNewPolls = false;
              self.myPacketOrderer.closeAfter(arg1, () {
                self.onClosed_();
              });
            } else {
              self.onClosed_();
            }
          } else {
            throw new Exception("Unrecognized command received: " + command);
          }
        }
      }, (pN, data) {
        self.incrementIncomingBytes_(arguments);
        self.myPacketOrderer.handleResponse(pN, data);
      }, () {
        self.onClosed_();
      }, self.urlFn);
      var urlParams = {};
      urlParams[FIREBASE_LONGPOLL_START_PARAM] = "t";
      urlParams[FIREBASE_LONGPOLL_SERIAL_PARAM] = Math.floor(Math.random() * 1E8);
      if (self.scriptTagHolder.uniqueCallbackIdentifier) {
        urlParams[FIREBASE_LONGPOLL_CALLBACK_ID_PARAM] = self.scriptTagHolder.uniqueCallbackIdentifier;
      }
      urlParams[fb.realtime.Constants.VERSION_PARAM] = fb.realtime.Constants.PROTOCOL_VERSION;
      if (self.transportSessionId) {
        urlParams[fb.realtime.Constants.TRANSPORT_SESSION_PARAM] = self.transportSessionId;
      }
      if (self.lastSessionId) {
        urlParams[fb.realtime.Constants.LAST_SESSION_PARAM] = self.lastSessionId;
      }
      if (!NODE_CLIENT && location!=null && location.href && location.href.indexOf(fb.realtime.Constants.FORGE_DOMAIN) != -1) {
        urlParams[fb.realtime.Constants.REFERER_PARAM] = fb.realtime.Constants.FORGE_REF;
      }
      var connectURL = self.urlFn(urlParams);
      self.log_("Connecting via long-poll to " + connectURL);
      self.scriptTagHolder.addTag(connectURL, () {
      });
    });
  }


  start() {
    this.scriptTagHolder.startLongPoll(this.id, this.password);
    this.addDisconnectPingFrame(this.id, this.password);
  }

  static forceAllow() {
    forceAllow_ = true;
  }

  static forceDisallow() {
    forceDisallow_ = true;
  }

  static get isAvailable {

    return fb.realtime.BrowserPollConnection.forceAllow_ ||
        !fb.realtime.BrowserPollConnection.forceDisallow_ &&
            document!=null && goog.isDefAndNotNull(document.createElement) &&
            !fb.core.util.isChromeExtensionContentScript() &&
            !fb.core.util.isWindowsStoreApp() && !NODE_CLIENT;
  }

  markConnectionHealthy() {
  }

  shutdown_() {
    this.isClosed_ = true;
    if (this.scriptTagHolder) {
      this.scriptTagHolder.close();
      this.scriptTagHolder = null;
    }
    if (this.myDisconnFrame) {
      document.body.removeChild(this.myDisconnFrame);
      this.myDisconnFrame = null;
    }
    if (this.connectTimeoutTimer_) {
      clearTimeout(this.connectTimeoutTimer_);
      this.connectTimeoutTimer_ = null;
    }
  }

  onClosed_() {
    if (!this.isClosed_) {
      this.log_("Longpoll is closing itself");
      this.shutdown_();
      if (this.onDisconnect_) {
        this.onDisconnect_(this.everConnected_);
        this.onDisconnect_ = null;
      }
    }
  }

  close() {
    if (!this.isClosed_) {
      this.log_("Longpoll is being closed.");
      this.shutdown_();
    }
  }

  send(data) {
    var dataStr = fb.util.json.stringify(data);
    this.bytesSent += dataStr.length;
    this.stats_.incrementCounter("bytes_sent", dataStr.length);
    var base64data = fb.core.util.base64Encode(dataStr);
    var dataSegs = fb.core.util.splitStringBySize(base64data, MAX_PAYLOAD_SIZE);
    for (var i = 0;i < dataSegs.length;i++) {
      this.scriptTagHolder.enqueueSegment(this.curSegmentNum, dataSegs.length, dataSegs[i]);
      this.curSegmentNum++;
    }
  }

  addDisconnectPingFrame(id, pw) {
    if (NODE_CLIENT) {
      return;
    }
    this.myDisconnFrame = document.createElement("iframe");
    var urlParams = {};
    urlParams[FIREBASE_LONGPOLL_DISCONN_FRAME_REQUEST_PARAM] = "t";
    urlParams[FIREBASE_LONGPOLL_ID_PARAM] = id;
    urlParams[FIREBASE_LONGPOLL_PW_PARAM] = pw;
    this.myDisconnFrame.src = this.urlFn(urlParams);
    this.myDisconnFrame.style.display = "none";
    document.body.appendChild(this.myDisconnFrame);
  }

  incrementIncomingBytes_(args) {
    var bytesReceived = fb.util.json.stringify(args).length;
    this.bytesReceived += bytesReceived;
    this.stats_.incrementCounter("bytes_received", bytesReceived);
  }

}

const WEBSOCKET_MAX_FRAME_SIZE = 16384;
const WEBSOCKET_KEEPALIVE_INTERVAL = 45E3;

class WebSocketConnection {

  fb.realtime.WebSocketConnection(connId, repoInfo, opt_transportSessionId, opt_lastSessionId) {
  this.connId = connId;
  this.log_ = fb.core.util.logWrapper(this.connId);
  this.keepaliveTimer = null;
  this.frames = null;
  this.totalFrames = 0;
  this.bytesSent = 0;
  this.bytesReceived = 0;
  this.stats_ = fb.core.stats.StatsManager.getCollection(repoInfo);
  this.connURL = this.connectionURL_(repoInfo, opt_transportSessionId, opt_lastSessionId);
  }


  connectionURL_(repoInfo, opt_transportSessionId, opt_lastSessionId) {
    var urlParams = {};
    urlParams[fb.realtime.Constants.VERSION_PARAM] = fb.realtime.Constants.PROTOCOL_VERSION;
    if (!NODE_CLIENT && location!=null && location.href &&
        location.href.indexOf(fb.realtime.Constants.FORGE_DOMAIN) != -1) {
      urlParams[fb.realtime.Constants.REFERER_PARAM] = fb.realtime.Constants.FORGE_REF;
    }
    if (opt_transportSessionId) {
      urlParams[fb.realtime.Constants.TRANSPORT_SESSION_PARAM] = opt_transportSessionId;
    }
    if (opt_lastSessionId) {
      urlParams[fb.realtime.Constants.LAST_SESSION_PARAM] = opt_lastSessionId;
    }
    return repoInfo.connectionURL(fb.realtime.Constants.WEBSOCKET, urlParams);
  }

  open(onMess, onDisconn) {
    this.onDisconnect = onDisconn;
    this.onMessage = onMess;
    this.log_("Websocket connecting to " + this.connURL);
    this.everConnected_ = false;
    fb.core.storage.PersistentStorage.set("previous_websocket_failure", true);
    try {
      if (NODE_CLIENT) {
        var options = {"headers":{"User-Agent":"Firebase/" + fb.realtime.Constants.PROTOCOL_VERSION + "/" + CLIENT_VERSION + "/" + process.platform + "/Node"}};
        this.mySock = new fb.WebSocket(this.connURL, [], options);
      } else {
        this.mySock = new fb.WebSocket(this.connURL);
      }
    } catch (e) {
      this.log_("Error instantiating WebSocket.");
      var error = e.message || e.data;
      if (error) {
        this.log_(error);
      }
      this.onClosed_();
      return;
    }
    var self = this;
    this.mySock.onopen = () {
      self.log_("Websocket connected.");
      self.everConnected_ = true;
    };
    this.mySock.onclose = () {
      self.log_("Websocket connection was disconnected.");
      self.mySock = null;
      self.onClosed_();
    };
    this.mySock.onmessage = (m) {
      self.handleIncomingFrame(m);
    };
    this.mySock.onerror = (e) {
      self.log_("WebSocket error.  Closing connection.");
      var error = e.message || e.data;
      if (error) {
        self.log_(error);
      }
      self.onClosed_();
    };
  }

  start() {
  }

  static forceDisallow() {
    forceDisallow_ = true;
  }

  static bool get isAvailable {
    var isOldAndroid = false;
    if (navigator!=null && navigator.userAgent) {
      var oldAndroidRegex = new RegExp(r"Android ([0-9]{0,}\.[0-9]{0,})");
      var oldAndroidMatch = navigator.userAgent.match(oldAndroidRegex);
      if (oldAndroidMatch && oldAndroidMatch.length > 1) {
        if (parseFloat(oldAndroidMatch[1]) < 4.4) {
          isOldAndroid = true;
        }
      }
    }
    return!isOldAndroid && fb.WebSocket != null && !fb.realtime.WebSocketConnection.forceDisallow_;
  }

  static const responsesRequiredToBeHealthy = 2;
  static const healthyTimeout = 3E4;

  static previouslyFailed() {
    return fb.core.storage.PersistentStorage.isInMemoryStorage || fb.core.storage.PersistentStorage.get("previous_websocket_failure") == true;
  }

  markConnectionHealthy() {
    fb.core.storage.PersistentStorage.remove("previous_websocket_failure");
  }

  appendFrame_(data) {
    this.frames.push(data);
    if (this.frames.length == this.totalFrames) {
      var fullMess = this.frames.join("");
      this.frames = null;
      var jsonMess = fb.util.json.eval(fullMess);
      this.onMessage(jsonMess);
    }
  }

  handleNewFrameCount_(frameCount) {
    this.totalFrames = frameCount;
    this.frames = [];
  }

  extractFrameCount_(data) {
    assert(this.frames == null);//, "We already have a frame buffer");
    if (data.length <= 6) {
      var frameCount = Number(data);
      if (!isNaN(frameCount)) {
        this.handleNewFrameCount_(frameCount);
        return null;
      }
    }
    this.handleNewFrameCount_(1);
    return data;
  }


  handleIncomingFrame(mess) {
    if (this.mySock == null) {
      return;
    }
    var data = mess["data"];
    this.bytesReceived += data.length;
    this.stats_.incrementCounter("bytes_received", data.length);
    this.resetKeepAlive();
    if (this.frames != null) {
      this.appendFrame_(data);
    } else {
      var remainingData = this.extractFrameCount_(data);
      if (remainingData != null) {
        this.appendFrame_(remainingData);
      }
    }
  }

  send(data) {
    this.resetKeepAlive();
    var dataStr = fb.util.json.stringify(data);
    this.bytesSent += dataStr.length;
    this.stats_.incrementCounter("bytes_sent", dataStr.length);
    var dataSegs = fb.core.util.splitStringBySize(dataStr, WEBSOCKET_MAX_FRAME_SIZE);
    if (dataSegs.length > 1) {
      this.sendString_(String(dataSegs.length));
    }
    for (var i = 0;i < dataSegs.length;i++) {
      this.sendString_(dataSegs[i]);
    }
  }

  shutdown_() {
    this.isClosed_ = true;
    if (this.keepaliveTimer) {
      clearInterval(this.keepaliveTimer);
      this.keepaliveTimer = null;
    }
    if (this.mySock) {
      this.mySock.close();
      this.mySock = null;
    }
  }

  onClosed_() {
    if (!this.isClosed_) {
      this.log_("WebSocket is closing itself");
      this.shutdown_();
      if (this.onDisconnect) {
        this.onDisconnect(this.everConnected_);
        this.onDisconnect = null;
      }
    }
  }

  close() {
    if (!this.isClosed_) {
      this.log_("WebSocket is being closed");
      this.shutdown_();
    }
  }

  resetKeepAlive() {
    var self = this;
    clearInterval(this.keepaliveTimer);
    this.keepaliveTimer = setInterval(() {
      if (self.mySock) {
        self.sendString_("0");
      }
      self.resetKeepAlive();
    }, Math.floor(WEBSOCKET_KEEPALIVE_INTERVAL));
  }
  sendString_(str) {
    try {
      this.mySock.send(str);
    } catch (e) {
      this.log_("Exception thrown from WebSocket.send():", e.message || e.data, "Closing connection.");
      setTimeout(goog.bind(this.onClosed_, this), 0);
    }
  }

}

class TransportManager {
  TransportManager(repoInfo) {
    this.initTransports_(repoInfo);
  }

  static const ALL_TRANSPORTS = const [BrowserPollConnection, WebSocketConnection];

  initTransports_(repoInfo) {
    var isWebSocketsAvailable = WebSocketConnection && WebSocketConnection["isAvailable"]();
    var isSkipPollConnection = isWebSocketsAvailable && !fb.realtime.WebSocketConnection.previouslyFailed();
    if (repoInfo.webSocketOnly) {
      if (!isWebSocketsAvailable) {
        fb.core.util.warn("wss:// URL used, but browser isn't known to support websockets.  Trying anyway.");
      }
      isSkipPollConnection = true;
    }
    if (isSkipPollConnection) {
      this.transports_ = [fb.realtime.WebSocketConnection];
    } else {
      var transports = this.transports_ = [];
      fb.core.util.each(
          fb.realtime.TransportManager.ALL_TRANSPORTS, (i, transport) {
        if (transport && transport["isAvailable"]()) {
          transports.push(transport);
        }
      });
    }

  }


  initialTransport() {
    if (this.transports_.length > 0) {
      return this.transports_[0];
    } else {
      throw new Error("No transports available");
    }
  }

  upgradeTransport() {
    if (this.transports_.length > 1) {
      return this.transports_[1];
    } else {
      return null;
    }
  }

}

var UPGRADE_TIMEOUT = 6E4;
var DELAY_BEFORE_SENDING_EXTRA_REQUESTS = 5E3;
var BYTES_SENT_HEALTHY_OVERRIDE = 10 * 1024;
var BYTES_RECEIVED_HEALTHY_OVERRIDE = 100 * 1024;
var REALTIME_STATE_CONNECTING = 0;
var REALTIME_STATE_CONNECTED = 1;
var REALTIME_STATE_DISCONNECTED = 2;
var MESSAGE_TYPE = "t";
var MESSAGE_DATA = "d";
var CONTROL_SHUTDOWN = "s";
var CONTROL_RESET = "r";
var CONTROL_ERROR = "e";
var CONTROL_PONG = "o";
var SWITCH_ACK = "a";
var END_TRANSMISSION = "n";
var PING = "p";
var SERVER_HELLO = "h";

class Connection {

  Connection(connId, repoInfo, onMessage, onReady, onDisconnect, onKill, lastSessionId) {
    this.id = connId;
    this.log_ = fb.core.util.logWrapper("c:" + this.id + ":");
    this.onMessage_ = onMessage;
    this.onReady_ = onReady;
    this.onDisconnect_ = onDisconnect;
    this.onKill_ = onKill;
    this.repoInfo_ = repoInfo;
    this.pendingDataMessages = [];
    this.connectionCount = 0;
    this.transportManager_ = new fb.realtime.TransportManager(repoInfo);
    this.state_ = REALTIME_STATE_CONNECTING;
    this.lastSessionId = lastSessionId;
    this.log_("Connection created");
    this.start_();
  }

  start_() {
    var conn = this.transportManager_.initialTransport();
    this.conn_ = new conn(this.nextTransportId_(), this.repoInfo_, undefined, this.lastSessionId);
    this.primaryResponsesRequired_ = conn["responsesRequiredToBeHealthy"] || 0;
    var onMessageReceived = this.connReceiver_(this.conn_);
    var onConnectionLost = this.disconnReceiver_(this.conn_);
    this.tx_ = this.conn_;
    this.rx_ = this.conn_;
    this.secondaryConn_ = null;
    this.isHealthy_ = false;
    var self = this;
    setTimeout(() {
      self.conn_ && self.conn_.open(onMessageReceived, onConnectionLost);
    }, Math.floor(0));
    var healthyTimeout_ms = conn["healthyTimeout"] || 0;
    if (healthyTimeout_ms > 0) {
      this.healthyTimeout_ = setTimeout(() {
        self.healthyTimeout_ = null;
        if (!self.isHealthy_) {
          if (self.conn_ && self.conn_.bytesReceived > BYTES_RECEIVED_HEALTHY_OVERRIDE) {
            self.log_("Connection exceeded healthy timeout but has received " + self.conn_.bytesReceived + " bytes.  Marking connection healthy.");
            self.isHealthy_ = true;
            self.conn_.markConnectionHealthy();
          } else {
            if (self.conn_ && self.conn_.bytesSent > BYTES_SENT_HEALTHY_OVERRIDE) {
              self.log_("Connection exceeded healthy timeout but has sent " + self.conn_.bytesSent + " bytes.  Leaving connection alive.");
            } else {
              self.log_("Closing unhealthy connection after timeout.");
              self.close();
            }
          }
        }
      }, Math.floor(healthyTimeout_ms));
    }
  }

  nextTransportId_() {
    return "c:" + this.id + ":" + this.connectionCount++;
  }

  disconnReceiver_(conn) {
    var self = this;
    return (everConnected) {
      if (conn == self.conn_) {
        self.onConnectionLost_(everConnected);
      } else {
        if (conn == self.secondaryConn_) {
          self.log_("Secondary connection lost.");
          self.onSecondaryConnectionLost_();
        } else {
          self.log_("closing an old connection");
        }
      }
    };
  }

  connReceiver_(conn) {
    var self = this;
    return (message) {
      if (self.state_ != REALTIME_STATE_DISCONNECTED) {
        if (conn == self.rx_) {
          self.onPrimaryMessageReceived_(message);
        } else {
          if (conn == self.secondaryConn_) {
            self.onSecondaryMessageReceived_(message);
          } else {
            self.log_("message on old connection");
          }
        }
      }
    };
  }

  sendRequest(dataMsg) {
    var msg = {"t":"d", "d":dataMsg};
    this.sendData_(msg);
  }

  tryCleanupConnection() {
    if (this.tx_ == this.secondaryConn_ && this.rx_ == this.secondaryConn_) {
      this.log_("cleaning up and promoting a connection: " + this.secondaryConn_.connId);
      this.conn_ = this.secondaryConn_;
      this.secondaryConn_ = null;
    }
  }

  onSecondaryControl_(controlData) {
    if (MESSAGE_TYPE in controlData) {
      var cmd = controlData[MESSAGE_TYPE];
      if (cmd == SWITCH_ACK) {
        this.upgradeIfSecondaryHealthy_();
      } else {
        if (cmd == CONTROL_RESET) {
          this.log_("Got a reset on secondary, closing it");
          this.secondaryConn_.close();
          if (this.tx_ == this.secondaryConn_ || this.rx_ == this.secondaryConn_) {
            this.close();
          }
        } else {
          if (cmd == CONTROL_PONG) {
            this.log_("got pong on secondary.");
            this.secondaryResponsesRequired_--;
            this.upgradeIfSecondaryHealthy_();
          }
        }
      }
    }
  }

  onSecondaryMessageReceived_(parsedData) {
    var layer = fb.core.util.requireKey("t", parsedData);
    var data = fb.core.util.requireKey("d", parsedData);
    if (layer == "c") {
      this.onSecondaryControl_(data);
    } else {
      if (layer == "d") {
        this.pendingDataMessages.push(data);
      } else {
        throw new Error("Unknown protocol layer: " + layer);
      }
    }
  }


  upgradeIfSecondaryHealthy_() {
    if (this.secondaryResponsesRequired_ <= 0) {
      this.log_("Secondary connection is healthy.");
      this.isHealthy_ = true;
      this.secondaryConn_.markConnectionHealthy();
      this.proceedWithUpgrade_();
    } else {
      this.log_("sending ping on secondary.");
      this.secondaryConn_.send({"t":"c", "d":{"t":PING, "d":{}}});
    }
  }

  proceedWithUpgrade_() {
    this.secondaryConn_.start();
    this.log_("sending client ack on secondary");
    this.secondaryConn_.send({"t":"c", "d":{"t":SWITCH_ACK, "d":{}}});
    this.log_("Ending transmission on primary");
    this.conn_.send({"t":"c", "d":{"t":END_TRANSMISSION, "d":{}}});
    this.tx_ = this.secondaryConn_;
    this.tryCleanupConnection();
  }

  onPrimaryMessageReceived_(parsedData) {
    var layer = fb.core.util.requireKey("t", parsedData);
    var data = fb.core.util.requireKey("d", parsedData);
    if (layer == "c") {
      this.onControl_(data);
    } else {
      if (layer == "d") {
        this.onDataMessage_(data);
      }
    }
  }

  onDataMessage_(message) {
    this.onPrimaryResponse_();
    this.onMessage_(message);
  }

  onPrimaryResponse_() {
    if (!this.isHealthy_) {
      this.primaryResponsesRequired_--;
      if (this.primaryResponsesRequired_ <= 0) {
        this.log_("Primary connection is healthy.");
        this.isHealthy_ = true;
        this.conn_.markConnectionHealthy();
      }
    }
  }

  onControl_(controlData) {
    var cmd = fb.core.util.requireKey(MESSAGE_TYPE, controlData);
    if (MESSAGE_DATA in controlData) {
      var payload = controlData[MESSAGE_DATA];
      if (cmd == SERVER_HELLO) {
        this.onHandshake_(payload);
      } else {
        if (cmd == END_TRANSMISSION) {
          this.log_("recvd end transmission on primary");
          this.rx_ = this.secondaryConn_;
          for (var i = 0;i < this.pendingDataMessages.length;++i) {
            this.onDataMessage_(this.pendingDataMessages[i]);
          }
          this.pendingDataMessages = [];
          this.tryCleanupConnection();
        } else {
          if (cmd == CONTROL_SHUTDOWN) {
            this.onConnectionShutdown_(payload);
          } else {
            if (cmd == CONTROL_RESET) {
              this.onReset_(payload);
            } else {
              if (cmd == CONTROL_ERROR) {
                fb.core.util.error("Server Error: " + payload);
              } else {
                if (cmd == CONTROL_PONG) {
                  this.log_("got pong on primary.");
                  this.onPrimaryResponse_();
                  this.sendPingOnPrimaryIfNecessary_();
                } else {
                  fb.core.util.error("Unknown control packet command: " + cmd);
                }
              }
            }
          }
        }
      }
    }
  }

  onHandshake_(handshake) {
    var timestamp = handshake["ts"];
    var version = handshake["v"];
    var host = handshake["h"];
    this.sessionId = handshake["s"];
    this.repoInfo_.updateHost(host);
    if (this.state_ == REALTIME_STATE_CONNECTING) {
      this.conn_.start();
      this.onConnectionEstablished_(this.conn_, timestamp);
      if (fb.realtime.Constants.PROTOCOL_VERSION != version) {
        fb.core.util.warn("Protocol version mismatch detected");
      }
      this.tryStartUpgrade_();
    }
  }

  tryStartUpgrade_() {
    var conn = this.transportManager_.upgradeTransport();
    if (conn) {
      this.startUpgrade_(conn);
    }
  }

  startUpgrade_(conn) {
    this.secondaryConn_ = new conn(this.nextTransportId_(), this.repoInfo_, this.sessionId);
    this.secondaryResponsesRequired_ = conn["responsesRequiredToBeHealthy"] || 0;
    var onMessage = this.connReceiver_(this.secondaryConn_);
    var onDisconnect = this.disconnReceiver_(this.secondaryConn_);
    this.secondaryConn_.open(onMessage, onDisconnect);
    var self = this;
    setTimeout(function() {
    if (self.secondaryConn_) {
    self.log_("Timed out trying to upgrade.");
    self.secondaryConn_.close();
    }
    }, Math.floor(UPGRADE_TIMEOUT));
  }

  onReset_(host) {
    this.log_("Reset packet received.  New host: " + host);
    this.repoInfo_.updateHost(host);
    if (this.state_ == REALTIME_STATE_CONNECTED) {
      this.close();
    } else {
      this.closeConnections_();
      this.start_();
    }
  }

  onConnectionEstablished_(conn, timestamp) {
    this.log_("Realtime connection established.");
    this.conn_ = conn;
    this.state_ = REALTIME_STATE_CONNECTED;
    if (this.onReady_) {
      this.onReady_(timestamp, this.sessionId);
      this.onReady_ = null;
    }
    var self = this;
    if (this.primaryResponsesRequired_ == 0) {
      this.log_("Primary connection is healthy.");
      this.isHealthy_ = true;
    } else {
      setTimeout(() {
        self.sendPingOnPrimaryIfNecessary_();
      }, Math.floor(DELAY_BEFORE_SENDING_EXTRA_REQUESTS));
    }
  }

  sendPingOnPrimaryIfNecessary_() {
    if (!this.isHealthy_ && this.state_ == REALTIME_STATE_CONNECTED) {
      this.log_("sending ping on primary.");
      this.sendData_({"t":"c", "d":{"t":PING, "d":{}}});
    }
  }

  onSecondaryConnectionLost_() {
    var conn = this.secondaryConn_;
    this.secondaryConn_ = null;
    if (this.tx_ == conn || this.rx_ == conn) {
      this.close();
    }
  }

  onConnectionLost_(everConnected) {
    this.conn_ = null;
    if (!everConnected && this.state_ == REALTIME_STATE_CONNECTING) {
      this.log_("Realtime connection failed.");
      if (this.repoInfo_.isCacheableHost()) {
        fb.core.storage.PersistentStorage.remove("host:" + this.repoInfo_.host);
        this.repoInfo_.internalHost = this.repoInfo_.host;
      }
    } else {
      if (this.state_ == REALTIME_STATE_CONNECTED) {
        this.log_("Realtime connection lost.");
      }
    }
    this.close();
  }

  onConnectionShutdown_(reason) {
    this.log_("Connection shutdown command received. Shutting down...");
    if (this.onKill_) {
      this.onKill_(reason);
      this.onKill_ = null;
    }
    this.onDisconnect_ = null;
    this.close();
  }

  sendData_(data) {
    if (this.state_ != REALTIME_STATE_CONNECTED) {
      throw "Connection is not connected";
    } else {
      this.tx_.send(data);
    }
  }

  close() {
    if (this.state_ != REALTIME_STATE_DISCONNECTED) {
      this.log_("Closing realtime connection.");
      this.state_ = REALTIME_STATE_DISCONNECTED;
      this.closeConnections_();
      if (this.onDisconnect_) {
        this.onDisconnect_();
        this.onDisconnect_ = null;
      }
    }
  }

  closeConnections_() {
    this.log_("Shutting down all connections");
    if (this.conn_) {
      this.conn_.close();
      this.conn_ = null;
    }
    if (this.secondaryConn_) {
      this.secondaryConn_.close();
      this.secondaryConn_ = null;
    }
    if (this.healthyTimeout_) {
      clearTimeout(this.healthyTimeout_);
      this.healthyTimeout_ = null;
    }
  }

}
