goog.provide("fb.core.PersistentConnection");
goog.require("fb.core.ServerActions");
goog.require("fb.core.util");
goog.require("fb.core.util.OnlineMonitor");
goog.require("fb.core.util.VisibilityMonitor");
goog.require("fb.login.util.environment");
goog.require("fb.realtime.Connection");
goog.require("fb.util.json");
goog.require("fb.util.jwt");
var RECONNECT_MIN_DELAY = 1E3;
var RECONNECT_MAX_DELAY_DEFAULT = 60 * 5 * 1E3;
var RECONNECT_MAX_DELAY_FOR_ADMINS = 30 * 1E3;
var RECONNECT_DELAY_MULTIPLIER = 1.3;
var RECONNECT_DELAY_RESET_TIMEOUT = 3E4;
fb.core.PersistentConnection = goog.defineClass(null, {constructor:function(repoInfo, onDataUpdate, onConnectStatus, onServerInfoUpdate) {
    this.id = fb.core.PersistentConnection.nextPersistentConnectionId_++;
    this.log_ = fb.core.util.logWrapper("p:" + this.id + ":");
    this.interrupted_ = false;
    this.killed_ = false;
    this.listens_ = {};
    this.outstandingPuts_ = [];
    this.outstandingPutCount_ = 0;
    this.onDisconnectRequestQueue_ = [];
    this.connected_ = false;
    this.reconnectDelay_ = RECONNECT_MIN_DELAY;
    this.maxReconnectDelay_ = RECONNECT_MAX_DELAY_DEFAULT;
    this.onDataUpdate_ = onDataUpdate;
    this.onConnectStatus_ = onConnectStatus;
    this.onServerInfoUpdate_ = onServerInfoUpdate;
    this.repoInfo_ = repoInfo;
    this.securityDebugCallback_ = null;
    this.lastSessionId = null;
    this.realtime_ = null;
    this.credential_ = null;
    this.establishConnectionTimer_ = null;
    this.visible_ = false;
    this.requestCBHash_ = {};
    this.requestNumber_ = 0;
    this.firstConnection_ = true;
    this.lastConnectionAttemptTime_ = null;
    this.lastConnectionEstablishedTime_ = null;
    this.scheduleConnect_(0);
    fb.core.util.VisibilityMonitor.getInstance().on("visible", this.onVisible_, this);
    if (repoInfo.host.indexOf("fblocal") === -1) {
        fb.core.util.OnlineMonitor.getInstance().on("online", this.onOnline_, this);
    }
}, statics:{nextPersistentConnectionId_:0, nextConnectionId_:0}, sendRequest:function(action, body, onResponse) {
    var curReqNum = ++this.requestNumber_;
    var msg = {"r":curReqNum, "a":action, "b":body};
    this.log_(fb.util.json.stringify(msg));
    fb.core.util.assert(this.connected_, "sendRequest call when we're not connected not allowed.");
    this.realtime_.sendRequest(msg);
    if (onResponse) {
        this.requestCBHash_[curReqNum] = onResponse;
    }
}, listen:function(query, currentHashFn, tag, onComplete) {
    var queryId = query.queryIdentifier();
    var pathString = query.path.toString();
    this.log_("Listen called for " + pathString + " " + queryId);
    this.listens_[pathString] = this.listens_[pathString] || {};
    fb.core.util.assert(query.getQueryParams().isDefault() || !query.getQueryParams().loadsAllData(), "listen() called for non-default but complete query");
    fb.core.util.assert(!this.listens_[pathString][queryId], "listen() called twice for same path/queryId.");
    var listenSpec = {onComplete:onComplete, hashFn:currentHashFn, query:query, tag:tag};
    this.listens_[pathString][queryId] = listenSpec;
    if (this.connected_) {
        this.sendListen_(listenSpec);
    }
}, sendListen_:function(listenSpec) {
    var query = listenSpec.query;
    var pathString = query.path.toString();
    var queryId = query.queryIdentifier();
    var self = this;
    this.log_("Listen on " + pathString + " for " + queryId);
    var req = {"p":pathString};
    var action = "q";
    if (listenSpec.tag) {
        req["q"] = query.queryObject();
        req["t"] = listenSpec.tag;
    }
    req["h"] = listenSpec.hashFn();
    this.sendRequest(action, req, function(message) {
        var payload = message["d"];
        var status = message["s"];
        self.warnOnListenWarnings_(payload, query);
        var currentListenSpec = self.listens_[pathString] && self.listens_[pathString][queryId];
        if (currentListenSpec === listenSpec) {
            self.log_("listen response", message);
            if (status !== "ok") {
                self.removeListen_(pathString, queryId);
            }
            if (listenSpec.onComplete) {
                listenSpec.onComplete(status, payload);
            }
        }
    });
}, warnOnListenWarnings_:function(payload, query) {
    if (payload && typeof payload === "object" && fb.util.obj.contains(payload, "w")) {
        var warnings = fb.util.obj.get(payload, "w");
        if (goog.isArray(warnings) && goog.array.contains(warnings, "no_index")) {
            var indexSpec = '".indexOn": "' + query.getQueryParams().getIndex().toString() + '"';
            var indexPath = query.path.toString();
            fb.core.util.warn("Using an unspecified index. Consider adding " + indexSpec + " at " + indexPath + " to your security rules for better performance");
        }
    }
}, auth:function(cred, opt_callback, opt_cancelCallback) {
    this.credential_ = {cred:cred, firstRequestSent:false, callback:opt_callback, cancelCallback:opt_cancelCallback};
    this.log_("Authenticating using credential: " + cred);
    this.tryAuth();
    this.reduceReconnectDelayIfAdminCredential_(cred);
}, reduceReconnectDelayIfAdminCredential_:function(credential) {
    var isFirebaseSecret = credential.length == 40;
    if (isFirebaseSecret || fb.util.jwt.isAdmin(credential)) {
        this.log_("Admin auth credential detected.  Reducing max reconnect time.");
        this.maxReconnectDelay_ = RECONNECT_MAX_DELAY_FOR_ADMINS;
    }
}, unauth:function(onComplete) {
    delete this.credential_;
    if (this.connected_) {
        this.sendRequest("unauth", {}, function(result) {
            var status = result["s"];
            var errorReason = result["d"];
            onComplete(status, errorReason);
        });
    }
}, tryAuth:function() {
    var authdata = this.credential_;
    var self = this;
    if (this.connected_ && authdata) {
        var requestData = {"cred":authdata.cred};
        this.sendRequest("auth", requestData, function(res) {
            var status = res["s"];
            var data = res["d"] || "error";
            if (status !== "ok" && self.credential_ === authdata) {
                delete self.credential_;
            }
            if (!authdata.firstRequestSent) {
                authdata.firstRequestSent = true;
                if (authdata.callback) {
                    authdata.callback(status, data);
                }
            } else {
                if (status !== "ok" && authdata.cancelCallback) {
                    authdata.cancelCallback(status, data);
                }
            }
        });
    }
}, unlisten:function(query, tag) {
    var pathString = query.path.toString();
    var queryId = query.queryIdentifier();
    this.log_("Unlisten called for " + pathString + " " + queryId);
    fb.core.util.assert(query.getQueryParams().isDefault() || !query.getQueryParams().loadsAllData(), "unlisten() called for non-default but complete query");
    var listen = this.removeListen_(pathString, queryId);
    if (listen && this.connected_) {
        this.sendUnlisten_(pathString, queryId, query.queryObject(), tag);
    }
}, sendUnlisten_:function(pathString, queryId, queryObj, tag) {
    this.log_("Unlisten on " + pathString + " for " + queryId);
    var self = this;
    var req = {"p":pathString};
    var action = "n";
    if (tag) {
        req["q"] = queryObj;
        req["t"] = tag;
    }
    this.sendRequest(action, req);
}, onDisconnectPut:function(pathString, data, opt_onComplete) {
    if (this.connected_) {
        this.sendOnDisconnect_("o", pathString, data, opt_onComplete);
    } else {
        this.onDisconnectRequestQueue_.push({pathString:pathString, action:"o", data:data, onComplete:opt_onComplete});
    }
}, onDisconnectMerge:function(pathString, data, opt_onComplete) {
    if (this.connected_) {
        this.sendOnDisconnect_("om", pathString, data, opt_onComplete);
    } else {
        this.onDisconnectRequestQueue_.push({pathString:pathString, action:"om", data:data, onComplete:opt_onComplete});
    }
}, onDisconnectCancel:function(pathString, opt_onComplete) {
    if (this.connected_) {
        this.sendOnDisconnect_("oc", pathString, null, opt_onComplete);
    } else {
        this.onDisconnectRequestQueue_.push({pathString:pathString, action:"oc", data:null, onComplete:opt_onComplete});
    }
}, sendOnDisconnect_:function(action, pathString, data, opt_onComplete) {
    var self = this;
    var request = {"p":pathString, "d":data};
    self.log_("onDisconnect " + action, request);
    this.sendRequest(action, request, function(response) {
        if (opt_onComplete) {
            setTimeout(function() {
                opt_onComplete(response["s"], response["d"]);
            }, Math.floor(0));
        }
    });
}, put:function(pathString, data, opt_onComplete, opt_hash) {
    this.putInternal("p", pathString, data, opt_onComplete, opt_hash);
}, merge:function(pathString, data, onComplete, opt_hash) {
    this.putInternal("m", pathString, data, onComplete, opt_hash);
}, putInternal:function(action, pathString, data, opt_onComplete, opt_hash) {
    var request = {"p":pathString, "d":data};
    if (goog.isDef(opt_hash)) {
        request["h"] = opt_hash;
    }
    this.outstandingPuts_.push({action:action, request:request, onComplete:opt_onComplete});
    this.outstandingPutCount_++;
    var index = this.outstandingPuts_.length - 1;
    if (this.connected_) {
        this.sendPut_(index);
    } else {
        this.log_("Buffering put: " + pathString);
    }
}, sendPut_:function(index) {
    var self = this;
    var action = this.outstandingPuts_[index].action;
    var request = this.outstandingPuts_[index].request;
    var onComplete = this.outstandingPuts_[index].onComplete;
    this.outstandingPuts_[index].queued = this.connected_;
    this.sendRequest(action, request, function(message) {
        self.log_(action + " response", message);
        delete self.outstandingPuts_[index];
        self.outstandingPutCount_--;
        if (self.outstandingPutCount_ === 0) {
            self.outstandingPuts_ = [];
        }
        if (onComplete) {
            onComplete(message["s"], message["d"]);
        }
    });
}, reportStats:function(stats) {
    if (this.connected_) {
        var request = {"c":stats};
        this.log_("reportStats", request);
        this.sendRequest("s", request, function(result) {
            var status = result["s"];
            if (status !== "ok") {
                var errorReason = result["d"];
                this.log_("reportStats", "Error sending stats: " + errorReason);
            }
        });
    }
}, onDataMessage_:function(message) {
    if ("r" in message) {
        this.log_("from server: " + fb.util.json.stringify(message));
        var reqNum = message["r"];
        var onResponse = this.requestCBHash_[reqNum];
        if (onResponse) {
            delete this.requestCBHash_[reqNum];
            onResponse(message["b"]);
        }
    } else {
        if ("error" in message) {
            throw "A server-side error has occurred: " + message["error"];
        } else {
            if ("a" in message) {
                this.onDataPush_(message["a"], message["b"]);
            }
        }
    }
}, onDataPush_:function(action, body) {
    this.log_("handleServerMessage", action, body);
    if (action === "d") {
        this.onDataUpdate_(body["p"], body["d"], false, body["t"]);
    } else {
        if (action === "m") {
            this.onDataUpdate_(body["p"], body["d"], true, body["t"]);
        } else {
            if (action === "c") {
                this.onListenRevoked_(body["p"], body["q"]);
            } else {
                if (action === "ac") {
                    this.onAuthRevoked_(body["s"], body["d"]);
                } else {
                    if (action === "sd") {
                        this.onSecurityDebugPacket_(body);
                    } else {
                        fb.core.util.error("Unrecognized action received from server: " + fb.util.json.stringify(action) + "\nAre you using the latest client?");
                    }
                }
            }
        }
    }
}, onReady_:function(timestamp, sessionId) {
    this.log_("connection ready");
    this.connected_ = true;
    this.lastConnectionEstablishedTime_ = (new Date).getTime();
    this.handleTimestamp_(timestamp);
    this.lastSessionId = sessionId;
    if (this.firstConnection_) {
        this.sendConnectStats_();
    }
    this.restoreState_();
    this.firstConnection_ = false;
    this.onConnectStatus_(true);
}, scheduleConnect_:function(timeout) {
    fb.core.util.assert(!this.realtime_, "Scheduling a connect when we're already connected/ing?");
    if (this.establishConnectionTimer_) {
        clearTimeout(this.establishConnectionTimer_);
    }
    var self = this;
    this.establishConnectionTimer_ = setTimeout(function() {
        self.establishConnectionTimer_ = null;
        self.establishConnection_();
    }, Math.floor(timeout));
}, onVisible_:function(visible) {
    if (visible && !this.visible_ && this.reconnectDelay_ === this.maxReconnectDelay_) {
        this.log_("Window became visible.  Reducing delay.");
        this.reconnectDelay_ = RECONNECT_MIN_DELAY;
        if (!this.realtime_) {
            this.scheduleConnect_(0);
        }
    }
    this.visible_ = visible;
}, onOnline_:function(online) {
    if (online) {
        this.log_("Browser went online.");
        this.reconnectDelay_ = RECONNECT_MIN_DELAY;
        if (!this.realtime_) {
            this.scheduleConnect_(0);
        }
    } else {
        this.log_("Browser went offline.  Killing connection.");
        if (this.realtime_) {
            this.realtime_.close();
        }
    }
}, onRealtimeDisconnect_:function() {
    this.log_("data client disconnected");
    this.connected_ = false;
    this.realtime_ = null;
    this.cancelSentTransactions_();
    this.requestCBHash_ = {};
    if (this.shouldReconnect_()) {
        if (!this.visible_) {
            this.log_("Window isn't visible.  Delaying reconnect.");
            this.reconnectDelay_ = this.maxReconnectDelay_;
            this.lastConnectionAttemptTime_ = (new Date).getTime();
        } else {
            if (this.lastConnectionEstablishedTime_) {
                var timeSinceLastConnectSucceeded = (new Date).getTime() - this.lastConnectionEstablishedTime_;
                if (timeSinceLastConnectSucceeded > RECONNECT_DELAY_RESET_TIMEOUT) {
                    this.reconnectDelay_ = RECONNECT_MIN_DELAY;
                }
                this.lastConnectionEstablishedTime_ = null;
            }
        }
        var timeSinceLastConnectAttempt = (new Date).getTime() - this.lastConnectionAttemptTime_;
        var reconnectDelay = Math.max(0, this.reconnectDelay_ - timeSinceLastConnectAttempt);
        reconnectDelay = Math.random() * reconnectDelay;
        this.log_("Trying to reconnect in " + reconnectDelay + "ms");
        this.scheduleConnect_(reconnectDelay);
        this.reconnectDelay_ = Math.min(this.maxReconnectDelay_, this.reconnectDelay_ * RECONNECT_DELAY_MULTIPLIER);
    }
    this.onConnectStatus_(false);
}, establishConnection_:function() {
    if (this.shouldReconnect_()) {
        this.log_("Making a connection attempt");
        this.lastConnectionAttemptTime_ = (new Date).getTime();
        this.lastConnectionEstablishedTime_ = null;
        var onDataMessage = goog.bind(this.onDataMessage_, this);
        var onReady = goog.bind(this.onReady_, this);
        var onDisconnect = goog.bind(this.onRealtimeDisconnect_, this);
        var connId = this.id + ":" + fb.core.PersistentConnection.nextConnectionId_++;
        var self = this;
        var lastSessionId = this.lastSessionId;
        this.realtime_ = new fb.realtime.Connection(connId, this.repoInfo_, onDataMessage, onReady, onDisconnect, function(reason) {
            fb.core.util.warn(reason + " (" + self.repoInfo_.toString() + ")");
            self.killed_ = true;
        }, lastSessionId);
    }
}, interrupt:function() {
    this.interrupted_ = true;
    if (this.realtime_) {
        this.realtime_.close();
    } else {
        if (this.establishConnectionTimer_) {
            clearTimeout(this.establishConnectionTimer_);
            this.establishConnectionTimer_ = null;
        }
        if (this.connected_) {
            this.onRealtimeDisconnect_();
        }
    }
}, resume:function() {
    this.interrupted_ = false;
    this.reconnectDelay_ = RECONNECT_MIN_DELAY;
    if (!this.realtime_) {
        this.scheduleConnect_(0);
    }
}, handleTimestamp_:function(timestamp) {
    var delta = timestamp - (new Date).getTime();
    this.onServerInfoUpdate_({"serverTimeOffset":delta});
}, cancelSentTransactions_:function() {
    for (var i = 0;i < this.outstandingPuts_.length;i++) {
        var put = this.outstandingPuts_[i];
        if (put && "h" in put.request && put.queued) {
            if (put.onComplete) {
                put.onComplete("disconnect");
            }
            delete this.outstandingPuts_[i];
            this.outstandingPutCount_--;
        }
    }
    if (this.outstandingPutCount_ === 0) {
        this.outstandingPuts_ = [];
    }
}, onListenRevoked_:function(pathString, opt_query) {
    var queryId;
    if (!opt_query) {
        queryId = "default";
    } else {
        queryId = goog.array.map(opt_query, function(q) {
            return fb.core.util.ObjectToUniqueKey(q);
        }).join("$");
    }
    var listen = this.removeListen_(pathString, queryId);
    if (listen && listen.onComplete) {
        listen.onComplete("permission_denied");
    }
}, removeListen_:function(pathString, queryId) {
    var normalizedPathString = (new fb.core.util.Path(pathString)).toString();
    var listen;
    if (goog.isDef(this.listens_[normalizedPathString])) {
        listen = this.listens_[normalizedPathString][queryId];
        delete this.listens_[normalizedPathString][queryId];
        if (goog.object.getCount(this.listens_[normalizedPathString]) === 0) {
            delete this.listens_[normalizedPathString];
        }
    } else {
        listen = undefined;
    }
    return listen;
}, onAuthRevoked_:function(statusCode, explanation) {
    var cred = this.credential_;
    delete this.credential_;
    if (cred && cred.cancelCallback) {
        cred.cancelCallback(statusCode, explanation);
    }
}, onSecurityDebugPacket_:function(body) {
    if (this.securityDebugCallback_) {
        this.securityDebugCallback_(body);
    } else {
        if ("msg" in body && typeof console !== "undefined") {
            console.log("FIREBASE: " + body["msg"].replace("\n", "\nFIREBASE: "));
        }
    }
}, restoreState_:function() {
    this.tryAuth();
    var self = this;
    goog.object.forEach(this.listens_, function(queries, pathString) {
        goog.object.forEach(queries, function(listenSpec) {
            self.sendListen_(listenSpec);
        });
    });
    for (var i = 0;i < this.outstandingPuts_.length;i++) {
        if (this.outstandingPuts_[i]) {
            this.sendPut_(i);
        }
    }
    while (this.onDisconnectRequestQueue_.length) {
        var request = this.onDisconnectRequestQueue_.shift();
        this.sendOnDisconnect_(request.action, request.pathString, request.data, request.onComplete);
    }
}, sendConnectStats_:function() {
    var stats = {};
    stats["sdk.js." + CLIENT_VERSION.replace(/\./g, "-")] = 1;
    if (fb.login.util.environment.isMobileCordova()) {
        stats["framework.cordova"] = 1;
    } else {
        if (fb.login.util.environment.isReactNative()) {
            stats["framework.reactnative"] = 1;
        }
    }
    this.reportStats(stats);
}, shouldReconnect_:function() {
    var online = fb.core.util.OnlineMonitor.getInstance().currentlyOnline();
    return!this.killed_ && !this.interrupted_ && online;
}});
