goog.provide("fb.core.SyncPoint");
goog.require("fb.core.util");
goog.require("fb.core.util.ImmutableTree");
goog.require("fb.core.view.ViewCache");
goog.require("fb.core.view.EventRegistration");
goog.require("fb.core.view.View");
goog.require("goog.array");
fb.core.SyncPoint = function() {
    this.views_ = {};
};
fb.core.SyncPoint.prototype.isEmpty = function() {
    return goog.object.isEmpty(this.views_);
};
fb.core.SyncPoint.prototype.applyOperation = function(operation, writesCache, optCompleteServerCache) {
    var queryId = operation.source.queryId;
    if (queryId !== null) {
        var view = fb.util.obj.get(this.views_, queryId);
        fb.core.util.assert(view != null, "SyncTree gave us an op for an invalid query.");
        return view.applyOperation(operation, writesCache, optCompleteServerCache);
    } else {
        var events = [];
        goog.object.forEach(this.views_, function(view) {
            events = events.concat(view.applyOperation(operation, writesCache, optCompleteServerCache));
        });
        return events;
    }
};
fb.core.SyncPoint.prototype.addEventRegistration = function(query, eventRegistration, writesCache, serverCache, serverCacheComplete) {
    var queryId = query.queryIdentifier();
    var view = fb.util.obj.get(this.views_, queryId);
    if (!view) {
        var eventCache = writesCache.calcCompleteEventCache(serverCacheComplete ? serverCache : null);
        var eventCacheComplete = false;
        if (eventCache) {
            eventCacheComplete = true;
        } else {
            if (serverCache instanceof fb.core.snap.ChildrenNode) {
                eventCache = writesCache.calcCompleteEventChildren(serverCache);
                eventCacheComplete = false;
            } else {
                eventCache = fb.core.snap.EMPTY_NODE;
                eventCacheComplete = false;
            }
        }
        var viewCache = new fb.core.view.ViewCache(new fb.core.view.CacheNode((eventCache), eventCacheComplete, false), new fb.core.view.CacheNode((serverCache), serverCacheComplete, false));
        view = new fb.core.view.View(query, viewCache);
        this.views_[queryId] = view;
    }
    view.addEventRegistration(eventRegistration);
    return view.getInitialEvents(eventRegistration);
};
fb.core.SyncPoint.prototype.removeEventRegistration = function(query, eventRegistration, cancelError) {
    var queryId = query.queryIdentifier();
    var removed = [];
    var cancelEvents = [];
    var hadCompleteView = this.hasCompleteView();
    if (queryId === "default") {
        var self = this;
        goog.object.forEach(this.views_, function(view, viewQueryId) {
            cancelEvents = cancelEvents.concat(view.removeEventRegistration(eventRegistration, cancelError));
            if (view.isEmpty()) {
                delete self.views_[viewQueryId];
                if (!view.getQuery().getQueryParams().loadsAllData()) {
                    removed.push(view.getQuery());
                }
            }
        });
    } else {
        var view = fb.util.obj.get(this.views_, queryId);
        if (view) {
            cancelEvents = cancelEvents.concat(view.removeEventRegistration(eventRegistration, cancelError));
            if (view.isEmpty()) {
                delete this.views_[queryId];
                if (!view.getQuery().getQueryParams().loadsAllData()) {
                    removed.push(view.getQuery());
                }
            }
        }
    }
    if (hadCompleteView && !this.hasCompleteView()) {
        removed.push(new Firebase(query.repo, query.path));
    }
    return{removed:removed, events:cancelEvents};
};
fb.core.SyncPoint.prototype.getQueryViews = function() {
    return goog.array.filter(goog.object.getValues(this.views_), function(view) {
        return!view.getQuery().getQueryParams().loadsAllData();
    });
};
fb.core.SyncPoint.prototype.getCompleteServerCache = function(path) {
    var serverCache = null;
    goog.object.forEach(this.views_, function(view) {
        serverCache = serverCache || view.getCompleteServerCache(path);
    });
    return serverCache;
};
fb.core.SyncPoint.prototype.viewForQuery = function(query) {
    var params = query.getQueryParams();
    if (params.loadsAllData()) {
        return this.getCompleteView();
    } else {
        var queryId = query.queryIdentifier();
        return fb.util.obj.get(this.views_, queryId);
    }
};
fb.core.SyncPoint.prototype.viewExistsForQuery = function(query) {
    return this.viewForQuery(query) != null;
};
fb.core.SyncPoint.prototype.hasCompleteView = function() {
    return this.getCompleteView() != null;
};
fb.core.SyncPoint.prototype.getCompleteView = function() {
    var completeView = goog.object.findValue(this.views_, function(view) {
        return view.getQuery().getQueryParams().loadsAllData();
    });
    return completeView || null;
};
goog.provide("fb.core.SyncTree");
goog.require("fb.core.Operation");
goog.require("fb.core.SyncPoint");
goog.require("fb.core.WriteTree");
goog.require("fb.core.util");
fb.core.ListenProvider;
fb.core.SyncTree = function(listenProvider) {
    this.syncPointTree_ = fb.core.util.ImmutableTree.Empty;
    this.pendingWriteTree_ = new fb.core.WriteTree;
    this.tagToQueryMap_ = {};
    this.queryToTagMap_ = {};
    this.listenProvider_ = listenProvider;
};
fb.core.SyncTree.prototype.applyUserOverwrite = function(path, newData, writeId, visible) {
    this.pendingWriteTree_.addOverwrite(path, newData, writeId, visible);
    if (!visible) {
        return[];
    } else {
        return this.applyOperationToSyncPoints_(new fb.core.operation.Overwrite(fb.core.OperationSource.User, path, newData));
    }
};
fb.core.SyncTree.prototype.applyUserMerge = function(path, changedChildren, writeId) {
    this.pendingWriteTree_.addMerge(path, changedChildren, writeId);
    var changeTree = fb.core.util.ImmutableTree.fromObject(changedChildren);
    return this.applyOperationToSyncPoints_(new fb.core.operation.Merge(fb.core.OperationSource.User, path, changeTree));
};
fb.core.SyncTree.prototype.ackUserWrite = function(writeId, revert) {
    revert = revert || false;
    var write = this.pendingWriteTree_.getWrite(writeId);
    var needToReevaluate = this.pendingWriteTree_.removeWrite(writeId);
    if (!needToReevaluate) {
        return[];
    } else {
        var affectedTree = fb.core.util.ImmutableTree.Empty;
        if (write.snap != null) {
            affectedTree = affectedTree.set(fb.core.util.Path.Empty, true);
        } else {
            fb.util.obj.foreach(write.children, function(pathString, node) {
                affectedTree = affectedTree.set(new fb.core.util.Path(pathString), node);
            });
        }
        return this.applyOperationToSyncPoints_(new fb.core.operation.AckUserWrite(write.path, affectedTree, revert));
    }
};
fb.core.SyncTree.prototype.applyServerOverwrite = function(path, newData) {
    return this.applyOperationToSyncPoints_(new fb.core.operation.Overwrite(fb.core.OperationSource.Server, path, newData));
};
fb.core.SyncTree.prototype.applyServerMerge = function(path, changedChildren) {
    var changeTree = fb.core.util.ImmutableTree.fromObject(changedChildren);
    return this.applyOperationToSyncPoints_(new fb.core.operation.Merge(fb.core.OperationSource.Server, path, changeTree));
};
fb.core.SyncTree.prototype.applyListenComplete = function(path) {
    return this.applyOperationToSyncPoints_(new fb.core.operation.ListenComplete(fb.core.OperationSource.Server, path));
};
fb.core.SyncTree.prototype.applyTaggedQueryOverwrite = function(path, snap, tag) {
    var queryKey = this.queryKeyForTag_(tag);
    if (queryKey != null) {
        var r = this.parseQueryKey_(queryKey);
        var queryPath = r.path, queryId = r.queryId;
        var relativePath = fb.core.util.Path.relativePath(queryPath, path);
        var op = new fb.core.operation.Overwrite(fb.core.OperationSource.forServerTaggedQuery(queryId), relativePath, snap);
        return this.applyTaggedOperation_(queryPath, queryId, op);
    } else {
        return[];
    }
};
fb.core.SyncTree.prototype.applyTaggedQueryMerge = function(path, changedChildren, tag) {
    var queryKey = this.queryKeyForTag_(tag);
    if (queryKey) {
        var r = this.parseQueryKey_(queryKey);
        var queryPath = r.path, queryId = r.queryId;
        var relativePath = fb.core.util.Path.relativePath(queryPath, path);
        var changeTree = fb.core.util.ImmutableTree.fromObject(changedChildren);
        var op = new fb.core.operation.Merge(fb.core.OperationSource.forServerTaggedQuery(queryId), relativePath, changeTree);
        return this.applyTaggedOperation_(queryPath, queryId, op);
    } else {
        return[];
    }
};
fb.core.SyncTree.prototype.applyTaggedListenComplete = function(path, tag) {
    var queryKey = this.queryKeyForTag_(tag);
    if (queryKey) {
        var r = this.parseQueryKey_(queryKey);
        var queryPath = r.path, queryId = r.queryId;
        var relativePath = fb.core.util.Path.relativePath(queryPath, path);
        var op = new fb.core.operation.ListenComplete(fb.core.OperationSource.forServerTaggedQuery(queryId), relativePath);
        return this.applyTaggedOperation_(queryPath, queryId, op);
    } else {
        return[];
    }
};
fb.core.SyncTree.prototype.addEventRegistration = function(query, eventRegistration) {
    var path = query.path;
    var serverCache = null;
    var foundAncestorDefaultView = false;
    this.syncPointTree_.foreachOnPath(path, function(pathToSyncPoint, sp) {
        var relativePath = fb.core.util.Path.relativePath(pathToSyncPoint, path);
        serverCache = serverCache || sp.getCompleteServerCache(relativePath);
        foundAncestorDefaultView = foundAncestorDefaultView || sp.hasCompleteView();
    });
    var syncPoint = this.syncPointTree_.get(path);
    if (!syncPoint) {
        syncPoint = new fb.core.SyncPoint;
        this.syncPointTree_ = this.syncPointTree_.set(path, syncPoint);
    } else {
        foundAncestorDefaultView = foundAncestorDefaultView || syncPoint.hasCompleteView();
        serverCache = serverCache || syncPoint.getCompleteServerCache(fb.core.util.Path.Empty);
    }
    var serverCacheComplete;
    if (serverCache != null) {
        serverCacheComplete = true;
    } else {
        serverCacheComplete = false;
        serverCache = fb.core.snap.EMPTY_NODE;
        var subtree = this.syncPointTree_.subtree(path);
        subtree.foreachChild(function(childName, childSyncPoint) {
            var completeCache = childSyncPoint.getCompleteServerCache(fb.core.util.Path.Empty);
            if (completeCache) {
                serverCache = serverCache.updateImmediateChild(childName, completeCache);
            }
        });
    }
    var viewAlreadyExists = syncPoint.viewExistsForQuery(query);
    if (!viewAlreadyExists && !query.getQueryParams().loadsAllData()) {
        var queryKey = this.makeQueryKey_(query);
        fb.core.util.assert(!goog.object.containsKey(this.queryToTagMap_, queryKey), "View does not exist, but we have a tag");
        var tag = fb.core.SyncTree.getNextQueryTag_();
        this.queryToTagMap_[queryKey] = tag;
        this.tagToQueryMap_["_" + tag] = queryKey;
    }
    var writesCache = this.pendingWriteTree_.childWrites(path);
    var events = syncPoint.addEventRegistration(query, eventRegistration, writesCache, serverCache, serverCacheComplete);
    if (!viewAlreadyExists && !foundAncestorDefaultView) {
        var view = (syncPoint.viewForQuery(query));
        events = events.concat(this.setupListener_(query, view));
    }
    return events;
};
fb.core.SyncTree.prototype.removeEventRegistration = function(query, eventRegistration, cancelError) {
    var path = query.path;
    var maybeSyncPoint = this.syncPointTree_.get(path);
    var cancelEvents = [];
    if (maybeSyncPoint && (query.queryIdentifier() === "default" || maybeSyncPoint.viewExistsForQuery(query))) {
        var removedAndEvents = maybeSyncPoint.removeEventRegistration(query, eventRegistration, cancelError);
        if (maybeSyncPoint.isEmpty()) {
            this.syncPointTree_ = this.syncPointTree_.remove(path);
        }
        var removed = removedAndEvents.removed;
        cancelEvents = removedAndEvents.events;
        var removingDefault = -1 !== goog.array.findIndex(removed, function(query) {
                return query.getQueryParams().loadsAllData();
            });
        var covered = this.syncPointTree_.findOnPath(path, function(relativePath, parentSyncPoint) {
            return parentSyncPoint.hasCompleteView();
        });
        if (removingDefault && !covered) {
            var subtree = this.syncPointTree_.subtree(path);
            if (!subtree.isEmpty()) {
                var newViews = this.collectDistinctViewsForSubTree_(subtree);
                for (var i = 0;i < newViews.length;++i) {
                    var view = newViews[i], newQuery = view.getQuery();
                    var listener = this.createListenerForView_(view);
                    this.listenProvider_.startListening(this.queryForListening_(newQuery), this.tagForQuery_(newQuery), listener.hashFn, listener.onComplete);
                }
            } else {
            }
        }
        if (!covered && removed.length > 0 && !cancelError) {
            if (removingDefault) {
                var defaultTag = null;
                this.listenProvider_.stopListening(this.queryForListening_(query), defaultTag);
            } else {
                var self = this;
                goog.array.forEach(removed, function(queryToRemove) {
                    var queryIdToRemove = queryToRemove.queryIdentifier();
                    var tagToRemove = self.queryToTagMap_[self.makeQueryKey_(queryToRemove)];
                    self.listenProvider_.stopListening(self.queryForListening_(queryToRemove), tagToRemove);
                });
            }
        }
        this.removeTags_(removed);
    } else {
    }
    return cancelEvents;
};
fb.core.SyncTree.prototype.calcCompleteEventCache = function(path, writeIdsToExclude) {
    var includeHiddenSets = true;
    var writeTree = this.pendingWriteTree_;
    var serverCache = this.syncPointTree_.findOnPath(path, function(pathSoFar, syncPoint) {
        var relativePath = fb.core.util.Path.relativePath(pathSoFar, path);
        var serverCache = syncPoint.getCompleteServerCache(relativePath);
        if (serverCache) {
            return serverCache;
        }
    });
    return writeTree.calcCompleteEventCache(path, serverCache, writeIdsToExclude, includeHiddenSets);
};
fb.core.SyncTree.prototype.collectDistinctViewsForSubTree_ = function(subtree) {
    return subtree.fold(function(relativePath, maybeChildSyncPoint, childMap) {
        if (maybeChildSyncPoint && maybeChildSyncPoint.hasCompleteView()) {
            var completeView = maybeChildSyncPoint.getCompleteView();
            return[completeView];
        } else {
            var views = [];
            if (maybeChildSyncPoint) {
                views = maybeChildSyncPoint.getQueryViews();
            }
            goog.object.forEach(childMap, function(childViews) {
                views = views.concat(childViews);
            });
            return views;
        }
    });
};
fb.core.SyncTree.prototype.removeTags_ = function(queries) {
    for (var j = 0;j < queries.length;++j) {
        var removedQuery = queries[j];
        if (!removedQuery.getQueryParams().loadsAllData()) {
            var removedQueryKey = this.makeQueryKey_(removedQuery);
            var removedQueryTag = this.queryToTagMap_[removedQueryKey];
            delete this.queryToTagMap_[removedQueryKey];
            delete this.tagToQueryMap_["_" + removedQueryTag];
        }
    }
};
fb.core.SyncTree.prototype.queryForListening_ = function(query) {
    if (query.getQueryParams().loadsAllData() && !query.getQueryParams().isDefault()) {
        return(query.ref());
    } else {
        return query;
    }
};
fb.core.SyncTree.prototype.setupListener_ = function(query, view) {
    var path = query.path;
    var tag = this.tagForQuery_(query);
    var listener = this.createListenerForView_(view);
    var events = this.listenProvider_.startListening(this.queryForListening_(query), tag, listener.hashFn, listener.onComplete);
    var subtree = this.syncPointTree_.subtree(path);
    if (tag) {
        fb.core.util.assert(!subtree.value.hasCompleteView(), "If we're adding a query, it shouldn't be shadowed");
    } else {
        var queriesToStop = subtree.fold(function(relativePath, maybeChildSyncPoint, childMap) {
            if (!relativePath.isEmpty() && maybeChildSyncPoint && maybeChildSyncPoint.hasCompleteView()) {
                return[maybeChildSyncPoint.getCompleteView().getQuery()];
            } else {
                var queries = [];
                if (maybeChildSyncPoint) {
                    queries = queries.concat(goog.array.map(maybeChildSyncPoint.getQueryViews(), function(view) {
                        return view.getQuery();
                    }));
                }
                goog.object.forEach(childMap, function(childQueries) {
                    queries = queries.concat(childQueries);
                });
                return queries;
            }
        });
        for (var i = 0;i < queriesToStop.length;++i) {
            var queryToStop = queriesToStop[i];
            this.listenProvider_.stopListening(this.queryForListening_(queryToStop), this.tagForQuery_(queryToStop));
        }
    }
    return events;
};
fb.core.SyncTree.prototype.createListenerForView_ = function(view) {
    var self = this;
    var query = view.getQuery();
    var tag = this.tagForQuery_(query);
    return{hashFn:function() {
        var cache = view.getServerCache() || fb.core.snap.EMPTY_NODE;
        return cache.hash();
    }, onComplete:function(status, data) {
        if (status === "ok") {
            if (tag) {
                return self.applyTaggedListenComplete(query.path, tag);
            } else {
                return self.applyListenComplete(query.path);
            }
        } else {
            var error = fb.core.util.errorForServerCode(status);
            return self.removeEventRegistration(query, null, error);
        }
    }};
};
fb.core.SyncTree.prototype.makeQueryKey_ = function(query) {
    return query.path.toString() + "$" + query.queryIdentifier();
};
fb.core.SyncTree.prototype.parseQueryKey_ = function(queryKey) {
    var splitIndex = queryKey.indexOf("$");
    fb.core.util.assert(splitIndex !== -1 && splitIndex < queryKey.length - 1, "Bad queryKey.");
    return{queryId:queryKey.substr(splitIndex + 1), path:new fb.core.util.Path(queryKey.substr(0, splitIndex))};
};
fb.core.SyncTree.prototype.queryKeyForTag_ = function(tag) {
    return goog.object.get(this.tagToQueryMap_, "_" + tag);
};
fb.core.SyncTree.prototype.tagForQuery_ = function(query) {
    var queryKey = this.makeQueryKey_(query);
    return fb.util.obj.get(this.queryToTagMap_, queryKey);
};
fb.core.SyncTree.nextQueryTag_ = 1;
fb.core.SyncTree.getNextQueryTag_ = function() {
    return fb.core.SyncTree.nextQueryTag_++;
};
fb.core.SyncTree.prototype.applyTaggedOperation_ = function(queryPath, queryId, operation) {
    var syncPoint = this.syncPointTree_.get(queryPath);
    fb.core.util.assert(syncPoint, "Missing sync point for query tag that we're tracking");
    var writesCache = this.pendingWriteTree_.childWrites(queryPath);
    return syncPoint.applyOperation(operation, writesCache, null);
};
fb.core.SyncTree.prototype.applyOperationToSyncPoints_ = function(operation) {
    return this.applyOperationHelper_(operation, this.syncPointTree_, null, this.pendingWriteTree_.childWrites(fb.core.util.Path.Empty));
};
fb.core.SyncTree.prototype.applyOperationHelper_ = function(operation, syncPointTree, serverCache, writesCache) {
    if (operation.path.isEmpty()) {
        return this.applyOperationDescendantsHelper_(operation, syncPointTree, serverCache, writesCache);
    } else {
        var syncPoint = syncPointTree.get(fb.core.util.Path.Empty);
        if (serverCache == null && syncPoint != null) {
            serverCache = syncPoint.getCompleteServerCache(fb.core.util.Path.Empty);
        }
        var events = [];
        var childName = operation.path.getFront();
        var childOperation = operation.operationForChild(childName);
        var childTree = syncPointTree.children.get(childName);
        if (childTree && childOperation) {
            var childServerCache = serverCache ? serverCache.getImmediateChild(childName) : null;
            var childWritesCache = writesCache.child(childName);
            events = events.concat(this.applyOperationHelper_(childOperation, childTree, childServerCache, childWritesCache));
        }
        if (syncPoint) {
            events = events.concat(syncPoint.applyOperation(operation, writesCache, serverCache));
        }
        return events;
    }
};
fb.core.SyncTree.prototype.applyOperationDescendantsHelper_ = function(operation, syncPointTree, serverCache, writesCache) {
    var syncPoint = syncPointTree.get(fb.core.util.Path.Empty);
    if (serverCache == null && syncPoint != null) {
        serverCache = syncPoint.getCompleteServerCache(fb.core.util.Path.Empty);
    }
    var events = [];
    var self = this;
    syncPointTree.children.inorderTraversal(function(childName, childTree) {
        var childServerCache = serverCache ? serverCache.getImmediateChild(childName) : null;
        var childWritesCache = writesCache.child(childName);
        var childOperation = operation.operationForChild(childName);
        if (childOperation) {
            events = events.concat(self.applyOperationDescendantsHelper_(childOperation, childTree, childServerCache, childWritesCache));
        }
    });
    if (syncPoint) {
        events = events.concat(syncPoint.applyOperation(operation, writesCache, serverCache));
    }
    return events;
};
