
import 'sync.dart';
import 'tree.dart';
import 'snap.dart';
import 'query.dart';

import 'dart:collection';
import 'dart:convert';

typedef void EventRunner();

/*
abstract class Event {

  final Change _change;
  Path get path;

  Event._(path, this._change);

  DataSnapshot get snapshot => new DataSnapshot(_change.snapshotNode);
  ChangeType get type => _change.type;

  EventRunner get eventRunner;
}
*/

abstract class Event {

  Path get path;
  final String eventType;

  Event(this.eventType);

  get eventRunner;

  toString();

}

abstract class EventRegistration {
  final CancelCallback _cancelCallback;

  EventRegistration._([this._cancelCallback]);

  factory EventRegistration(ChangeType type, EventCallback callback, [CancelCallback cancelCallback]) {
    switch (type) {
      case ChangeType.VALUE:
        return new ValueEventRegistration(callback, cancelCallback);
      default:
        return new ChildEventRegistration();
    }
  }

  cancel(Exception reason) {
    if (_cancelCallback==null) return;
    _cancelCallback(reason);
  }

  notify(Change change) {
    _callback(new Event._(change));
  }
}

class ValueEventRegistration extends EventRegistration {

  final _callback;

  ValueEventRegistration(this._callback, _cancelCallback) : super._(_cancelCallback);


  bool respondsTo(String eventType) => eventType=="value";

  Event createEvent(Change change, Query query) {
    return new DataEvent("value", this, new DataSnapshot(change.snapshotNode, query.ref, query.queryParams.index));
  }

  getEventRunner(Event eventData) {
    var ctx = this._context;
    if (eventData.eventType == "cancel") {
      assert(this
          ._cancelCallback); //, "Raising a cancel event on a listener with no cancel callback");
      var cancelCB = this._cancelCallback;
      return () {
        cancelCB.call(ctx, (eventData as CancelEvent).error);
      };
    } else {
      var cb = this._callback;
      return () {
        cb.call(ctx, (eventData as DataEvent).snapshot);
      };
    }
  }


  bool matches(EventRegistration other) {
    if (other is ValueEventRegistration) {
      if (other._callback==null || this._callback==null) {
        return true;
      } else {
        return other._callback == this._callback &&
            other._context == this._context;
      }
    } else {
      return false;
    }
  }

  bool get hasAnyCallback => _callback!=null;


}

class ChildEventRegistration extends EventRegistration {

  final Map<String, Function> _callbacks = {};

  ChildEventRegistration(_cancelCallback, _context)
      : super(_cancelCallback, _context);

  bool respondsTo(String eventType) {
    var eventToCheck = eventType == "children_added"
        ? "child_added"
        : eventType;
    eventToCheck =
    eventToCheck == "children_removed" ? "child_removed" : eventToCheck;
    return _callbacks.containsKey(eventToCheck);
  }

  Event createEvent(Change change, Query query) {
    assert(change.childName !=
        null); //, "Child events should have a childName.");
    var ref = query.ref.child((change.childName));
    var index = query.queryParams.index;
    return new DataEvent(
        change.type, this, new DataSnapshot(change.snapshotNode, ref, index),
        change.prevName);
  }

  getEventRunner(Event eventData) {
    var ctx = this._context;
    if (eventData is CancelEvent) {
      assert(this._cancelCallback !=
          null); //, "Raising a cancel event on a listener with no cancel callback");
      var cancelCB = this._cancelCallback;
      return () {
        cancelCB.call(ctx, eventData.error);
      };
    } else {
      var cb = this._callbacks[eventData.eventType];
      return () {
        cb.call(ctx, (eventData as DataEvent).snapshot,
            (eventData as DataEvent).prevName);
      };
    }
  }

  bool matches(EventRegistration other) {
    if (other is ChildEventRegistration) {
      if (this._callbacks == null || other._callbacks == null) {
        return true;
      } else {
        if (this._context == other._context) {
          var otherCount = other._callbacks.length;
          var thisCount = _callbacks.length;
          if (otherCount == thisCount) {
            if (otherCount == 1) {
              var otherKey = other._callbacks.keys.first;
              var thisKey = _callbacks.keys.first;
              return thisKey == otherKey &&
                  (other._callbacks[otherKey]==null || this._callbacks[thisKey]==null
                      ||
                      other._callbacks[otherKey] == this._callbacks[thisKey]);
            } else {
              return _callbacks.keys.every((k)=>other._callbacks[k]==_callbacks[k]);
            }
          }
        }
      }
    }
    return false;
  }


  bool get hasAnyCallback => _callbacks.isNotEmpty;

}


class DataEvent extends Event {

  final eventRegistration;
  final DataSnapshot snapshot;
  final String prevName;

  DataEvent(String eventType, this.eventRegistration, this.snapshot, [this.prevName]) : super(eventType);

  Path get path {
    var ref = this.snapshot.ref;
    if (this.eventType == "value") {
      return ref.path;
    } else {
      return ref.parent.path;
    }
  }

  get eventRunner => eventRegistration.getEventRunner(this);

  toString() => "$path:$eventType:${JSON.encode(snapshot.exportVal())}";
}


class CancelEvent extends Event {

  final eventRegistration;
  final error;
  final Path path;

  CancelEvent(this.eventRegistration, this.error, this.path) : super("cancel");

  get eventRunner => eventRegistration.getEventRunner(this);

  toString() => "$path:cancel";
}


class EventList extends UnmodifiableListView<Event> {

  final List<Event> _source;
  final Path path;

  EventList(Path path) : this._(path, []);
  EventList._(this.path, List<Event> source) : super(source), _source = source;

  @override
  void add(Event value) => value==null ? null : _source.add(value);

  void raise() {
    for (var eventData in _source) {
      eventData.eventRunner();
    }
    _source.clear();
  }

}

class EventQueue {
  final List<EventList> eventLists = [];

  int _recursionDepth = 0;

  queueEvents(Iterable<Event> events) {
    EventList currList = null;
    for (var eventData in events) {
      var eventPath = eventData.path;
      if (currList != null && eventPath!=currList.path) {
        this.eventLists.add(currList);
        currList = null;
      }
      if (currList == null) {
        currList = new EventList(eventPath);
      }
      currList.add(eventData);
    }
    if (currList!=null) {
      this.eventLists.add(currList);
    }
  }

  void raiseEventsAtPath(Path path, eventDataList) {
    this.queueEvents(eventDataList);
    this._raiseQueuedEventsMatchingPredicate((eventPath) {
      return eventPath == (path);
    });
  }

  void raiseEventsForChangedPath(Path changedPath, Iterable<Event> events) {
    print("raiseEventsForChangedPath $events");
    this.queueEvents(events);
    this._raiseQueuedEventsMatchingPredicate((eventPath) {
      return eventPath.contains(changedPath) || changedPath.contains(eventPath);
    });
  }

  void _raiseQueuedEventsMatchingPredicate(bool predicate(Path path)) {
    this._recursionDepth++;
    var sentAll = true;
    for (var eventList in eventLists) {
      if (eventList!=null) {
        var eventPath = eventList.path;
        if (predicate(eventPath)) {
          eventList.raise();
          this.eventLists[i] = null;
        } else {
          sentAll = false;
        }
      }
    }
    if (sentAll) {
      this.eventLists.clear();
    }
    this._recursionDepth--;
  }
}


class EventGenerator {

  final Query query;

  Index get index => query.queryParams.index;

  EventGenerator(this.query);

  List<Event> generateEventsForChanges(List<Change> changes, Node eventCache,
      List<EventRegistration> eventRegistrations) {
    var events = [];
    var moves = changes
        .where((change)=>change.type==ChangeType.CHILD_CHANGED &&index.indexedValueChanged(change.oldSnap, change.snapshotNode))
        .map((change)=>new Change.childMoved(change.childName, change.snapshotNode))
        .toList();

    this._generateEventsForType(events, ChangeType.CHILD_REMOVED, changes, eventRegistrations, eventCache);
    this._generateEventsForType(events, ChangeType.CHILD_ADDED, changes, eventRegistrations, eventCache);
    this._generateEventsForType(events, ChangeType.CHILD_MOVED, moves, eventRegistrations, eventCache);
    this._generateEventsForType(events, ChangeType.CHILD_CHANGED, changes, eventRegistrations, eventCache);
    this._generateEventsForType(events, ChangeType.VALUE, changes, eventRegistrations, eventCache);
    return events;
  }


  void _generateEventsForType(List<Event> events, ChangeType eventType,
      List<Change> changes, Iterable<EventRegistration> registrations, Node eventCache) { //TODO sync*
    var filteredChanges = changes.where((c)=>c.type==eventType).toList()..sort(_compareChanges);

    filteredChanges.forEach((change) {
      var materializedChange = _materializeSingleChange(change, eventCache);
      registrations.forEach((registration) {
        if (registration.respondsTo(change.type)) {
          events.add(registration.createEvent(materializedChange, query));
        }
      });
    });
  }

  _materializeSingleChange(Change change, eventCache) {
    if (change.type == ChangeType.VALUE || change.type == ChangeType.CHILD_REMOVED) {
      return change;
    } else {
      change.prevName = eventCache.getPredecessorChildName((change.childName),
          change.snapshotNode, index);
      return change;
    }
  }

  _compareChanges(Change a, Change b) {
    if (a.childName == null || b.childName == null) {
      throw new Exception("Should only compare child_ events.");
    }
    var aWrapped = new NamedNode(a.childName, a.snapshotNode);
    var bWrapped = new NamedNode(b.childName, b.snapshotNode);
    return this.index.compare(aWrapped, bWrapped);
  }
}
