
import 'snap.dart';
import 'tree.dart';
import 'query.dart';

abstract class NodeFilter {

  Node updateChild(Node snap, Name key, Node newChild, Path<Name> affectedPath);

  Node updateFullNode(Node oldSnap, Node newSnap);

  Node updatePriority(Node oldSnap, Node newPriority);

  bool get filtersNodes;

  IndexedFilter get indexedFilter;

  Index get index => indexedFilter.index;

}


class IndexedFilter extends NodeFilter {
  final Index index;

  IndexedFilter(this.index);

  @override
  Node updateChild(Node snap, Name key, Node newChild, Path<Name> affectedPath) {
    var oldChild = snap.children[key] ?? Node.nil;

    if (oldChild.subtree(affectedPath)==newChild.subtree(affectedPath)) {
      if (oldChild.isNil == newChild.isNil) {
        return snap;
      }
    }

    return snap.isLeaf && newChild.isNil
        ? snap : snap.updateChild(key, newChild).withIndex(index);
  }

  @override
  updateFullNode(Node oldSnap, Node newSnap) => newSnap?.withIndex(this.index);

  @override
  Node updatePriority(Node oldSnap, Node newPriority) =>
      oldSnap.isNil ? Node.nil : oldSnap.updatePriority(newPriority);

  bool get filtersNodes => false;

  IndexedFilter get indexedFilter => this;

}

class RangedFilter extends NodeFilter {

  final IndexedFilter indexedFilter;
  final Pair<Name,Node> startPost;
  final Pair<Name,Node> endPost;

  RangedFilter(QueryParams params) : indexedFilter = new IndexedFilter(params.index),
        startPost = _getStartPost(params), endPost = _getEndPost(params);

  bool matches(Pair<Name,Node> node) =>
      index.compare(startPost, node) <= 0 && index.compare(node, endPost) <= 0;


  @override
  updateChild(Node snap, Name key, Node newChild, Path affectedPath) {
    if (!this.matches(new Pair(key, newChild))) {
      newChild = Node.nil;
    }
    return indexedFilter.updateChild(snap, key, newChild, affectedPath);
  }

  @override
  updateFullNode(Node oldSnap, Node newSnap) {
    if (newSnap.isLeaf) {
      newSnap = Node.nil;
    }
    var filtered = newSnap
        .withIndex(index)
        .updatePriority(Node.nil);

    newSnap.children.forEach((key, childNode) {
      if (!matches(new Pair(key, childNode))) {
        filtered = filtered.updateChild(key, Node.nil);
      }
    });
    return indexedFilter.updateFullNode(oldSnap, filtered);
  }

  @override
  updatePriority(Node oldSnap, Node newPriority) => oldSnap;

  bool get filtersNodes => true;

  static _getStartPost(QueryParams params) {
    if (params.hasStart) {
      var startName = params.indexStartName;
      return params.index.makePost(params.indexStartValue, startName);
    } else {
      return params.index.minPost;
    }
  }

  static _getEndPost(QueryParams params) {
    if (params.hasEnd) {
      var endName = params.indexEndName;
      return params.index.makePost(params.indexEndValue, endName);
    } else {
      return params.index.maxPost;
    }
  }

}

class LimitedFilter extends NodeFilter {

  final RangedFilter rangedFilter;

  final int limit;
  final bool reverse;

  IndexedFilter get indexedFilter => rangedFilter.indexedFilter;

  LimitedFilter(QueryParams params) : rangedFilter = new RangedFilter(params),
        limit = params.limit, reverse = params.desc;

  @override
  Node updateChild(Node snap, Name key, Node newChild, Path affectedPath) {

    if (!rangedFilter.matches(new Pair(key, newChild))) {
      newChild = Node.nil;
    }
    if (snap[key]==newChild) {
      return snap;
    } else {
      if (snap.children.length < this.limit) {
        return indexedFilter.updateChild(snap, key, newChild, affectedPath);
      } else {
        return _fullLimitUpdateChild(snap, key, newChild);
      }
    }
  }

  @override
  Node updateFullNode(Node oldSnap, Node newSnap,
      [ChildChangeAccumulator changeAccumulator]) {
    var filtered;
    if (newSnap.isLeafNode || newSnap.isEmpty) {
      filtered = EMPTY_NODE.withIndex(index);
    } else {
      if (this.limit * 2 < newSnap.numChildren && newSnap.isIndexed(this.index)) {
        filtered = EMPTY_NODE.withIndex(index);
        var iterator;
        newSnap = (newSnap);
        if (this.reverse) {
          iterator = newSnap.reverseIteratorFrom(this.rangedFilter.endPost, index);
        } else {
          iterator = newSnap.iteratorFrom(rangedFilter.startPost, index);
        }
        var count = 0;
        while (iterator.hasNext() && count < this.limit) {
          var next = iterator.getNext();
          var inRange;
          if (this.reverse) {
            inRange = index.compare(this.rangedFilter.startPost, next) <= 0;
          } else {
            inRange = index.compare(next, this.rangedFilter.endPost) <= 0;
          }
          if (inRange) {
            filtered = filtered.updateImmediateChild(next.name, next.node);
            count++;
          } else {
            break;
          }
        }
      } else {
        filtered = newSnap.withIndex(this.index);
        filtered = (filtered.updatePriority(EMPTY_NODE));
        var startPost;
        var endPost;
        var cmp;
        var iterator;
        if (this.reverse) {
          iterator = filtered.getReverseIterator(this.index);
          startPost = this.rangedFilter.endPost;
          endPost = this.rangedFilter.startPost;
          var indexCompare = this.index.compare;
          cmp = (a, b) {
            return indexCompare(b, a);
          };
        } else {
          iterator = filtered.getIterator(this.index);
          startPost = this.rangedFilter.startPost;
          endPost = this.rangedFilter.endPost;
          cmp = this.index.compare;
        }
        var count = 0;
        var foundStartPost = false;
        var next;
        while (iterator.hasNext()) {
          next = iterator.getNext();
          if (!foundStartPost && cmp(startPost, next) <= 0) {
            foundStartPost = true;
          }
          var inRange = foundStartPost && count < this.limit && cmp(next, endPost) <= 0;
          if (inRange) {
            count++;
          } else {
            filtered = filtered.updateImmediateChild(next.name, EMPTY_NODE);
          }
        }
      }
    }
    return indexedFilter.updateFullNode(oldSnap, filtered, changeAccumulator);
  }

  @override
  Node updatePriority(Node oldSnap, Node newPriority) => oldSnap;

  bool get filtersNodes => true;

  Node _fullLimitUpdateChild(Node snap, Name childKey, Node childSnap) {
    var cmp = reverse ? (a,b)=>index.compare(b,a) : index.compare;

    assert(snap.children.length==limit);

    var newChildNamedNode = new Pair(childKey, childSnap);
    var windowBoundary = reverse ? snap.getFirstChild(this.index) :
    snap.getLastChild(this.index);

    var inRange = rangedFilter.matches(newChildNamedNode);

    if (snap.children.containsKey(childKey)) {
      var oldChildSnap = snap.children[childKey];

      var nextChild = source.getChildAfterChild(this.index, windowBoundary, this.reverse);
      while (nextChild != null && (nextChild.name == childKey || oldEventCache.hasChild(nextChild.name))) {
        nextChild = source.getChildAfterChild(this.index, nextChild, this.reverse);
      }
      var compareNext = nextChild == null ? 1 : cmp(nextChild, newChildNamedNode);
      var remainsInWindow = inRange && !childSnap.isEmpty && compareNext >= 0;
      if (remainsInWindow) {
        if (changeAccumulator != null) {
          changeAccumulator.trackChildChange(new Change.childChanged(childKey, childSnap, oldChildSnap));
        }
        return oldEventCache.updateImmediateChild(childKey, childSnap);
      } else {
        if (changeAccumulator != null) {
          changeAccumulator.trackChildChange(new Change.childRemoved(childKey, oldChildSnap));
        }
        var newEventCache = oldEventCache.updateImmediateChild(childKey, EMPTY_NODE);
        var nextChildInRange = nextChild != null && this.rangedFilter.matches(nextChild);
        if (nextChildInRange) {
          if (changeAccumulator != null) {
            changeAccumulator.trackChildChange(new Change.childAdded(nextChild.name, nextChild.node));
          }
          return newEventCache.updateImmediateChild(nextChild.name, nextChild.node);
        } else {
          return newEventCache;
        }
      }
    } else {
      if (childSnap.isEmpty) {
        return snap;
      } else {
        if (inRange) {
          if (cmp(windowBoundary, newChildNamedNode) >= 0) {
            if (changeAccumulator != null) {
              changeAccumulator.trackChildChange(new Change.childRemoved(windowBoundary.name, windowBoundary.node));
              changeAccumulator.trackChildChange(new Change.childAdded(childKey, childSnap));
            }
            return oldEventCache.updateImmediateChild(childKey, childSnap)
                .updateImmediateChild(windowBoundary.name, EMPTY_NODE);
          } else {
            return snap;
          }
        } else {
          return snap;
        }
      }
    }
  }

}

