
import 'snap.dart';
import 'tree.dart';
import 'sync.dart';
import 'filter.dart';

abstract class Operation {

  final OperationSource source;
  final Path<Name> path;

  Operation(this.source, this.path);

  Operation operationForChild(Name childName);

  OperationResult apply(View view, WriteTreeRef writesCahe);


  ViewCache _generateEventCacheAfterServerEvent(ViewCache viewCache, NodeFilter filter, WriteTreeRef writesCache) {

    var oldEventSnap = viewCache.eventCache;

    var newEventCache, serverNode;
    if (path.isEmpty) {
      assert(viewCache.serverCache.isFullyInitialized);//, "If change path is empty, we must have complete server data");
      if (viewCache.serverCache.isFiltered) {
        var serverCache = viewCache.completeServerSnap;
        var completeChildren = serverCache.isNotEmpty ? serverCache : Node.nil;
        var completeEventChildren = writesCache.calcCompleteEventChildren(completeChildren);
        newEventCache = filter.updateFullNode(viewCache.eventCache.node, completeEventChildren);
      } else {
        var completeNode = (writesCache.calcCompleteEventCache(viewCache.completeServerSnap));
        newEventCache = filter.updateFullNode(viewCache.eventCache.node, completeNode);
      }
    } else {
      var childKey = path.first;
      if (childKey == Name.priority) {
        assert(path.length == 1);//, "Can't have a priority with additional path components");
        var oldEventNode = oldEventSnap.node;
        serverNode = viewCache.serverCache.node;
        var updatedPriority = writesCache.calcEventCacheAfterServerOverwrite(path, oldEventNode, serverNode);
        if (updatedPriority != null) {
          newEventCache = filter.updatePriority(oldEventNode, updatedPriority);
        } else {
          newEventCache = oldEventSnap.node;
        }
      } else {
        var childChangePath = path.skip(1);
        var newEventChild;
        if (oldEventSnap.isCompleteForChild(childKey)) {
          serverNode = viewCache.serverCache.node;
          var eventChildUpdate = writesCache.calcEventCacheAfterServerOverwrite(
              path, oldEventSnap.node, serverNode);
          if (eventChildUpdate != null) {
            newEventChild = oldEventSnap.node.children[childKey]
                .updateSubtree(childChangePath, eventChildUpdate);
          } else {
            newEventChild = oldEventSnap.node.children[childKey];
          }
        } else {
          newEventChild = writesCache.calcCompleteChild(childKey, viewCache.serverCache);
        }
        if (newEventChild != null) {
          newEventCache = filter.updateChild(oldEventSnap.node, childKey,
              newEventChild, childChangePath);
        } else {
          newEventCache = oldEventSnap.node;
        }
      }
    }
    return viewCache.updateEventSnap(newEventCache,
        oldEventSnap.isFullyInitialized || path.isEmpty,
        filter.filtersNodes);
  }


}

enum OperationSource {user, server}

class Overwrite extends Operation {

  final Node snap;

  Overwrite(OperationSource source, Path<Name> path, this.snap) :
        super(source, path);


  @override
  Overwrite operationForChild(Name childName) {
    if (path.isEmpty) {
      return new Overwrite(source, Path.empty, snap.children[childName]);
    } else {
      return new Overwrite(source, path.skip(1), snap);
    }
  }

  @override
  OperationResult apply(View view, WriteTreeRef writesCache) {
    if (source==OperationSource.user) {
      return _applyUserOverwrite(view);
    } else {
      return _applyServerOverwrite(view, writesCache);
    }

  }

  OperationResult _applyServerOverwrite(View view, WriteTreeRef writesCache) {
    var filter = view.query.queryParams.nodeFilter;

    var oldViewCache = view.viewCache;
    var oldServerSnap = oldViewCache.serverCache;

    var newServerCache;
    if (path.isEmpty) {
      newServerCache = filter.updateFullNode(oldServerSnap.node, snap);
    } else if (filter.filtersNodes && !oldServerSnap.isFiltered) {
      var newServerNode = oldServerSnap.node.updateSubtree(path, snap);
      newServerCache = filter.updateFullNode(oldServerSnap.node, newServerNode);
    } else if (!oldServerSnap.isCompleteForPath(path) && path.length > 1) {
      newServerCache = oldServerSnap.node;
    } else {
      var childKey = path.first;

      var childChangePath = path.skip(1);
      var childNode = oldServerSnap.node.children[childKey] ?? Node.nil;
      var newChildNode = childNode.updateSubtree(childChangePath, snap);
      if (childKey == Name.priority) {
        newServerCache = filter.updatePriority(oldServerSnap.node, newChildNode);
      } else {
        newServerCache = filter?.updateChild(oldServerSnap.node, childKey,
            newChildNode, childChangePath);
      }
    }

    var newViewCache = oldViewCache.updateServerSnap(
        newServerCache,
        oldServerSnap.isFullyInitialized || path.isEmpty,
        filter.filtersNodes);

    newViewCache = _generateEventCacheAfterServerEvent(newViewCache, filter, writesCache);

    view.viewCache = newViewCache;

    return new OperationResult(
        oldViewCache,
        newViewCache
        );
  }

  OperationResult _applyUserOverwrite(View view) {

    var oldViewCache = view.viewCache;
    var filter = view.query.queryParams.nodeFilter;

    var oldEventSnap = oldViewCache.eventCache;
    var newViewCache, newEventCache;

    if (path.isEmpty) {
      newEventCache = filter.updateFullNode(oldViewCache.eventCache.node, snap);
      newViewCache = oldViewCache.updateEventSnap(newEventCache, true, filter.filtersNodes);
    } else {
      var childKey = path.first;
      if (childKey == Name.priority) {
        newEventCache = filter.updatePriority(oldViewCache.eventCache.node, snap);
        newViewCache = oldViewCache.updateEventSnap(newEventCache,
            oldEventSnap.isFullyInitialized, oldEventSnap.isFiltered);
      } else {
        var childChangePath = path.skip(1);
        var oldChild = oldEventSnap.node.children[childKey];
        var newChild;
        if (childChangePath.isEmpty) {
          newChild = snap;
        } else {
          var childNode = source.getCompleteChild(childKey);
          if (childNode != null) {
            if (childChangePath.last == Name.priority
                && childNode.getChild((childChangePath.parent)).isEmpty) {
              newChild = childNode;
            } else {
              newChild = childNode.updateChild(childChangePath, changedSnap);
            }
          } else {
            newChild = Node.nil;
          }
        }
        if (oldChild!=(newChild)) {
          var newEventSnap = this.filter.updateChild(oldEventSnap.node,
              childKey, newChild, childChangePath, source, accumulator);
          newViewCache = oldViewCache.updateEventSnap(newEventSnap,
              oldEventSnap.isFullyInitialized, this.filter.filtersNodes);
        } else {
          newViewCache = oldViewCache;
        }
      }
    }
    return newViewCache;
  }


  toString() => "Operation($path: $source overwrite: $snap)";
}


class ListenComplete extends Operation {

  ListenComplete(OperationSource source, Path<Name> path) :
        super(source, path);

  ListenComplete operationForChild(Name childName) {
    if (path.isEmpty) {
      return new ListenComplete(source, Path.empty);
    } else {
      return new ListenComplete(source, path.skip(1));
    }
  }

  toString() => "Operation($path: $source listen_complete)";


  ViewCache apply(ViewCache viewCache) {
    var oldServerNode = viewCache.serverCache;
    var newViewCache = viewCache.updateServerSnap(oldServerNode.node,
        oldServerNode.isFullyInitialized || path.isEmpty, oldServerNode.isFiltered);

    return this._generateEventCacheAfterServerEvent(newViewCache, path, writesCache,
        NO_COMPLETE_CHILD_SOURCE, accumulator);
  }

}

class Merge extends Operation {
  final ImmutableTree<Node> children;

  Merge(OperationSource source, Path<Name> path, this.children) :
        super(source, path);

  Operation operationForChild(Name childName) {
    if (this.path.isEmpty) {
      var childTree = this.children.subtree(NamePath.parse(childName));
      if (childTree.isEmpty) {
        return null;
      } else {
        if (childTree.value) {
          return new Overwrite(this.source, Path.empty, childTree.value);
        } else {
          return new Merge(this.source, Path.empty, childTree);
        }
      }
    } else {
      assert(this.path.first == childName);//, "Can't get a merge for a child not on the path of the operation");
      return new Merge(this.source, this.path.skip(1), this.children);
    }
  }

  toString() => "Operation($path: $source merge: $children)";

}

class AckUserWrite extends Operation {

  final ImmutableTree<Node> affectedTree;
  final bool revert;

  AckUserWrite(util.Path path, this.affectedTree, this.revert)
      : super(OperationSource.user, OperationType.ACK_USER_WRITE, path);

  operationForChild(String childName) {
    if (!this.path.isEmpty) {
      assert(this.path.front == childName);//, "operationForChild called for unrelated child.");
      return new AckUserWrite(this.path.popFront(), this.affectedTree, this.revert);
    } else {
      if (this.affectedTree.value != null) {
        assert(this.affectedTree.children.isEmpty);//, "affectedTree should not have overlapping affected paths.");
        return this;
      } else {
        var childTree = this.affectedTree.subtree(new util.Path.fromString(childName));
        return new AckUserWrite(util.Path.nil, childTree, this.revert);
      }
    }
  }

  toString() => "Operation($path: $source ack_write revert=$revert affectedTree=$affectedTree)";

}

class OperationResult {
  final ViewCache newViewCache;
  final ViewCache oldViewCache;

  OperationResult(this.oldViewCache, this.newViewCache);

  Change get valueChange {
    var eventSnap = newViewCache.eventCache;
    if (eventSnap.isFullyInitialized) {
      var isLeafOrNil = eventSnap.node.isLeaf || eventSnap.node.isNil;
      var oldCompleteSnap = oldViewCache.completeEventSnap;
      if (!oldViewCache.eventCache.isFullyInitialized
          || isLeafOrNil && eventSnap.node!=oldCompleteSnap
          || eventSnap.node.priority!=oldCompleteSnap.priority) {
        return new Change.value((newViewCache.completeEventSnap));
      }
    }
    return null;
  }
}

class ViewProcessor {
  final NodeFilter filter;

  const ViewProcessor(this.filter);

  OperationResult applyOperation(ViewCache oldViewCache, Operation operation) {
    var newViewCache = operation.apply(oldViewCache);

    return new OperationResult(oldViewCache, newViewCache);

/*
    var accumulator = new ChildChangeAccumulator();
    var newViewCache, filterServerNode;


    if (operation is Overwrite) {
    } else {
      if (operation is Merge) {
        var merge = (operation);
        if (merge.source.fromUser) {
          newViewCache = this._applyUserMerge(oldViewCache, merge.path,
              merge.children, writesCache, completeCache, accumulator);
        } else {
          assert(merge.source.fromServer);//, "Unknown source.");
          filterServerNode = merge.source.tagged || oldViewCache.serverCache.isFiltered;
          newViewCache = this._applyServerMerge(oldViewCache, merge.path,
              merge.children, writesCache, completeCache, filterServerNode, accumulator);
        }
      } else {
        if (operation is AckUserWrite) {
          var ackUserWrite = (operation);
          if (!ackUserWrite.revert) {
            newViewCache = this._ackUserWrite(oldViewCache, ackUserWrite.path,
                ackUserWrite.affectedTree, writesCache, completeCache, accumulator);
          } else {
            newViewCache = this._revertUserWrite(oldViewCache, ackUserWrite.path,
                writesCache, completeCache, accumulator);
          }
        } else {
          if (operation.type == OperationType.LISTEN_COMPLETE) {
            newViewCache = this._listenComplete(oldViewCache, operation.path,
                writesCache, completeCache, accumulator);
          } else {
            throw new Exception("Unknown operation type: $operation");
          }
        }
      }
    }
    var changes = accumulator.changes.toList();
    this._maybeAddValueEvent(oldViewCache, newViewCache, changes);
*/
  }

  void _maybeAddValueEvent(ViewCache oldViewCache, ViewCache newViewCache, List<Change> accumulator) {
    var eventSnap = newViewCache.eventCache;
    if (eventSnap.isFullyInitialized) {
      var isLeafOrEmpty = eventSnap.node.isLeaf || eventSnap.node.isEmpty;
      var oldCompleteSnap = oldViewCache.completeEventSnap;
      if (accumulator.length > 0 || !oldViewCache.eventCache.isFullyInitialized
          || isLeafOrEmpty && eventSnap.node!=((oldCompleteSnap))
          || eventSnap.node.priority!=(oldCompleteSnap.priority)) {
        accumulator.add(new Change.value((newViewCache.completeEventSnap)));
      }
    }
  }


  ViewCache _applyServerOverwrite(ViewCache oldViewCache, Path changePath,
      Node changedSnap, WriteTreeRef writesCache, Node completeCache,
      bool filterServerNode, ChildChangeAccumulator accumulator) {
    var oldServerSnap = oldViewCache.serverCache;
    var newServerCache;
    var serverFilter = filterServerNode ? this.filter : this.filter.indexedFilter;
    if (changePath.isEmpty) {
      newServerCache = serverFilter.updateFullNode(oldServerSnap.node, changedSnap, null);
    } else {
      if (serverFilter.filtersNodes && !oldServerSnap.isFiltered) {
        var newServerNode = oldServerSnap.node.updateChild(changePath, changedSnap);
        newServerCache = serverFilter.updateFullNode(oldServerSnap.node, newServerNode, null);
      } else {
        var childKey = changePath.front;
        if (!oldServerSnap.isCompleteForPath(changePath) && changePath.length > 1) {
          return oldViewCache;
        }
        var childChangePath = changePath.popFront();
        var childNode = oldServerSnap.node.getImmediateChild(childKey);
        var newChildNode = childNode.updateChild(childChangePath, changedSnap);
        if (childKey == ".priority") {
          newServerCache = serverFilter.updatePriority(oldServerSnap.node, newChildNode);
        } else {
          newServerCache = serverFilter.updateChild(oldServerSnap.node, childKey,
              newChildNode, childChangePath, NO_COMPLETE_CHILD_SOURCE, null);
        }
      }
    }
    var newViewCache = oldViewCache.updateServerSnap(newServerCache,
        oldServerSnap.isFullyInitialized || changePath.isEmpty, serverFilter.filtersNodes);
    var source = new WriteTreeCompleteChildSource(writesCache, newViewCache, completeCache);
    return this._generateEventCacheAfterServerEvent(newViewCache, changePath, writesCache, source, accumulator);
  }


  static _cacheHasChild(ViewCache viewCache, String childKey) {
    return viewCache.eventCache.isCompleteForChild(childKey);
  }

  ViewCache _applyUserMerge(ViewCache viewCache, Path path,
      UnmodifiableTree<Name,Node> changedChildren, WriteTreeRef writesCache, Node serverCache, ChildChangeAccumulator accumulator) {
    var curViewCache = viewCache;
    changedChildren.forEach((relativePath, childNode) {
      var writePath = path.child(relativePath);
      if (_cacheHasChild(viewCache, writePath.front)) {
        curViewCache = _applyUserOverwrite(curViewCache, writePath, childNode, writesCache, serverCache, accumulator);
      }
    });
    changedChildren.forEach((relativePath, childNode) {
      var writePath = path.child(relativePath);
      if (!_cacheHasChild(viewCache, writePath.front)) {
        curViewCache = _applyUserOverwrite(curViewCache, writePath, childNode, writesCache, serverCache, accumulator);
      }
    });
    return curViewCache;
  }

  Node _applyMerge(Node node, Map<Path, Node> merge) {
    merge.forEach((relativePath, childNode) {
      node = node.updateChild(relativePath, childNode);
    });
    return node;
  }

  ViewCache _applyServerMerge(ViewCache viewCache, Path path, changedChildren,
      WriteTreeRef writesCache, Node serverCache, bool filterServerNode, ChildChangeAccumulator accumulator) {
    if (viewCache.serverCache.node.isEmpty && !viewCache.serverCache.isFullyInitialized) {
      return viewCache;
    }
    var curViewCache = viewCache;
    var viewMergeTree;
    if (path.isEmpty) {
      viewMergeTree = changedChildren;
    } else {
      viewMergeTree = ImmutableTree.nil.setTree(path, changedChildren);
    }
    var serverNode = viewCache.serverCache.node;
    var self = this;
    viewMergeTree.children.inorderTraversal((childKey, childTree) {
      if (serverNode.hasChild(childKey)) {
        var serverChild = viewCache.serverCache.node.getImmediateChild(childKey);
        var newChild = self._applyMerge(serverChild, childTree);
        curViewCache = self._applyServerOverwrite(curViewCache,
            new util.Path.fromString(childKey), newChild, writesCache, serverCache, filterServerNode, accumulator);
      }
    });
    viewMergeTree.children.inorderTraversal((childKey, childMergeTree) {
      var isUnknownDeepMerge = !viewCache.serverCache.isCompleteForChild(childKey) && childMergeTree.value == null;
      if (!serverNode.hasChild(childKey) && !isUnknownDeepMerge) {
        var serverChild = viewCache.serverCache.node.getImmediateChild(childKey);
        var newChild = self._applyMerge(serverChild, childMergeTree);
        curViewCache = self._applyServerOverwrite(curViewCache, new util.Path.fromString(childKey), newChild, writesCache, serverCache, filterServerNode, accumulator);
      }
    });
    return curViewCache;
  }

  ViewCache _ackUserWrite(ViewCache viewCache, Path ackPath, UnmodifiableTree<Name,Node> affectedTree,
      WriteTreeRef writesCache, Node completeCache, ChildChangeAccumulator accumulator) {
    if (writesCache.shadowingWrite(ackPath) != null) {
      return viewCache;
    }
    var filterServerNode = viewCache.serverCache.isFiltered;
    var serverCache = viewCache.serverCache;
    if (affectedTree.value != null) {
      if (ackPath.isEmpty && serverCache.isFullyInitialized || serverCache.isCompleteForPath(ackPath)) {
        return this._applyServerOverwrite(viewCache, ackPath,
            serverCache.node.getChild(ackPath), writesCache, completeCache, filterServerNode, accumulator);
      } else {
        if (ackPath.isEmpty) {
          var changedChildren = ImmutableTree.nil;
          serverCache.node.forEachChild(keyIndex, (name, node) {
            changedChildren = changedChildren.set(new util.Path.fromString(name), node);
          });
          return this._applyServerMerge(viewCache, ackPath, changedChildren,
              writesCache, completeCache, filterServerNode, accumulator);
        } else {
          return viewCache;
        }
      }
    } else {
      var changedChildren = ImmutableTree.nil;
      affectedTree.forEach((mergePath, value) {
        var serverCachePath = ackPath.child(mergePath);
        if (serverCache.isCompleteForPath(serverCachePath)) {
          changedChildren = changedChildren.set(mergePath, serverCache.node.getChild(serverCachePath));
        }
      });
      return this._applyServerMerge(viewCache, ackPath, changedChildren, writesCache, completeCache, filterServerNode, accumulator);
    }
  }

  ViewCache _revertUserWrite(ViewCache viewCache, Path path, WriteTreeRef writesCache,
      Node completeServerCache, ChildChangeAccumulator accumulator) {
    var complete;
    if (writesCache.shadowingWrite(path) != null) {
      return viewCache;
    } else {
      var source = new WriteTreeCompleteChildSource(writesCache, viewCache, completeServerCache);
      var oldEventCache = viewCache.eventCache.node;
      var newEventCache;
      if (path.isEmpty || path.front == ".priority") {
        var newNode;
        if (viewCache.serverCache.isFullyInitialized) {
          newNode = writesCache.calcCompleteEventCache(viewCache.completeServerSnap);
        } else {
          var serverChildren = viewCache.serverCache.node;
          assert(serverChildren is ChildrenNode);//, "serverChildren would be complete if leaf node");
          newNode = writesCache.calcCompleteEventChildren((serverChildren));
        }
        newNode = (newNode);
        newEventCache = this.filter.updateFullNode(oldEventCache, newNode, accumulator);
      } else {
        var childKey = path.front;
        var newChild = writesCache.calcCompleteChild(childKey, viewCache.serverCache);
        if (newChild == null && viewCache.serverCache.isCompleteForChild(childKey)) {
          newChild = oldEventCache.getImmediateChild(childKey);
        }
        if (newChild != null) {
          newEventCache = this.filter.updateChild(oldEventCache, childKey,
              newChild, path.popFront(), source, accumulator);
        } else {
          if (viewCache.eventCache.node.hasChild(childKey)) {
            newEventCache = this.filter.updateChild(oldEventCache, childKey,
                EMPTY_NODE, path.popFront(), source, accumulator);
          } else {
            newEventCache = oldEventCache;
          }
        }
        if (newEventCache.isEmpty() && viewCache.serverCache.isFullyInitialized) {
          complete = writesCache.calcCompleteEventCache(viewCache.completeServerSnap);
          if (complete.isLeafNode) {
            newEventCache = this.filter.updateFullNode(newEventCache, complete, accumulator);
          }
        }
      }
      complete = viewCache.serverCache.isFullyInitialized || writesCache.shadowingWrite(util.Path.nil) != null;
      return viewCache.updateEventSnap(newEventCache, complete, this.filter.filtersNodes);
    }
  }

  ViewCache _listenComplete(ViewCache viewCache, Path path, WriteTreeRef writesCache,
      Node serverCache, ChildChangeAccumulator accumulator) {
    var oldServerNode = viewCache.serverCache;
    var newViewCache = viewCache.updateServerSnap(oldServerNode.node,
        oldServerNode.isFullyInitialized || path.isEmpty, oldServerNode.isFiltered);
    return this._generateEventCacheAfterServerEvent(newViewCache, path, writesCache,
        NO_COMPLETE_CHILD_SOURCE, accumulator);
  }

}
