
import 'tree.dart';
import 'snap.dart';
import 'dart:async';
import 'sync.dart';
import 'connection.dart';
import 'filter.dart';
import 'package:quiver/core.dart' as quiver;
import 'pointer.dart';
import 'events.dart';

class DataSnapshot {
  final Node node;

  DataSnapshot(this.node);

  dynamic get val => node.toJson();
}

class RepoInfo {

  final Uri uri;

  RepoInfo(this.uri);

  int get hashCode => uri.hashCode;
  bool operator==(other) => other is RepoInfo&&other.uri==uri;

}



class Repo {
  final RepoInfo info;
  Backend _backend;
  final EventQueue eventQueue = new EventQueue();

  Backend get backend => _backend;

  static final PointerMap<RepoInfo, SyncTree> _syncTrees = new PointerMap();

  Repo(Uri uri) : info = new RepoInfo(uri.resolve("/")) {
    var uri = info.uri;
    switch (uri.scheme) {
      case "mem":
        _backend = new MemoryBackend(uri.host);
        break;
      default:
        throw new UnimplementedError(
            "Backend for scheme ${uri.scheme} not implemented");
    }
  }

  Pointer<SyncTree> get _syncTree => _syncTrees.putIfAbsent(info, () => new SyncTree(this));

  Future<DataSnapshot> setWithPriority(Path<Name> path, value, priority) {
    var node = new Node.fromJson(value, priority);
    return _backend.put(path, node).then((n) => new DataSnapshot(n));
  }

  int get hashCode => info.hashCode;
  bool operator==(other) => other is Repo&&other.info==info;
}

class QueryParams {

  final int limit;
  final bool desc;
  final Index index;
  final startValue;
  final endValue;

  const QueryParams({this.limit, this.desc: false, this.index: Index.priority, this.startValue, this.endValue});

  int get hashCode => quiver.hashObjects([limit, desc, index, startValue, endValue]);
  bool operator==(other) => other is QueryParams&&other.limit==limit&&other.desc==desc
      &&other.index==index&&other.startValue==startValue&&other.endValue==endValue;

  QueryParams _limit(int newLimit, [bool desc = false]) {
    if (limit <= 0) {
      throw new ArgumentError("Limit must be a positive integer.");
    }
    if (this.limit!=null) {
      throw new StateError("Limit was already set (by another call to limitToFirst, or limitToLast).");
    }
    return new QueryParams(
        limit: newLimit,
        desc: desc,
        index: index,
        startValue: startValue,
        endValue: endValue
    );
  }

  QueryParams limitToFirst(int newLimit) => _limit(newLimit, false);

  QueryParams limitToLast(int newLimit) => _limit(newLimit, true);

  QueryParams startAt(value) {
    return new QueryParams(
        limit: limit,
        desc: desc,
        index: index,
        startValue: value,
        endValue: endValue
    );
  }

  QueryParams endAt(value) {
    return new QueryParams(
        limit: limit,
        desc: desc,
        index: index,
        startValue: startValue,
        endValue: value
    );
  }

  QueryParams orderBy(Index index) {
    return new QueryParams(
        limit: limit,
        desc: desc,
        index: index,
        startValue: startValue,
        endValue: endValue
    );
  }

  bool get loadsAllData => startValue==null && endValue==null && limit==null;

  bool get hasStart => startValue!=null;
  bool get hasEnd => endValue!=null;

  bool get isDefault => loadsAllData && index == Index.priority;

  NodeFilter get nodeFilter {
    if (this.loadsAllData) {
      return new IndexedFilter(this.index);
    } else {
      if (this.hasLimit) {
        return new LimitedFilter(this);
      } else {
        return new RangedFilter(this);
      }
    }
  }
}



class Query {
  final Repo repo;
  final Path path;
  final QueryParams queryParams;

  const Query(this.repo, this.path, this.queryParams);

  Firebase get ref => new Firebase.fromRepo(this.repo, this.path);

  int get hashCode => quiver.hash3(repo, path, queryParams);

  bool operator==(other) => other is Query && other.repo==repo && other.path==path && other.queryParams==queryParams;


/*
  Stream<Event> get onValue;
  Stream<Event> get onChildAdded;
  Stream<Event> get onChildChanged;
  Stream<Event> get onChildMoved;
  Stream<Event> get onChildRemoved;
*/

  on(ChangeType type, EventCallback callback) {
    print("on $path $type");
    var syncTree = repo._syncTree;
    syncTree.value.addEventRegistration(this,
        new EventRegistration(type, callback, (_) => syncTree.release()));
  }


  Query limitToFirst(int limit) => new Query(repo, path, queryParams.limitToFirst(limit));

  Query limitToLast(int limit) => new Query(repo, path, queryParams.limitToLast(limit));

  Query orderByChild(String path) => _orderBy(new PathIndex(NamePath.parse(path)));

  Query orderByKey() => _orderBy(Index.key);

  Query orderByPriority() => _orderBy(Index.priority);

  Query orderByValue() => _orderBy(Index.value);

  Query _orderBy(Index index) =>
      new Query(repo, path, queryParams.orderBy(index));


  Query startAt(value) => new Query(repo, path, queryParams.startAt(value));
  Query endAt(value) => new Query(repo, path, queryParams.endAt(value));
  Query equalTo(value) => new Query(repo, path, queryParams.startAt(value).endAt(value));

}

class Firebase extends Query {
  const Firebase.fromRepo(Repo repo, [Path path])
      : super(repo, path, const QueryParams());

  factory Firebase(Uri url) => new Firebase.fromRepo(new Repo(url), NamePath.parse(url.path));

  Firebase child(String path) => new Firebase.fromRepo(repo, this.path.append(NamePath.parse(path)));

  Firebase get parent => path.parent==null ? null : new Firebase.fromRepo(repo, path.parent);

  Firebase get root {
    var ref = this;
    while (ref.parent != null) {
      ref = ref.parent;
    }
    return ref;
  }

  Future<DataSnapshot> set(newVal) => repo.setWithPriority(path, newVal, null);

  Future<DataSnapshot> setWithPriority(newVal, newPriority) =>
      repo.setWithPriority(path, newVal, newPriority);

  Future<DataSnapshot> remove() => set(null);

}


main() async {
  var fredNameRef = new Firebase(Uri.parse('mem://docs-examples.firebaseio.com/samplechat/users/fred/name'));

  fredNameRef.child('first').on(ChangeType.VALUE, (change) {
    print("change1 ${change.type} ${change.snapshot.val}");
  });

  fredNameRef.child('first').on(ChangeType.VALUE, (change) {
    print("change2 ${change.type} ${change.snapshot.val}");
  });

  fredNameRef = new Firebase(Uri.parse('mem://docs-examples.firebaseio.com/samplechat/users/fred/name'));

  fredNameRef.child('first').on(ChangeType.VALUE, (change) {
    print("change3 ${change.type} ${change.snapshot.val}");
  });

  fredNameRef.on(ChangeType.VALUE, (change) {
    print("change4 ${change.type} ${change.snapshot.val}");
  });

  print("setting first to Fred");
  await fredNameRef.child('first').set('Fred');

  print("setting last to Flintstone");
  await fredNameRef.child('last').set('Flintstone');

  print("setting to null");
  await fredNameRef.set(null);

  print('setting to {"first": "Fred"}');
  await fredNameRef.set({
    "first": "Fred"
  });

  await new Future.delayed(new Duration(seconds: 1));

}