
import 'tree.dart';
import 'dart:convert';
import 'dart:collection';

class Name implements Pattern, Comparable<Name> {

  static const priority = const Name(".priority");
  static const min = const Name("[MIN_NAME]");
  static const max = const Name("[MAX_NAME]");

  final String _value;

  const Name(this._value);

  toString() => _value;

  @override
  Iterable<Match> allMatches(String string, [int start = 0]) =>
      _value.allMatches(string, start);

  @override
  Match matchAsPrefix(String string, [int start = 0]) =>
      _value.matchAsPrefix(string, start);

  int get hashCode => _value.hashCode;
  bool operator==(other) => other is Name&&other._value==_value;

  @override
  int compareTo(Name other) => compare(this, other);


  asInt() => int.parse(_value, onError: (_)=> null);

  int get length => _value.length;

  static int compare(Name a, Name b) {
    if (a == b) {
      return 0;
    } else {
      if (a == min || b == max) {
        return-1;
      }
      if (b == min || a == max) {
        return 1;
      }

      var aAsInt = a.asInt();
      var bAsInt = b.asInt();
      if (aAsInt != null) {
        if (bAsInt != null) {
          return aAsInt - bAsInt == 0 ? a.length - b.length : aAsInt - bAsInt;
        } else {
          return-1;
        }
      } else {
        if (bAsInt != null) {
          return 1;
        } else {
          return a.compareTo(b);
        }
      }
    }
  }
}

class NamePath extends Path<Name> {
  static NamePath parse(String str) => new NamePath.fromString(str);

  NamePath(Iterable<Name> pieces) : super(pieces);

  factory NamePath.fromString(String str) {
    var pieces = str.split("/").where((v)=>v.isNotEmpty).map((p)=>new Name(p));
    return new NamePath(pieces);
  }
}

abstract class Index {

  static const priority = const _PriorityIndex();
  static const key = const _KeyIndex();
  static const value = const _ValueIndex();

  const Index();

  int compare(Pair<Name, Node> a, Pair<Name, Node> b);
}

class PathIndex extends Index {

  final Path _indexPath;

  const PathIndex(this._indexPath);

  int compare(Pair<Name, Node> a, Pair<Name, Node> b) {
    var aChild = a.value.subtree(_indexPath);
    var bChild = b.value.subtree(_indexPath);
    var indexCmp = aChild.compareTo(bChild);
    if (indexCmp == 0) {
      return Name.compare(a.key, b.key);
    } else {
      return indexCmp;
    }
  }

  int get hashCode => _indexPath.hashCode;
  bool operator==(other) => other is PathIndex&&other._indexPath==_indexPath;
}

class _PriorityIndex extends Index {
  const _PriorityIndex();

  int compare(Pair<Name, Node> a, Pair<Name, Node> b) {
    var aPriority = a.value.priority;
    var bPriority = b.value.priority;
    var indexCmp = aPriority.compareTo(bPriority);
    if (indexCmp == 0) {
      return Name.compare(a.key, b.key);
    } else {
      return indexCmp;
    }
  }
}

class _KeyIndex extends Index {

  const _KeyIndex();

  int compare(Pair<Name, Node> a, Pair<Name, Node> b) => Name.compare(a.key, b.key);
}

class _ValueIndex extends Index {

  const _ValueIndex();

  int compare(Pair<Name, Node> a, Pair<Name, Node> b) {
    var indexCmp = a.value.compareTo(b.value);
    if (indexCmp == 0) {
      return Name.compare(a.key, b.key);
    } else {
      return indexCmp;
    }
  }
}

class ServerValue {
  final String type;
  const ServerValue._(this.type);

  static const timestamp = const ServerValue._("TIMESTAMP");

  static const values = const {
    "TIMESTAMP": timestamp
  };

  toJson() => {".sv": type};
}

class Value implements Comparable<Value> {

  final dynamic value;

  const Value._(this.value);
  const Value.bool(bool value) : this._(value);
  const Value.num(num value) : this._(value);
  const Value.string(String value) : this._(value);
  Value.server(String type) : this._(ServerValue.values[type]);

  factory Value(dynamic value) {
    if (value is bool) return new Value.bool(value);
    if (value is num) return new Value.num(value);
    if (value is String) return new Value.string(value);
    if (value is Map&&value.containsKey(".sv")) return new Value.server(value[".sv"]);
    throw new ArgumentError("Unsupported value type ${value.runtimeType}");
  }

  bool get isBool => value is bool;
  bool get isNum => value is num;
  bool get isString => value is String;
  bool get isServerValue => value is ServerValue;

  int get typeOrder => isServerValue ? 0 : isBool ? 1 : isNum ? 2 : isString ? 3 : 4;

  @override
  int compareTo(Value other) {
    var thisIndex = typeOrder;
    var otherIndex = other.typeOrder;

    if (otherIndex == thisIndex) {
      if (isServerValue) return 0;
      return Comparable.compare(value, other.value);
    }
    return thisIndex - otherIndex;
  }

  int get hashCode => value.hashCode;
  bool operator==(other) => other is Value&&value==other.value;

  toJson() => value;
}

class _NodeChildrenMap<K,V> extends MapView<K,V> {
  _NodeChildrenMap(Map<K,V> map) : super(map);

  @override
  V operator [](Object key) => super[key] ?? Node.nil;

}



class Node extends UnmodifiableTree<Name,Value> implements Comparable<Node> {

  static final nil = new Node._();
  static final max = new Node._();

  Node._() : this.leaf(null, null);
  Node.leaf(value, [this.priority]) : super(value==null ? null : new Value(value), new _NodeChildrenMap(const{}));
  Node.nonLeaf(Map<Name,Node> children, [this.priority]) : super(null, new _NodeChildrenMap(children));

  factory Node.fromJson(json, [priority]) {
    if (json == null) {
      return Node.nil;
    }
    if (json is Map && json.containsKey(".priority")) {
      priority = json[".priority"];
    }
    priority = priority==null ? null : new Node.leaf(priority);

    if (json is Map && json.containsKey(".value") && json[".value"] != null) {
      json = json[".value"];
    }

    if (json is !Map || json.containsKey(".sv") ) {
      return new Node.leaf(json, priority);
    }

    var children = new Map.fromIterable(
        json.keys.where((k)=>!k.startsWith(".")).map((k)=>new Name(k)),
        value: (k)=>new Node.fromJson(json[k]) );

    return new Node.nonLeaf(children, priority);
  }

  bool get isLeaf => children.isEmpty&&value!=null;
  bool get isNil => this==nil;
  bool get isMax => this==max;

  final Node priority;

  Map<Name, Node> get children => super.children;
  Node subtree(Path<Name> path) => super.subtree(path) ?? Node.nil;

  String toString() => JSON.encode(toJson());

  Iterable<Pair<Name,Node>> get orderedByPriority;

  dynamic toJson([bool exportFormat = false]) {
    if (isLeaf) {
      if (exportFormat && !this.priority.isEmpty) {
        return {
          ".value":value.toJson(),
          ".priority":this.priority.toJson()
        };
      } else {
        return this.value.toJson();
      }
    }
    if (isEmpty) {
      return null;
    }
    var obj = new Map.fromIterables(
        children.keys.map((k)=>k.toString()),
        children.values.map((v)=>v.toJson(exportFormat))
    );
    if (exportFormat && !this.priority.isEmpty) {
      obj[".priority"] = this.priority.toJson();
    }
    return obj;
  }

  int compareTo(Node other) {
    if (isLeaf) {
      if (other.isLeaf) {
        return value.compareTo(other.value);
      }
      if (other.isEmpty) {
        return 1;
      }
      return -1;
    }
    if (isEmpty) {
      return other.isEmpty&&!other.isLeaf ? 0 : -1;
    } else {
      if (other.isEmpty) {
        return 1;
      } else {
        return other == max ?  -1 : 0;
      }
    }
  }

  Node updateSubtree(Path<Name> path, Node subtree) {
    if (path.isEmpty) return subtree;
    var children = new Map.from(this.children);
    children[path.first] = (this.children[path.first] ?? Node.nil).updateSubtree(path.skip(1), subtree);
    return new Node.nonLeaf(children, priority);
  }

  Node updateChild(Name childName, Node child) =>
      updateSubtree(new Path([childName]), child);

  Node updatePriority(Node priority) =>
      isLeaf ? new Node.leaf(value, priority) : new Node.nonLeaf(children, priority);

  Node withIndex(Index index) => this; //TODO
}
