
import 'query.dart';
import 'tree.dart';
import 'dart:collection';
import 'snap.dart';
import 'filter.dart';
import 'operation.dart';
import 'connection.dart';
import 'dart:async';
import 'events.dart';

/**
 * A CacheNode contains the partially known value of a node in a tree.
 */
class CacheNode {

  /**
   * The partially known value
   */
  final Node node;

  /**
   * Has the value been fully initialized
   */
  final bool isFullyInitialized;

  /**
   * When true the value is filtered and therefore not complete
   */
  final bool isFiltered;

  const CacheNode(this.node, this.isFullyInitialized, this.isFiltered);

  /**
   * Checks if the subvalue for a particular path is complete
   */
  bool isCompleteForPath(Path<Name> path) =>
      path.isEmpty ? isFullyInitialized && !isFiltered
          : isCompleteForChild(path.first);

  /**
   * Checks if the value for a child is complete
   */
  bool isCompleteForChild(Name key) =>
      isFullyInitialized && !isFiltered || node.children.containsKey(key);

}

/**
 * Represents the current known state of a node in a tree.
 */
class ViewCache {
  final CacheNode eventCache;
  final CacheNode serverCache;

  const ViewCache(this.eventCache, this.serverCache);

  static final empty = new ViewCache(
      new CacheNode(Node.nil, false, false),
      new CacheNode(Node.nil, false, false));

  ViewCache updateEventSnap(Node eventSnap, bool complete, bool filtered) =>
      new ViewCache(new CacheNode(eventSnap, complete, filtered), this.serverCache);


  ViewCache updateServerSnap(Node serverSnap, bool complete, bool filtered) =>
      new ViewCache(this.eventCache, new CacheNode(serverSnap, complete, filtered));


  Node get completeEventSnap => eventCache.isFullyInitialized ? eventCache.node : null;

  Node get completeServerSnap => serverCache.isFullyInitialized ? serverCache.node : null;

}

typedef EventCallback(Event event);
typedef CancelCallback(Exception reason);



/**
 * Contains the current known state of a node in a tree and the event
 * registrations for a particular query.
 *
 */
class View {

  /**
   * The query for which this is a view
   */
  final Query query;

  final EventGenerator _eventGenerator;

  ViewCache _viewCache = ViewCache.empty;

  /**
   * The known remote and local state
   */
  ViewCache get viewCache => _viewCache;

  void set viewCache(ViewCache value) {
    _viewCache = value;
  }

  List<EventRegistration> _eventRegistrations = [];

  View(Query query, ViewCache initialViewCache) :
        query = query, _eventGenerator = new EventGenerator(query) {
    var params = query.queryParams;
    var indexFilter = new IndexedFilter(params.index);
    var filter = params.nodeFilter;
    var initialServerCache = initialViewCache.serverCache;
    var initialEventCache = initialViewCache.eventCache;
    var serverSnap = indexFilter.updateFullNode(Node.nil, initialServerCache.node);
    var eventSnap = filter.updateFullNode(Node.nil, initialEventCache.node);
    var newServerCache = new CacheNode(serverSnap, initialServerCache.isFullyInitialized, indexFilter.filtersNodes);
    var newEventCache = new CacheNode(eventSnap, initialEventCache.isFullyInitialized, filter.filtersNodes);
    this._viewCache = new ViewCache(newEventCache, newServerCache);
  }

  /**
   *
   */
  Node getCompleteServerCache(Path path) {
    var cache = _viewCache.completeServerSnap;
    if (cache!=null) {
      if (this.query.queryParams.loadsAllData || !path.isEmpty
          && !cache.children[path.first].isEmpty) {
        return cache.subtree(path);
      }
    }
    return null;
  }

  bool get isEmpty => _eventRegistrations.isEmpty;

  Iterable<Change> get initialChanges sync* {
    var eventSnap = _viewCache.eventCache;
    if (!eventSnap.node.isLeaf) {
      yield* eventSnap.node.orderedByPriority
          .map((pair)=>new Change.childAdded(pair.key, pair.value));
    }
    if (eventSnap.isFullyInitialized) {
      yield new Change.value(eventSnap.node);
    }
  }

  Iterable<Event> getInitialEvents(EventRegistration registration) =>
      _generateEventsForChanges(initialChanges, _viewCache.eventCache.node, registration);

  Iterable<Event> addEventRegistration(EventRegistration eventRegistration) {
    _eventRegistrations.add(eventRegistration);
    return getInitialEvents(eventRegistration);
  }

  void cancelAllEventRegistrations(Exception cancelError) {
    var path = this.query.path;
    _eventRegistrations.forEach((registration) {
      var maybeEvent = registration.createCancelEvent(cancelError, path);
      if (maybeEvent!=null) {
        cancelEvents.add(maybeEvent);
      }
    });
  }

  void removeEventRegistration(EventRegistration eventRegistration) {
    _eventRegistrations = eventRegistration==null ? []
        : _eventRegistrations.where((r)=>!r.matches(eventRegistration));
  }

  /**
   * Applies an operation to this view
   */
  Event applyOperation(Operation operation, WriteTreeRef writesCache) {
    var result = operation.apply(this, writesCache);
    this.viewCache = result.newViewCache;
    return _generateEventsForChanges([result.valueChange],
        result.newViewCache.eventCache.node);
  }


  Iterable<Change> get childAddedEvents sync* {
    var eventSnap = this._viewCache.eventCache;
    if (!eventSnap.node.isLeaf) {
      yield* eventSnap.node.children.keys.map(
          (key) => new Change.childAdded(key, eventSnap.node.children[key]));
    }
  }

  Change get valueEvent {
    var eventSnap = this._viewCache.eventCache;
    if (eventSnap.isFullyInitialized) {
      return new Change.value(eventSnap.node);
    }
    return null;
  }

  _generateEventsForChanges(List<Change> changes, Node eventCache, [EventRegistration eventRegistration]) {
    var registrations = eventRegistration!=null ? [eventRegistration] : _eventRegistrations;
    return this._eventGenerator.generateEventsForChanges(changes, eventCache, registrations);
  }


}

/**
 * A SyncPoint represents a node in a SyncTree.
 *
 * It contains the different views of the data for different queries for that
 * location.
 *
 */
class SyncPoint extends MapView<Query, View> {

  SyncPoint() : super({});

  /**
   * Applies an operation to the SyncPoint.
   *
   * When the operation source references a particular query, the operation is
   * only applied to the corresponding view. Otherwise, the operation is
   * applied to all views.
   *
   */
  Iterable<Event> applyOperation(Operation operation, WriteTreeRef writesCache,
      [Node completeServerCache]) sync* {
    print("SyncPoint::applyOperation $operation");
    var queryId = null;// operation.source.queryId;
    if (queryId != null) {
      var view = this[queryId];

      assert(view != null); //, "SyncTree gave us an op for an invalid query.");
      yield view.applyOperation(
          operation, writesCache, completeServerCache);
    } else {
      yield* values.map((view) => view.applyOperation(operation, writesCache));
    }
  }

  Iterable<Event> addEventRegistration(Query query, EventRegistration eventRegistration,
      WriteTreeRef writesCache, Node serverCache, bool serverCacheComplete) {

    var view = putIfAbsent(query, () {

      var eventCache = writesCache.calcCompleteEventCache(
          serverCacheComplete ? serverCache : null);
      var eventCacheComplete = eventCache != null;
      eventCache ??= serverCache.isLeaf ? Node.nil : writesCache.calcCompleteEventChildren(serverCache);

      var viewCache = new ViewCache(
          new CacheNode(eventCache, eventCacheComplete, false),
          new CacheNode(serverCache, serverCacheComplete, false));
      return new View(query, viewCache);
    });
    return view.addEventRegistration(eventRegistration);
  }

  void removeEventRegistration(Query query, EventRegistration eventRegistration, Exception cancelError) {
    var removed = <Query>[];
    var cancelEvents = [];
    var hadCompleteView = hasCompleteView;
    if (query.queryParams.isDefault) {
      forEach((viewQuery, view) {
        cancelEvents.addAll(
            view.removeEventRegistration(eventRegistration, cancelError));
        if (view.isEmpty) {
          remove(viewQuery);
          if (!view.query.queryParams.loadsAllData) {
            removed.add(view.query);
          }
        }
      });
    } else {
      var view = this[query];
      if (view!=null) {
        cancelEvents.addAll(
            view.removeEventRegistration(eventRegistration, cancelError));
        if (view.isEmpty) {
          remove(query);
          if (!view.query.queryParams.loadsAllData) {
            removed.add(view.query);
          }
        }
      }
    }
    if (hadCompleteView && !hasCompleteView) {
      removed.add(new Firebase.fromRepo(query.repo, query.path));
    }
    return new RemovedAndEvents(removed, cancelEvents);
  }


  Iterable<View> get queryViews => values.where((view)=>!view.query.queryParams.loadsAllData);

  Node getCompleteServerCache(Path path) =>
    values.map((view)=>view.getCompleteServerCache(path)).firstWhere((v)=>v!=null, orElse: ()=>null);

  View viewForQuery(Query query) => query.queryParams.loadsAllData ? completeView : this[query];

  bool viewExistsForQuery(Query query) => viewForQuery(query) != null;

  bool get hasCompleteView => completeView !=null;

  View get completeView => values.firstWhere((view)=>view.query.queryParams.loadsAllData, orElse: ()=>null);

}

/**
 * A SyncTree represents a local view on a remote tree structured database.
 *
 * It contains a partial local copy of the database as well as pending user
 * operations and event registrations.
 *
 */
class SyncTree {

  /**
   * A Repo representing the remote database
   */
  final Repo repo;

  SyncTree(this.repo);

  Tree<Name, SyncPoint> _syncPointTree = new Tree();
  WriteTree _pendingWriteTree = new WriteTree();

  Node _getCompleteChildren(Path<Name> path) {
    var subtree = this._syncPointTree.subtree(path);
    if (subtree==null) return Node.nil;

    var children = {};
    subtree.children.forEach((name, Tree<Name,SyncPoint> tree) {
      var n = tree.value.getCompleteServerCache(Path.empty);
      if (n!=null) {
        children[name] = n;
      }
    });
    return new Node.nonLeaf(children);
  }

  /**
   * Adds an event listener for a particular query
   */
  Iterable<Event> addEventRegistration(Query query, EventRegistration eventRegistration) sync* {
    print("SyncTree::addEventRegistration");
    var path = query.path;

    // get complete server cache from ancestors
    var serverCache = _syncPointTree.pairsOnPath(path)
    .map((p)=>p.value.getCompleteServerCache(p.key.relativeFrom(path)))
    .firstWhere((c)=>c!=null, orElse: ()=>null);

    // checks if ancestor has complete view
    var foundAncestorDefaultView = _syncPointTree.pairsOnPath(path).any((p)=>p.value.hasCompleteView);

    var syncPoint = _syncPointTree[path];
    if (syncPoint==null) {
      syncPoint = _syncPointTree[path] = new SyncPoint();
    } else {
      foundAncestorDefaultView = foundAncestorDefaultView || syncPoint.hasCompleteView;
      serverCache = serverCache || syncPoint.getCompleteServerCache(Path.empty);
    }

    var serverCacheComplete = serverCache!=null;
    if (!serverCacheComplete) {
      serverCache = Node.nil;
      var subtree = _syncPointTree.subtree(path);
      subtree.children.forEach((childName, childTree) {
        var completeCache = childTree.value.getCompleteServerCache(Path.empty);
        if (completeCache!=null) {
          serverCache = serverCache.updateImmediateChild(childName, completeCache);
        }
      });
    }


    var viewAlreadyExists = syncPoint.viewExistsForQuery(query);
    var writesCache = this._pendingWriteTree.childWrites(path);
    yield* syncPoint.addEventRegistration(query, eventRegistration,
        writesCache, serverCache, serverCacheComplete);
    if (!viewAlreadyExists && !foundAncestorDefaultView) {
      var view = (syncPoint.viewForQuery(query));
      yield * _installListener(query, view);
    }

/*
    Pair<Path,SyncPoint> completePair = _syncPointTree.pairsOnPath(path)
        .firstWhere((p)=>p.value?.hasCompleteView ?? false, orElse: ()=>null);

    Node serverCache = completePair?.value
        ?.getCompleteServerCache(path.relativeFrom(completePair.key));
    var serverCacheComplete = serverCache!=null;
    serverCache ??= _getCompleteChildren(path);

    var writesCache = _pendingWriteTree.childWrites(path);


    var syncPoint = _syncPointTree[path] ??= new SyncPoint();
    var viewAlreadyExists = syncPoint.viewExistsForQuery(query);

    yield* syncPoint.addEventRegistration(
        query, eventRegistration, writesCache, serverCache,
        serverCacheComplete);

    if (!viewAlreadyExists && !foundAncestorDefaultView) {
      var view = (syncPoint.viewForQuery(query));
      events = events.concat(this.setupListener_(query, view));
    }
    return events;
    _installListener(query, syncPoint.viewForQuery(query));
*/
  }


  void applyUserOverwrite(Path path, Node newData, int writeId, [bool visible = true]) {
    _pendingWriteTree.addOverwrite(path, newData, writeId, visible);
    if (visible) {
      this._applyOperationToSyncPoints(
          new Overwrite(OperationSource.user, path, newData));
    }
  }

  /**
   * Apply an overwrite of the data at path with the value newData, received
   * from the remote database
   *
   */
  Iterable<Event> applyServerOverwrite(Path path, Node newData) =>
    _applyOperationToSyncPoints(
        new Overwrite(OperationSource.server, path, newData));


  /**
   * Applies an operation to all the appropriate nodes in the sync tree
   *
   */
  Iterable<Event> _applyOperationToSyncPoints(Operation operation) =>
    _applyOperationHelper(operation, _syncPointTree, null,
        _pendingWriteTree.childWrites(Path.empty));

  /**
   * Applies an operation to all the appropriate nodes in a subtree of the
   * sync tree.
   *
   */
  Iterable<Event> _applyOperationHelper(Operation operation,
      Tree<Name,SyncPoint> syncPointTree, Node serverCache,
      WriteTreeRef writesCache) sync* {
    print("SyncTree::_applyOperationHelper $operation");

    if (operation.path.isEmpty) {
      yield* _applyOperationDescendantsHelper(
          operation, syncPointTree, serverCache, writesCache);
    } else {

      var syncPoint = syncPointTree[Path.empty];

      if (serverCache == null && syncPoint != null) {
        serverCache = syncPoint.getCompleteServerCache(Path.empty);
      }
      var childName = operation.path.first;
      var childOperation = operation.operationForChild(childName);
      var childTree = syncPointTree.children[childName];
      if (childTree!=null && childOperation!=null) {
        var childServerCache = serverCache==null ? null : serverCache.children[childName];
        var childWritesCache = writesCache.child(childName);
        yield* this._applyOperationHelper(
            childOperation, childTree, childServerCache, childWritesCache);
      }
      if (syncPoint!=null) {
        print("SyncTree::_applyOperationHelper2 $operation");
        yield syncPoint.applyOperation(operation, writesCache, serverCache);
      }
    }
  }

  Iterable<Event> _applyOperationDescendantsHelper(Operation operation,
      Tree<Name, SyncPoint> syncPointTree, Node serverCache,
      WriteTreeRef writesCache) sync* {
    print("SyncTree::_applyOperationDescendantsHelper $operation");

    var syncPoint = syncPointTree.value;

    if (serverCache == null && syncPoint != null) {
      serverCache = syncPoint.getCompleteServerCache(Path.empty);
    }

    syncPointTree.children.forEach((childName, childTree) {

      var childServerCache = (serverCache?.children ?? Node.nil)[childName];
      var childWritesCache = writesCache.child(childName);

      var childOperation = operation.operationForChild(childName);
      if (childOperation!=null) {

        return this._applyOperationDescendantsHelper(
            childOperation, childTree, childServerCache, childWritesCache);
      }
    });

    print("SyncTree::_applyOperationDescendantsHelper2 $operation");
    yield syncPoint?.applyOperation(operation, writesCache, serverCache);
  }


  final Map<Query,StreamSubscription> _subscriptions = {};
  Iterable<Event> _installListener(Query query, View view) sync* {
    _subscriptions.putIfAbsent(query, () {
      return repo.backend.changes(query).listen((Change change) {
        switch (change.type) {
          case ChangeType.VALUE:
            repo.eventQueue.raiseEventsForChangedPath(
                query.path,
                applyServerOverwrite(query.path, change.snapshotNode));
            break;
          default:

        }
      });
    });

    // TODO check already listening
    // TODO unlisten to others
  }

}


class WriteRecord {

  final Path<Name> path;
  final Map<Name,Node> children;
  final int writeId;
  final bool visible;
  final Node snap;

  WriteRecord({this.path, this.children, this.writeId, this.visible, this.snap});

  bool containsPath(Path<Name> path) {
    if (snap!=null) {
      return path.contains(path);
    } else {
      return children.keys.any((childName) {
        return path.child(childName).contains(path);
      });
    }
  }
}

class WriteTree {

  CompoundWrite _visibleWrites = CompoundWrite.empty;
  final SplayTreeMap<int, WriteRecord> _allWrites = new SplayTreeMap();
  int _lastWriteId = -1;


  WriteTreeRef childWrites(Path<Name> path) {
    return new WriteTreeRef(path, this);
  }

  void addOverwrite(Path<Name> path, Node snap, int writeId, [bool visible = true]) {
    assert(writeId > this._lastWriteId); //, "Stacking an older write on top of newer ones");

    this._allWrites[writeId] = new WriteRecord(
        path:path, snap:snap, writeId:writeId, visible:visible);

    if (visible) {
      this._visibleWrites = this._visibleWrites.addWrite(path, snap);
    }
    this._lastWriteId = writeId;
  }

  void addMerge(Path<Name> path, Map<Name,Node> changedChildren, int writeId) {
    assert(writeId > this._lastWriteId); //, "Stacking an older merge on top of newer ones");

    this._allWrites[writeId] = new WriteRecord(path:path, children:changedChildren,
        writeId:writeId, visible:true);

    this._visibleWrites = this._visibleWrites.addWrites(path, changedChildren);
    this._lastWriteId = writeId;
  }

  WriteRecord getWrite(int writeId) => _allWrites[writeId];


  bool removeWrite(int writeId) {
    var writeToRemove = _allWrites.remove(writeId);

    var removedWriteWasVisible = writeToRemove.visible;
    var removedWriteOverlapsWithOtherWrites = false;

    var i = this._allWrites.lastKey();
    while (removedWriteWasVisible && i >= _allWrites.firstKey()) {
      var currentWrite = this._allWrites[i];
      if (currentWrite.visible) {
        if (i >= writeId && currentWrite.containsPath(writeToRemove.path)) {
          removedWriteWasVisible = false;
        } else {
          if (writeToRemove.path.contains(currentWrite.path)) {
            removedWriteOverlapsWithOtherWrites = true;
          }
        }
      }
      i = _allWrites.lastKeyBefore(i);
    }
    if (!removedWriteWasVisible) {
      return false;
    } else {
      if (removedWriteOverlapsWithOtherWrites) {
        this.resetTree_();
        return true;
      } else {
        if (writeToRemove.snap) {
          this._visibleWrites =
              this._visibleWrites.removeWrite(writeToRemove.path);
        } else {
          var children = writeToRemove.children;
          var self = this;
          children.forEach((childName, childSnap) {
            self._visibleWrites = self._visibleWrites.removeWrite(
                writeToRemove.path.child(childName));
          });
        }
        return true;
      }
    }
  }

  Node getCompleteWriteData(Path path) {
    return this._visibleWrites.getCompleteNode(path);
  }

  Node calcCompleteEventCache(Path<Name> treePath, Node completeServerCache,
      [List<int> writeIdsToExclude = null, bool includeHiddenWrites = false]) {

    if (writeIdsToExclude==null && !includeHiddenWrites) {
      var shadowingNode = this._visibleWrites.getCompleteNode(treePath);
      if (shadowingNode != null) {
        return shadowingNode;
      } else {
        var subMerge = this._visibleWrites.childCompoundWrite(treePath);
        if (subMerge.isEmpty) {
          return completeServerCache;
        } else {
          if (completeServerCache == null &&
              !subMerge.hasCompleteWrite(Path.empty)) {
            return null;
          } else {
            var layeredCache = completeServerCache ?? Node.nil;
            return subMerge.apply(layeredCache);
          }
        }
      }
    } else {
      var merge = this._visibleWrites.childCompoundWrite(treePath);
      if (!includeHiddenWrites && merge.isEmpty) {
        return completeServerCache;
      } else {
        if (!includeHiddenWrites && completeServerCache == null &&
            !merge.hasCompleteWrite(Path.empty)) {
          return null;
        } else {
          var filter =
              (WriteRecord write) {
            return (write.visible || includeHiddenWrites) &&
                (writeIdsToExclude==null || writeIdsToExclude.contains(write.writeId)) &&
                (write.path.contains(treePath) ||
                    treePath.contains(write.path));
          };
          var mergeAtPath = WriteTree.layerTree_(
              this._allWrites, filter, treePath);
          var layeredCache = completeServerCache ?? Node.nil;
          return mergeAtPath.apply(layeredCache);
        }
      }
    }
  }

  Node calcCompleteEventChildren(Path treePath, [Node completeServerChildren]) {
    var completeChildren = Node.nil;
    var topLevelSet = this._visibleWrites.getCompleteNode(treePath);
    if (topLevelSet!=null) {
      if (!topLevelSet.isLeaf) {
        topLevelSet.children.forEach((childName, childSnap) {
          completeChildren =
              completeChildren.updateImmediateChild(childName, childSnap);
        });
      }
      return completeChildren;
    } else {
      if (completeServerChildren!=null) {
        var merge = this._visibleWrites.childCompoundWrite(treePath);
        completeServerChildren.children.forEach((childName, childNode) {
          var node = merge.childCompoundWrite(new Path([childName])).apply(
              childNode);
          completeChildren =
              completeChildren.updateImmediateChild(childName, node);
        });
        merge.getCompleteChildren().forEach((namedNode) {
          completeChildren = completeChildren.updateImmediateChild(
              namedNode.name, namedNode.node);
        });
        return completeChildren;
      } else {
        var merge = this._visibleWrites.childCompoundWrite(treePath);
        merge.getCompleteChildren().forEach((namedNode) {
          completeChildren = completeChildren.updateImmediateChild(
              namedNode.name, namedNode.node);
        });
        return completeChildren;
      }
    }
  }

  Node calcEventCacheAfterServerOverwrite(Path treePath, Path childPath,
      Node existingEventSnap, Node existingServerSnap) {
    assert(existingEventSnap!=null ||
        existingServerSnap!=null); //"Either existingEventSnap or existingServerSnap must exist");
    var path = treePath.child(childPath);
    if (this._visibleWrites.hasCompleteWrite(path)) {
      return null;
    } else {
      var childMerge = this._visibleWrites.childCompoundWrite(path);
      if (childMerge.isEmpty) {
        return existingServerSnap.subtree(childPath);
      } else {
        return childMerge.apply(existingServerSnap.subtree(childPath));
      }
    }
  }

  Node calcCompleteChild(Path treePath, Name childKey, CacheNode existingServerSnap) {
    var path = treePath.child(childKey);
    var shadowingNode = this._visibleWrites.getCompleteNode(path);
    if (shadowingNode != null) {
      return shadowingNode;
    } else {
      if (existingServerSnap.isCompleteForChild(childKey)) {
        var childMerge = this._visibleWrites.childCompoundWrite(path);
        return childMerge.apply(
            existingServerSnap.node.children[childKey]);
      } else {
        return null;
      }
    }
  }

  Node shadowingWrite(Path path) {
    return this._visibleWrites.getCompleteNode(path);
  }

  List<Node> calcIndexedSlice(Path treePath, Node completeServerData,
      String startPost, int count, bool reverse, Index index) {
    Node toIterate;
    var merge = this._visibleWrites.childCompoundWrite(treePath);
    var shadowingNode = merge.getCompleteNode(Path.empty);
    if (shadowingNode != null) {
      toIterate = shadowingNode;
    } else {
      if (completeServerData != null) {
        toIterate = merge.apply(completeServerData);
      } else {
        return [];
      }
    }
    toIterate = toIterate.withIndex(index);
    if (!toIterate.isEmpty && !toIterate.isLeaf) {
      List<Node> nodes = [];
      var cmp = index.compare;
      var iter = reverse
          ? toIterate.reverseIteratorFrom(startPost, index)
          : toIterate.iteratorFrom(startPost, index);
      var next = iter.getNext();
      while (next && nodes.length < count) {
        if (cmp(next, startPost) != 0) {
          nodes.add(next);
        }
        next = iter.getNext();
      }
      return nodes;
    } else {
      return [];
    }
  }



  void resetTree_() {
    this._visibleWrites = layerTree_(
        this._allWrites, DefaultFilter_, Path.empty);
    if (this._allWrites.length > 0) {
      this._lastWriteId = this._allWrites[this._allWrites.length - 1].writeId;
    } else {
      this._lastWriteId = -1;
    }
  }

  static bool DefaultFilter_(WriteRecord write) {
    return write.visible;
  }

  static layerTree_(List<WriteRecord> writes, bool filter(WriteRecord r),
      Path treeRoot) {
    var compoundWrite = CompoundWrite.empty;
    for (var i = 0; i < writes.length; ++i) {
      var write = writes[i];
      if (filter(write)) {
        var writePath = write.path;
        util.Path relativePath;
        if (write.snap) {
          if (treeRoot.contains(writePath)) {
            relativePath = util.Path.relativePath(treeRoot, writePath);
            compoundWrite = compoundWrite.addWrite(relativePath, write.snap);
          } else {
            if (writePath.contains(treeRoot)) {
              relativePath = util.Path.relativePath(writePath, treeRoot);
              compoundWrite = compoundWrite.addWrite(
                  util.Path.nil, write.snap.getChild(relativePath));
            } else {}
          }
        } else {
          if (write.children) {
            if (treeRoot.contains(writePath)) {
              relativePath = util.Path.relativePath(treeRoot, writePath);
              compoundWrite =
                  compoundWrite.addWrites(relativePath, write.children);
            } else {
              if (writePath.contains(treeRoot)) {
                relativePath = util.Path.relativePath(writePath, treeRoot);
                if (relativePath.isEmpty) {
                  compoundWrite =
                      compoundWrite.addWrites(util.Path.nil, write.children);
                } else {
                  var child = write.children[relativePath.front];
                  if (child) {
                    var deepNode = child.getChild(relativePath.popFront());
                    compoundWrite =
                        compoundWrite.addWrite(util.Path.nil, deepNode);
                  }
                }
              } else {}
            }
          } else {
            throw new Exception(
                "WriteRecord should have .snap or .children");
          }
        }
      }
    }
    return compoundWrite;
  }
}

class WriteTreeRef {
  final treePath_;
  final WriteTree writeTree_;

  WriteTreeRef(this.treePath_, this.writeTree_);

  Node calcCompleteEventCache(Node completeServerCache, [List<int> writeIdsToExclude = const[],
  bool includeHiddenWrites = false]) {
    return this.writeTree_.calcCompleteEventCache(
        this.treePath_, completeServerCache, writeIdsToExclude,
        includeHiddenWrites);
  }

  Node calcCompleteEventChildren(Node completeServerChildren) {
    return this.writeTree_.calcCompleteEventChildren(
        this.treePath_, completeServerChildren);
  }

  Node calcEventCacheAfterServerOverwrite(Path path, Node existingEventSnap,
      Node existingServerSnap) {
    return this.writeTree_.calcEventCacheAfterServerOverwrite(
        this.treePath_, path, existingEventSnap, existingServerSnap);
  }

  Node shadowingWrite(Path path) {
    return this.writeTree_.shadowingWrite(this.treePath_.child(path));
  }

  List<Node> calcIndexedSlice(Node completeServerData, String startPost,
      int count, bool reverse, Index index) {
    return this.writeTree_.calcIndexedSlice(
        this.treePath_, completeServerData, startPost, count, reverse, index);
  }

  Node calcCompleteChild(Name childKey, CacheNode existingServerCache) {
    return this.writeTree_.calcCompleteChild(
        this.treePath_, childKey, existingServerCache);
  }

  WriteTreeRef child(Name childName) {
    return new WriteTreeRef(
        this.treePath_.child(childName), this.writeTree_);
  }
}

class CompoundWrite {
  final UnmodifiableTree<Name, Node> writeTree_;

  const CompoundWrite(this.writeTree_);

  static final empty = new CompoundWrite(new UnmodifiableTree(null));

  CompoundWrite addWrite(Path path, Node node) {
    if (path.isEmpty) {
      return new CompoundWrite(new UnmodifiableTree(node));
    } else {
      var rootmost = this.writeTree_.findRootMostValueAndPath(path);
      if (rootmost != null) {
        var rootMostPath = rootmost.path, value = rootmost.value;
        var relativePath = Path.relativePath(rootMostPath, path);
        value = value.updateChild(relativePath, node);
        return new CompoundWrite(this.writeTree_.set(rootMostPath, value));
      } else {
        var subtree = new UnmodifiableTree(node);
        var newWriteTree = this.writeTree_.setTree(path, subtree);
        return new CompoundWrite(newWriteTree);
      }
    }
  }

  CompoundWrite addWrites(Path<Name> path, Map<Name,Node> updates) {
    var newWrite = this;
    updates.forEach((childKey, node) {
      newWrite = newWrite.addWrite(path.child(childKey), node);
    });
    return newWrite;
  }

  CompoundWrite removeWrite(Path path) {
    if (path.isEmpty) {
      return CompoundWrite.empty;
    } else {
      var newWriteTree = this.writeTree_.setTree(path, ImmutableTree.nil);
      return new CompoundWrite(newWriteTree);
    }
  }

  bool hasCompleteWrite(Path path) {
    return this.getCompleteNode(path) != null;
  }

  Node getCompleteNode(Path path) {
    var rootmost = this.writeTree_.pairsOnPath(path)
        .firstWhere((pair)=>pair.value!=null, orElse: ()=>null);
    if (rootmost != null) {
      return this.writeTree_.get(rootmost.path).getChild(util.Path.relativePath(rootmost.path, path));
    } else {
      return null;
    }
  }

  List<NamedNode> getCompleteChildren() {
    List<NamedNode> children = [];
    var node = this.writeTree_.value;
    if (node != null) {
      if (!node.isLeafNode) {
        node = (node);
        node.forEachChild(priorityIndex, (childName, childNode) {
          children.add(new NamedNode(childName, childNode));
        });
      }
    } else {
      this.writeTree_.children.forEach((childName, childTree) {
        if (childTree.value != null) {
          children.add(new NamedNode(childName, childTree.value));
        }
      });
    }
    return children;
  }

  CompoundWrite childCompoundWrite(Path path) {
    if (path.isEmpty) {
      return this;
    } else {
      var shadowingNode = this.getCompleteNode(path);
      if (shadowingNode != null) {
        return new CompoundWrite(new UnmodifiableTree(shadowingNode));
      } else {
        return new CompoundWrite(this.writeTree_.subtree(path) ?? new UnmodifiableTree(null));
      }
    }
  }

  bool get isEmpty {
    return this.writeTree_.isEmpty;
  }

  Node apply(Node node) {
    return CompoundWrite._applySubtreeWrite(Path.empty, this.writeTree_, node);
  }

  static Node _applySubtreeWrite(Path relativePath,
      UnmodifiableTree<Name, Node> writeTree, Node node) {
    if (writeTree.value != null) {
      return node.updateChild(relativePath, writeTree.value);
    } else {
      var priorityWrite = null;
      writeTree.children.forEach((childKey, childTree) {
        if (childKey == ".priority") {
          assert(childTree.value != null);//, "Priority writes must always be leaf nodes");
          priorityWrite = childTree.value;
        } else {
          node = CompoundWrite._applySubtreeWrite(relativePath.child(childKey), childTree, node);
        }
      });
      if (!node.subtree(relativePath).isEmpty && priorityWrite != null) {
        node = node.updateChild(relativePath.child(".priority"), (priorityWrite));
      }
      return node;
    }
  }

}

enum ChangeType {CHILD_ADDED, CHILD_CHANGED, CHILD_MOVED, CHILD_REMOVED,VALUE}

class Change {

  final ChangeType type;
  final Node snapshotNode;
  final Name childName;
  final Node oldSnap;
  final String prevName;

  const Change(this.type, this.snapshotNode, {this.childName, this.oldSnap, this.prevName});

  const Change.value(Node snapshot) : this(ChangeType.VALUE, snapshot);
  const Change.childAdded(Name childKey, Node snapshot) :
        this(ChangeType.CHILD_ADDED, snapshot, childName: childKey);

  const Change.childRemoved(Name childKey, Node snapshot) :
        this(ChangeType.CHILD_REMOVED, snapshot, childName: childKey);

  const Change.childChanged(Name childKey, Node newSnapshot, Node oldSnapshot) :
        this(ChangeType.CHILD_CHANGED, newSnapshot, childName: childKey, oldSnap: oldSnapshot);

  const Change.childMoved(Name childKey, Node snapshot) :
        this(ChangeType.CHILD_MOVED, snapshot, childName: childKey);

  Change changeWithPrevName(String prevName) {
    return new Change(this.type, this.snapshotNode, childName: this.childName,
        oldSnap: this.oldSnap, prevName: prevName);
  }
}
