
import 'dart:collection';

class Pair<K,V> {
  final K key;
  final V value;

  Pair(this.key, this.value);
}

class Path<T> extends UnmodifiableListView<T> {

  static final empty = new Path();

  Path([Iterable<T> pieces = const []]) : super(pieces);

  Path skip(int n) => new Path(super.skip(n));

  String toString() => "/"+join("/");

  String toUrlEncodedString() => "/"+map((p)=>Uri.encodeComponent(p.toString())).join("/");

  Path<T> get parent => new Path(this.take(length-1));

  Path<T> child(T child) => new Path(new List.from(this)..add(child));

  Path<T> append(Iterable<T> children) => new Path(new List.from(this)..addAll(children));


  Path<T> relativeFrom(Path<T> other) {
    if (!startsWith(other)) {
      throw new ArgumentError();
    }
    return skip(other.length);
  }

  bool operator==(Path other) {
    if (length != other.length) {
      return false;
    }
    for (var i = 0;i < length; i++) {
      if (this[i] != other[i]) {
        return false;
      }
    }
    return true;
  }

  int get hashCode => toUrlEncodedString().hashCode;

  bool startsWith(Path other) {
    if (length < other.length) {
      return false;
    }
    for (int i=0;i<other.length;i++) {
      if (this[i] != other[i]) {
        return false;
      }
    }
    return true;
  }
}

class UnmodifiableTree<K,V> extends Tree<K,V> {

  UnmodifiableTree(V value, [Map<K,Tree<K,V>> children = const{}]) :
        super._(value, new UnmodifiableMapView(children));

  @override
  set value(V value) => throw new UnsupportedError("Tree is unmodifiable");


  @override
  operator []=(Path<K> key, V value) => throw new UnsupportedError("Tree is unmodifiable");

  @override
  void clear() => throw new UnsupportedError("Tree is unmodifiable");

  @override
  V remove(Object key) => throw new UnsupportedError("Tree is unmodifiable");


}

class Tree<K,V> extends MapBase<Path<K>, V> {

  final Map<K,Tree<K,V>> children;
  V value;

  Tree._(this.value, this.children);

  Tree([V defaultValue]) : this._(null, {});

  @override
  V operator [](Object key) => key is! Path ? null : subtree(key)?.value;

  @override
  operator []=(Path<K> key, V value) {
    var c = this;
    for (var k in key) {
      c = c.children.putIfAbsent(k, ()=>new Tree());
    }
    c.value = value;
  }

  @override
  void clear() {
    children.clear();
    value = null;
  }

  @override
  Iterable<Path<K>> get keys => pairs.map((p)=>p.key);

  @override
  Iterable<V> get values => pairs.map((p)=>p.value);

  @override
  void forEach(fn(Path<K> k, V v)) => pairs.forEach((p)=>fn(p.key,p.value));

  @override
  V remove(Object key) {
    var c = this;
    for (var k in key) {
      if (!c.children.containsKey(k)) return null;
      c = c.children[k];
    }
    var v = c.value;
    c.value = null;
    return v;
  }

  Iterable<Pair<Path<K>,V>> get pairs => _getPairs(new Path<K>());

  Iterable<Pair<Path<K>,V>> _getPairs(Path<K> p) sync* {
    yield new Pair(p, value);
    for (var c in children.keys) {
      yield* children[c]._getPairs(p.child(c));
    }
  }

  bool get isEmpty => children.isEmpty;

  Tree<K,V> subtree(Path<K> p) => p.fold(this,
      (Tree<K,V> c, K k)=>c==null ? null : c.children[k]);

  Iterable<Pair<Path<K>,V>> pairsOnPath(Path<K> p) sync* {
    var c = this;
    var path = Path.empty;
    for (var n in p) {
      print("pairs on path $path $p");
      yield new Pair(path, c.value);
      c = c.children[n];
      path = path.child(n);
      if (c==null) return;
    }
  }

  Iterable<Pair<Path<K>,Tree<K,V>>> subtreesOnPath(Path<K> p) sync* {
    var c = this;
    var path = Path.empty;
    for (var n in p) {
      yield new Pair(path, c);
      c = c.children[n];
      path = path.child(n);
      if (c==null) return;
    }
  }

}

